/*hdr
**	FILE:		32bit.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		29/8/97
**	DESCRIPTION:	32 bit primitives
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

/****************************** Includes ************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"

/****************************** Classes *************************************************************************************/
#define AddPtr(p, i)		p = (ulong*) (((long)(p))+(i))

bool GdcApp32::SetSurface(GdcSurface *d, GdcPalette *p, GdcSurface *a)
{
	if (d AND d->Bits == 32)
	{
		Dest = d;
		Pal = p;
		Ptr = (ulong*) d->Base;
		Alpha = 0;
		return TRUE;
	}
	return FALSE;
}

void GdcApp32::SetPtr(int x, int y)
{
	ASSERT(Dest AND Dest->Base);
	Ptr = (ulong*) (Dest->Base + ((y * Dest->Line) + (x << 2)));
}

void GdcApp32::IncX()
{
	Ptr++;
}

void GdcApp32::IncY()
{
	AddPtr(Ptr, Dest->Line);
}

void GdcApp32::IncPtr(int X, int Y)
{
	AddPtr(Ptr, (Y * Dest->Line) + (X * 4));
}

COLOUR GdcApp32::Get()
{
	return *Ptr;
}

// 32 bit set sub functions
void GdcApp32Set::Set()
{
	*Ptr = c;
}

void GdcApp32Set::VLine(int height)
{
	while (height--)
	{
		*Ptr = c;
		AddPtr(Ptr, Dest->Line);
	}
}

void GdcApp32Set::Rectangle(int x, int y)
{
#ifdef GDC_USE_ASM

	ulong *p = Ptr;
	int Line = Dest->Line;
	COLOUR fill = c;

	if (x AND y)
	{
		_asm {
			mov esi, p
			mov eax, fill
			mov edx, Line
		LoopY:	mov edi, esi
			add esi, edx
			mov ecx, x
		LoopX:	mov [edi], eax
			add edi, 4
			dec ecx
			jnz LoopX
			dec y
			jnz LoopY
		}
	}
#else
	while (y--)
	{
		for (int n=0; n<x; n++)
		{
			*Ptr++ = c;
		}
		AddPtr(Ptr, Dest->Line - (x << 2));
	}
#endif
}

bool GdcApp32Set::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src)
	{
		switch (Src->Bits)
		{
			case 8:
			{
				uchar *s = Src->Base;
				if (SPal)
				{
					COLOUR c[256];
					for (int i=0; i<256; i++)
					{
						GdcRGB *p = (*SPal)[i];
						if (p)
						{
							c[i] = Rgb32(p->R, p->G, p->B);
						}
						else
						{
							c[i] = 0;
						}
					}

					for (int y=0; y<Src->y; y++)
					{
						for (int x=0; x<Src->x; x++)
						{
							*Ptr++ = c[*s++];
						}

						s += Src->Line - Src->x;
						AddPtr(Ptr, Dest->Line - (Src->x << 2));
					}
				}
				else
				{
					for (int y=0; y<Src->y; y++)
					{
						uchar *NextS = ((uchar*) s) + Src->Line;
						uchar *NextD = ((uchar*) Ptr) + Dest->Line;

						for (int x=0; x<Src->x; x++)
						{
							*Ptr++ = Rgb32(*s, *s, *s);
							s++;
						}
						
						s = NextS;
						Ptr = (ulong*) NextD;
					}
				}
				break;
			}
			case 16:
			{
				COLOUR c;
				ushort *s = (ushort*) Src->Base;

				for (int y=0; y<Src->y; y++)
				{
					uchar *NextS = ((uchar*) s) + Src->Line;
					uchar *NextD = ((uchar*) Ptr) + Dest->Line;
					
					for (int x=0; x<Src->x; x++)
					{
						c = *s++;
						*Ptr++ = Rgb16To32(c);
					}

					s = (ushort*) NextS;
					Ptr = (ulong*) NextD;
				}
				break;
			}
			case 24:
			{
				COLOUR c;
				uchar *s = Src->Base;

				for (int y=0; y<Src->y; y++)
				{
					uchar *NextS = ((uchar*) s) + Src->Line;
					uchar *NextD = ((uchar*) Ptr) + Dest->Line;
					
					for (int x=0; x<Src->x; x++)
					{
						*Ptr++ = Rgb32(s[2], s[1], s[0]);
						s += 3;
					}

					s = (uchar*) NextS;
					Ptr = (ulong*) NextD;
				}
				break;
			}
			case 32:
			{
				uchar *s = Src->Base;
				for (int y=0; y<Src->y; y++)
				{
					MemCpy(Ptr, s, Src->x << 2);
					s += Src->Line;
					AddPtr(Ptr, Dest->Line);
				}
				break;
			}
		}
	}
	return TRUE;
}

// 32 bit or sub functions
void GdcApp32Or::Set()
{
	*Ptr |= c;
}

void GdcApp32Or::VLine(int height)
{
	while (height--)
	{
		*Ptr |= c;
		AddPtr(Ptr, Dest->Line);
	}
}

void GdcApp32Or::Rectangle(int x, int y)
{
	while (y--)
	{
		for (int n=0; n<x; n++) *Ptr++ |= c;
		AddPtr(Ptr, Dest->Line - (x << 2));
	}
}

bool GdcApp32Or::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src)
	{
		switch (Src->Bits)
		{
			case 32:
			{
				uchar *s = Src->Base;
				for (int y=0; y<Src->y; y++)
				{
					MemOr(Ptr, s, Src->x << 2);
					s += Src->Line;
					AddPtr(Ptr, Dest->Line);
				}
				break;
			}
		}
	}
	return TRUE;
}

// 32 bit AND sub functions
void GdcApp32And::Set()
{
	*Ptr &= c;
}

void GdcApp32And::VLine(int height)
{
	while (height--)
	{
		*Ptr &= c;
		AddPtr(Ptr, Dest->Line);
	}
}

void GdcApp32And::Rectangle(int x, int y)
{
	while (y--)
	{
		for (int n=0; n<x; n++) *Ptr++ &= c;
		AddPtr(Ptr, Dest->Line - (x << 2));
	}
}

bool GdcApp32And::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src AND Src->Bits == Dest->Bits)
	{
		uchar *s = Src->Base;
		for (int y=0; y<Src->y; y++)
		{
			MemAnd(Ptr, s, Src->x << 2);
			s += Src->Line;
			AddPtr(Ptr, Dest->Line);
		}
	}
	return TRUE;
}

// 32 bit XOR sub functions
void GdcApp32Xor::Set()
{
	*Ptr ^= c;
}

void GdcApp32Xor::VLine(int height)
{
	while (height--)
	{
		*Ptr ^= c;
		AddPtr(Ptr, Dest->Line);
	}
}

void GdcApp32Xor::Rectangle(int x, int y)
{
	while (y--)
	{
		for (int n=0; n<x; n++) *Ptr++ ^= c;
		AddPtr(Ptr, Dest->Line - (x << 2));
	}
}

bool GdcApp32Xor::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src AND Src->Bits == Dest->Bits)
	{
		uchar *s = Src->Base;
		for (int y=0; y<Src->y; y++)
		{
			MemXor(Ptr, s, Src->x << 2);
			s += Src->Line;
			AddPtr(Ptr, Dest->Line);
		}
	}
	return TRUE;
}
