/*hdr
**      FILE:           GDevice.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           10/4/98
**      DESCRIPTION:    GDevice class
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

/****************************** Includes ********************************************/

#include <stdio.h>
#include <stdarg.h>

#include "defs.h"
#include "device.h"

/****************************** Classes *********************************************/

bool GDevice::Verbose = TRUE;
GDevice *GDevice::First = NULL;

GDevice::GDevice()
{
	AddDevice(this);
}

GDevice::~GDevice()
{
	RemoveDevice(this);
}

bool GDevice::AddDevice(GDevice *pDev)
{
	if (First)
	{
		GDevice *Temp = First;
		while (Temp->Next)
		{
			Temp = Temp->Next;
		}
		Temp->Next = pDev;
		pDev->Next = 0;
		pDev->Prev = Temp;
	}
	else
	{
		First = pDev;
		pDev->Next = 0;
		pDev->Prev = 0;
	}
	
	return TRUE;
}

bool GDevice::RemoveDevice(GDevice *pDev)
{
	if (First AND pDev)
	{
		GDevice *Temp = First;
		
		while (Temp AND Temp != pDev)
		{
			Temp = Temp->Next;
		}
		
		if (Temp == pDev)
		{
			if (pDev == First) First = pDev->Next;
			if (pDev->Next) pDev->Next->Prev = pDev->Prev;
			if (pDev->Prev) pDev->Prev->Next = pDev->Next;
			return TRUE;
		}
	}
	
	return FALSE;
}

GDevice *GDevice::FindDevice(char *Name)
{
	GDevice *Temp = First;
	while (Temp)
	{
		if (stricmp(Name, Temp->Name) == 0)
		{
			return Temp;
		}
		Temp = Temp->Next;
	}
	
	return 0;
}

bool GDevice::Print(char *str, ...)
{
	if (str)
	{
		if (Verbose)
		{
			// char buffer[256];
			va_list arg;

			va_start(arg, str);
			// vprintf(str, arg);
			va_end(arg);
		}
		
		return TRUE;
	}
	
	return FALSE;
}

bool GDevice::SetVerbose(bool v)
{
	bool Prev = Verbose;
	Verbose = v;
	return Prev;
}

void GDevice::CriticalExit(char *str, ...)
{
	if (str)
	{
		// char buffer[256];
		va_list arg;

		va_start(arg, str);
		vprintf(str, arg);
		va_end(arg);
		
	}

	exit(-1);
}


