/*hdr
**      FILE:           Base64.h
**      AUTHOR:         Matthew Allen
**      DATE:           1/6/98
**      DESCRIPTION:    Base64 encoding/decoding
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __BASE64_H_
#define __BASE64_H_

// These buffer length macros round up to the nearest block of
// bytes. So take the value returned by the convert routine as
// the actual value... but use these to allocate buffers.
#define BufferLen_64ToBin(l)		( ((l)*3)/4 )
#define BufferLen_BinTo64(l)		( ((((l)+2)/3)*4) )

// Character Conversion Routines
//
// Format of Base64 char:
// 7               0
// |-|-|-|-|-|-|-|-|
// |-U-|--- Data --|
//
// Data	= Bits, 0-63
// U	= Unused, must be 0
//
extern uchar Base64ToBin(char c);
extern char BinToBase64(uchar c);

// String Conversion Routines
//
// Arguments:
//	Binary:		Pointer to binary buffer
//	OutBuf:		Size of output buffer (in bytes)
//	Base64:		Pointer to Base64 buffer
//	InBuf:		Size of input buffer (in bytes)
//
// Returns:
//	Number of bytes converted.
//
extern int ConvertBase64ToBinary(uchar *Binary, int OutBuf, char *Base64, int InBuf);
extern int ConvertBinaryToBase64(char *Base64, int OutBuf, uchar *Binary, int InBuf);

#endif