/*hdr
**      FILE:           Contain.h
**      AUTHOR:         Matthew Allen
**      DATE:           27/11/96
**      DESCRIPTION:    Container class header
**
**      Copyright (C) 1996, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __CONTAIN_H
#define __CONTAIN_H

#define DEFAULT_ARRAY_BANK			16

typedef void *VoidPtr;

class PtrArray {

	int Used;
	int Size;
	int BankSize;
	void **Array;

public:
	PtrArray();
	virtual ~PtrArray();

	int GetSize() { return Used; }
	bool SetSize(int NewSize);
	bool SetBankSize(int NewBank);
	virtual bool Add(void *Ptr, int Where);
	virtual bool Remove(int Where);
	virtual bool Empty();

	VoidPtr &operator [](int i);
};

template <class ArrayType>
class Array : public PtrArray {
public:
	Array() {}
	~Array() {}

	bool SetSize(int Size, int Bank)
	{
		return PtrArray::SetSize(sizeof(ArrayType), Size, Bank);
	}

	bool Add(ArrayType *Obj, int Where)
	{
		return PtrArray::Add((char*) Obj, Where);
	}

	ArrayType *operator [](int i)
	{
		return (ArrayType*) (*((PtrArray*) this))[i];
	}
};

template <class StackType>
class Stack {

	class StackNode {
	public:
		StackNode *Next;
		StackType *Data;
	} *Top;

public:
	Stack();
	~Stack();

	void Push(StackType *Data);
	StackType *Pop();
	StackType *Peek();
	int IsEmpty();
	void Empty();
};

template <class QueType>
class Que {

	class QueNode {
	public:
		QueNode *Next;
		QueType *Data;
	} *First, *Last;

public:
	Que();
	~Que();

	void Insert(QueType *Data);
	QueType *Delete();
	QueType *Peek();
	int IsEmpty();
	void Empty();
};

#define DEFAULT_PQUE_SIZE			32

template <class PQueType>
class PriorityQue {

	class PQueNode {
	public:
		long Key;
		PQueType *Data;
	} *Top;

	long Nodes;
	long Size;
	
	void SetSize(long NewSize);
	void BubbleUp(long Start);
	void BubbleDown();

public:
	PriorityQue();
	~PriorityQue();

	void NewSize(long NewSize);
	void Insert(PQueType *Data,long Key);
	PQueType *DeleteMin();
	PQueType *Peek();
	int IsEmpty();
	void Empty();
};

class DLinkList;
class DLinkInterator;

class Item {
public:
	Item *Prev, *Next;
	void *Ptr;
};

class DLinkList {

	friend class DLinkIterator;

protected:
	friend class ListIterator;

	int Items;
	Item *FirstObj, *LastObj, *Cur;

	int FindIndex(void *p);

public:
	DLinkList();
	~DLinkList();

	int GetItems() { return Items; }
	virtual void Empty();
	void *operator [](int Index);

	bool Delete();
	bool Delete(int Index);
	bool Delete(void *p);
	bool Insert(void *p, int Index = -1);
	void *First();
	void *Last();
	void *Next();
	void *Prev();
	void *Find(void *p);
};

class DLinkIterator {
protected:
	DLinkList *List;
	Item *Cur;

public:
	DLinkIterator(DLinkList *list);
	~DLinkIterator();

	void *First();
	void *Last();
	void *Next();
	void *Prev();
	void *Find(void *p);
};

template <class Type>
class List : public DLinkList {
public:
	bool Delete() { return DLinkList::Delete(); }
	bool Delete(Type *p) { return DLinkList::Delete((void*)p); }
	bool Insert(Type *p, int Index = -1) { return DLinkList::Insert((void*)p, Index); }
	Type *First() { return (Type*) DLinkList::First(); }
	Type *Last() { return (Type*) DLinkList::Last(); }
	Type *Next() { return (Type*) DLinkList::Next(); }
	Type *Prev() { return (Type*) DLinkList::Prev(); }
	Type *Current() { return (Cur) ? (Type*) Cur->Ptr : 0; }
	Type *Find(Type *p) { return (Type*) DLinkList::Find((void*) p); }
	Type *operator [](int Index) { return ((Type*) ((*((DLinkList*)this))[Index])); }
};

template <class Type>
class Iterator : public DLinkIterator {
public:
	Iterator(DLinkList *list) : DLinkIterator(list) {}

	Type *First() { return (Type*) DLinkIterator::First(); }
	Type *Last() { return (Type*) DLinkIterator::Last(); }
	Type *Next() { return (Type*) DLinkIterator::Next(); }
	Type *Prev() { return (Type*) DLinkIterator::Prev(); }
	Type *Current() { return (Cur) ? (Type*) Cur->Ptr : 0; }
	Type *Find(Type *p) { return (Type*) DLinkIterator::Find((void*) p); }
	Type *operator [](int Index) { return ((Type*) ((*List)[Index])); }
};


class BytePipe {

	class Block {
	public:
		int Next;
		int Size;
		uchar *Ptr;

		Block()
		{
			Next = 0;
			Size = 0;
			Ptr = 0;
		}

		~Block()
		{
			Next = 0;
			Size = 0;
			DeleteArray(Ptr);
		}
	};

	List<Block> Mem;

public:
	BytePipe();
	~BytePipe();

	bool IsEmpty() { return Mem.GetItems() == 0; }
	void Empty();
	int Sizeof();
	bool Pop(uchar *Ptr, int Size);
	bool Push(uchar *Ptr, int Size);
};

#endif
