/*hdr
**      FILE:           GuiDlg.h
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Dialog components
**
**      Copyright (C) 1998 Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __DLG_H
#define __DLG_H

/////////////////////////////////////////////////////////////////////////////////////////
#define IDM_OK				99
#define IDM_CANCEL			98
#define IDM_STATIC			-1

#define CTRL_SUNKEN			0x80000000
#define CTRL_FLAT			0x40000000
#define CTRL_RAISED			0x20000000

/////////////////////////////////////////////////////////////////////////////////////////
class GDialog;
class GListItem;
class GTreeItem;

class GButton;
class GEdit;
class GCheckBox;
class GText;
class GRadioGroup;
class GRadioButton;
class GTabControl;
class GTabPage;
class GSlider;
class GCombo;
class GBitmap;
class GList;
class GTree;

///////////////////////////////////////////////////////////////////////////////////////////////
class GControlInfo : public GObject {
public:
};

class GDialog : public BView, public GWindow {

	friend class GControl;

private:
	bool IsResizable;
	sem_id ModalSem;
	int SizeOfControls();
	bool QuitRequested();

protected:
	void AttachedToWindow() { OnCreate(); }
	void Draw(BRect UpdateRect) { GdcViewDC DC(Sys_View); OnPaint(&DC); }
	void FrameMoved(BPoint Point) { Pos.Offset(Point.x-Pos.x1, Point.y-Pos.y1); OnPosChange(); }
	void FrameResized(float width, float height) { Pos.Dimension(width+1, height+1); OnPosChange(); }
	void Pulse() { OnPulse(); }
	void MessageReceived(BMessage *message) { OnEvent((GMessage*)message); }
	void MakeFocus(bool focused = true) { OnFocus(focused); }
	void KeyDown(const char *bytes, int32 numBytes) { Sys_KeyDown(bytes, numBytes); }
	void KeyUp(const char *bytes, int32 numBytes) { Sys_KeyUp(bytes, numBytes); }
	void MouseDown(BPoint point) { Sys_MouseDown(point); }
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) { Sys_MouseMoved(point, transit, message); }

	BWindow *Wnd;

	void MoveOnScreen();
	void MoveToCenter();
	void MoveToMouse();

	bool LoadFromResource(char *Resource);
	virtual GControl *CreateControl(GControlInfo *Info) { return 0; }
	
public:
	GDialog();
	~GDialog();

	virtual BWindow *GetBWindow() { return Wnd; }
	bool Resizable() { return IsResizable; }
	void Resizable(bool r) { IsResizable = r; }
	GControl *FindControl(int Id);
	
	virtual int DoModal();
	virtual void EndModal(int Code = 0);

	virtual int DoModeless();
	virtual void EndModeless(int Code = 0);

	int OnEvent(GMessage *Msg);
};

class GControl : public GWindow, public GId {

	friend class GDialog;
	friend class GControlGroup;

protected:
	bigtime_t Sys_LastClick;

	GdcPt2 SizeOfStr(char *Str);
	void MouseClickEvent(bool Down);

public:
	GControl(BView *view);
	~GControl();

	bool Sunken() { return (WndFlags & CTRL_SUNKEN) != 0; }
	void Sunken(bool n) { if (n) WndFlags |= CTRL_SUNKEN; else WndFlags &= ~CTRL_SUNKEN; }
	bool Flat() { return (WndFlags & CTRL_FLAT) != 0; }
	void Flat(bool n) { if (n) WndFlags |= CTRL_FLAT; else WndFlags &= ~CTRL_FLAT; }
	bool Raised() { return (WndFlags & CTRL_RAISED) != 0; }
	void Raised(bool n) { if (n) WndFlags |= CTRL_RAISED; else WndFlags &= ~CTRL_RAISED; }

	int OnEvent(GMessage *Msg);
	bool Attach(GWindow *parent);
	virtual GControl *IsControl() { return this; }
	virtual GControl *FindControl(int Id) { if (GetId() == Id) return this; else return 0; }
};

//////////////////////////////////////////////////////////////////////////////////////
// Controls

// Button
class GButton : public GControl, public BButton {

	status_t Invoke(BMessage *message = NULL);

public:
	GButton(int id, int x, int y, int cx, int cy, char *name);
	~GButton();
	
	bool Default() { return IsDefault(); }
	void Default(bool b) { MakeDefault(b); }
};

// Edit
class GEdit : public GControl, public BTextControl {
protected:
	bool MultiLineEdit;
	bool PasswordEdit;
	bool Enter;

public:
	GEdit(int id, int x, int y, int cx, int cy, char *name);
	~GEdit();

	bool WantReturn() { return Enter; }
	void WantReturn(bool b) { Enter = b; }
	bool MultiLine() { return MultiLineEdit; }
	void MultiLine(bool m) { MultiLineEdit = m; }
	bool Password() { return PasswordEdit; }
	void Password(bool m) { PasswordEdit = m; }

	void Value(int i);
	int Value();
	char *Name();
	bool Name(char *s);
	int OnEvent(GMessage *Msg);
};

// Check box
class GCheckBox : public GControl, public BCheckBox {

	int Val;
	bool Three;

public:
	GCheckBox(int id, int x, int y, int cx, int cy, char *name, int InitState = FALSE);
	~GCheckBox();

	bool ThreeState() { return Three; }
	void ThreeState(bool t) { Three = t; }

	int Value();
	void Value(int b);
	int OnEvent(GMessage *Msg);
};

// Static
class GText : public GControl, public BStringView {

	void Draw(BRect updateRect);

public:
	GText(int id, int x, int y, int cx, int cy, char *name);
	~GText();
	int OnEvent(GMessage *Msg);
};

// Group of radio buttons
typedef List<GRadioButton> RadioButtonList;

class GRadioGroup : public GControl, public BBox {

	static int NextId;

	bool Attach(GWindow *parent);

protected:
	RadioButtonList Buttons;
	int InitialValue;

public:
	GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init = 0);
	~GRadioGroup();

	int Value();
	void Value(int i);

	int OnEvent(GMessage *Msg);
	GRadioButton *Append(int x, int y, char *name);
};

class GRadioButton : public GControl, public BRadioButton {

	bool First;

public:
	GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first = FALSE);
	~GRadioButton();

	int Value();
	void Value(int i);
	int OnEvent(GMessage *Msg);

};

// Tab control
typedef List<GTabPage> TabPageList;
class GTabControl : public GControl, public BTabView {

	friend class GTabPage;

	bool Attach(GWindow *parent);
	GControl *FindControl(int Id);

	int Current;
	int NextTab;

protected:
	List<GTabPage> Tabs;
	int InitialValue;

public:
	GTabControl(int id, int x, int y, int cx, int cy, char *name, int Init = 0);
	~GTabControl();

	int Value();
	void Value(int i);

	void Select(int32 tab);
	int OnEvent(GMessage *Msg);
	GTabPage *Append(char *name);
};

class GTabPage : public GWindow, public BView {

	friend class GTabControl;
	GTabControl *Parent;
	GdcRegion Pos;
	WindowList Children;
	BTab *Tab;

	bool Attach(GWindow *parent);

public:
	GTabPage(char *name);
	~GTabPage();

	bool Enabled();
	void Enabled(bool e);
	bool Visible();
	void Visible(bool v);
	char *Name();
	bool Name(char *n);

	int OnEvent(GMessage *Msg);
	void Append(GWindow *Wnd);
	bool AppendDialogFromResource(char *Name);
};

// Slider
class GSlider : public GControl, public BScrollBar {

	bool Vertical;
	int Min, Max;
	int Val;

public:
	GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert);
	~GSlider();

	void Value(int i);
	int Value();
	void GetLimits(int &x, int &y);
	void SetLimits(int x, int y);
	int OnEvent(GMessage *Msg);
};

// Combo Box
class GCombo : public GControl, public BMenuField, public List<char> {

	void AttachedToWindow() { OnCreate(); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	bool SortItems;

public:
	GCombo(int id, int x, int y, int cx, int cy, char *name);
	~GCombo();

	bool Sort() { return SortItems; }
	void Sort(bool s) { SortItems = s; }
	void Value(int i);
	int Value();
	char *Name();
	bool Name(char *n);

	int OnEvent(GMessage *Msg);
	void OnCreate();
	bool Delete();
	bool Delete(int i);
	bool Delete(char *p);
	bool Insert(char *p, int Index = -1);
};

// Bitmap
class GBitmap : public GControl, public BControl {

	GdcBasePrimitives *pDC;

	void AttachedToWindow() { OnCreate(); }
	void Draw(BRect UpdateRect) { GdcViewDC DC(Sys_View); OnPaint(&DC); }
	void FrameMoved(BPoint Point) { Pos.Offset(Point.x-Pos.x1, Point.y-Pos.y1); OnPosChange(); }
	void FrameResized(float width, float height) { Pos.Dimension(width+1, height+1); OnPosChange(); }
	void Pulse() { OnPulse(); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	void MakeFocus(bool focused = true) { OnFocus(focused); }
	void KeyDown(const char *bytes, int32 numBytes) { Sys_KeyDown(bytes, numBytes); }
	void KeyUp(const char *bytes, int32 numBytes) { Sys_KeyUp(bytes, numBytes); }
	void MouseDown(BPoint point) { Sys_MouseDown(point); }
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) { Sys_MouseMoved(point, transit, message); }

public:
	GBitmap(int id, int x, int y, char *FileName);
	~GBitmap();

	int OnEvent(GMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);
	void OnMouseClick(VMouse &m);
};

//////////////////////////////////////////////////////////////////////////////
#include "BetterScrollView.h"
#include "ColumnListView.h"
#include "CLVEasyItem.h"

#define GIC_OWNER_DRAW			0x0001
#define GIC_ASK_TEXT			0x0002
#define GIC_ASK_IMAGE			0x0004
#define GIC_OWN_LIST			0x0008

class GItemContainer {
protected:
	int Flags;
	GImageList *ImageList;

public:
	GItemContainer();
	~GItemContainer();

	bool OwnerDraw() { return TestFlag(Flags, GIC_OWNER_DRAW); }
	void OwnerDraw(bool b) { if (b) SetFlag(Flags, GIC_OWNER_DRAW); else ClearFlag(Flags, GIC_OWNER_DRAW); }
	bool AskText() { return TestFlag(Flags, GIC_ASK_TEXT); }
	void AskText(bool b) { if (b) SetFlag(Flags, GIC_ASK_TEXT); else ClearFlag(Flags, GIC_ASK_TEXT); }
	bool AskImage() { return TestFlag(Flags, GIC_ASK_IMAGE); }
	void AskImage(bool b) { if (b) SetFlag(Flags, GIC_ASK_IMAGE); else ClearFlag(Flags, GIC_ASK_IMAGE); }
	bool OwnList() { return TestFlag(Flags, GIC_OWN_LIST); }
	void OwnList(bool b) { if (b) SetFlag(Flags, GIC_OWN_LIST); else ClearFlag(Flags, GIC_OWN_LIST); }
	GImageList *GetList() { return ImageList; }

	bool SetImageList(GImageList *List, bool Own = TRUE);
	bool Load(char *File);
};

class GItem : public GEvent, public CLVEasyItem {
protected:
public:
	GItem(float minheight = 0.0);

	// Events
	void OnMouseClick(VMouse &m) {}
	void OnPaint(GdcBasePrimitives *pDC) {}
	virtual void OnBeginDrag(VMouse &m) {}
	virtual void OnMeasure(void *Info) {}
	virtual void Update() {}

	// State
	virtual bool Selected() { return FALSE; }
	virtual void Selected(bool b) {}

	// Data
	virtual char *GetText(int i=0) { return 0; }
	virtual bool SetText(char *s, int i=0) { return FALSE; }
	virtual int GetImage(int Flags = 0) { return -1; }
	virtual void SetImage(int i) {}
};

// List view
class GListItem : public GItem {

	friend class GList;

protected:
	GList *Parent;
	int Cols;
	char **Str;

	bool GetBitmapInfo(BBitmap **bmp, BRect *rect);

public:
	GListItem();
	virtual ~GListItem();

	int GetCols() { return Cols; }
	char *GetText(int i);
	bool SetText(char *s, int i=0);
	void Update();
	bool Selected();
	void Selected(bool b);
};

class GList : public GControl, public ColumnListView, public GItemContainer {

	friend class GListItem;

protected:
	void AttachedToWindow() { ColumnListView::AttachedToWindow(); OnCreate(); }
	bool Attach(GWindow *parent);

	BetterScrollView *ScrollView;
	bool ColumnHeader;
	bool EditLabels;
	int Val;
	
	void SelectionChanged();
	void MouseDown(BPoint point);

public:
	GList(int id, int x, int y, int cx, int cy, char *name = "List");
	~GList();

	bool SetPos(GdcRegion &p, bool Repaint = FALSE);

	// Events
	virtual void OnItemClick(GListItem *Item, VMouse &m);
	virtual void OnItemBeginDrag(GListItem *Item, int Flags);
	virtual void OnItemSelect(GListItem *Item) {}

	// properties
	void AddColumn(char *Name, int Width = 50, int Align = 0);
	void EmptyColumns();
	bool ShowColumnHeader();
	void ShowColumnHeader(bool Show);
	bool AllowEditLabels() { return EditLabels; }
	void AllowEditLabels(bool b) { EditLabels = b; }

	// methods
	int OnEvent(BMessage *Msg);
	void OnMouseClick(VMouse &m);
	int Value();
	void Value(int i);
	bool Select(GListItem *Obj);
	bool GetSelection(List<GListItem> &l);
	bool GetAll(List<GListItem> &l);

	// list
	GListItem *Current();
	bool IsEmpty() { return CountItems() == 0; }
	bool Delete();
	bool Delete(int Index);
	bool Delete(GListItem *p);
	bool Insert(GListItem *p, int Index = -1);
	bool Remove(GListItem *Obj);
	GListItem *First();
	GListItem *Last();
	GListItem *Next();
	GListItem *Prev();

	void Empty();			// delete all items
	void RemoveAll();		// remove references to externally owned items
};

// Tree view
extern int GTreeCompareFunc(const BListItem *a, const BListItem *b);
class GTreeItem : public GItem {

	friend class GTree;
	friend int GTreeCompareFunc(const BListItem *a, const BListItem *b);

protected:
	char *Str;
	GTree *Tree;
	GTreeItem *Parent;
	int Index;

	void DrawItem(BView *owner, BRect frame, bool complete);

public:
	GTreeItem();
	virtual ~GTreeItem();

	char *GetText(int i=0);
	bool SetText(char *s, int i=0);
	GTreeItem *Insert(GTreeItem *Obj = NULL, int Pos = -1);
	void Update();
	bool Selected();
	void Selected(bool b);
	bool Expanded();
	void Expanded(bool b);

	GTreeItem *GetChild();
	GTreeItem *GetNext();
};

class GTree : public GControl, public BOutlineListView, public GItemContainer {

	friend class GTreeItem;

	// redirector stubs
	void MouseDown(BPoint point);

protected:
	BScrollView *ScrollView;
	bool Attach(GWindow *parent);

	bool Lines;
	bool Buttons;
	bool LinesAtRoot;
	bool EditLabels;

	void SelectionChanged();
	
public:
	GTree(int id, int x, int y, int cx, int cy, char *name);
	~GTree();

	// Move the scrollview and the tree
	bool SetPos(GdcRegion &p, bool Repaint = FALSE);

	// Events
	virtual void OnItemClick(GTreeItem *Item, VMouse &m);
	virtual void OnItemBeginDrag(GTreeItem *Item, int Flags);
	virtual void OnItemExpand(GTreeItem *Item, bool Expand) {}
	virtual void OnItemSelect(GTreeItem *Item) {}
	
	// Procedures
	void OnMouseClick(VMouse &m);
	int OnEvent(GMessage *Msg);
	GTreeItem *Insert(GTreeItem *Obj = 0, int Pos = -1);
	bool Delete(GTreeItem *Obj);
	bool Select(GTreeItem *Obj);
	bool Remove(GTreeItem *Obj);

	// Container
	void Empty();		// delete all items
	void RemoveAll();	// remove reference to externally owned items
};

// Progress control
class GProgress : public GControl, public BStatusBar, public Progress {

	void AttachedToWindow(void);

public:
	GProgress(int id, int x, int y, int cx, int cy, char *name);
	~GProgress();

	void SetLimits(int l, int h);
	void GetLimits(int &l, int &h);
	void Value(int v);
	int Value();
	int OnEvent(GMessage *Msg);
};

#endif