/*hdr
**      FILE:           Gui.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Memory subsystem
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"

#include "gdc2.h"
#include "gui.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
BMessage NewMsg(int m, int a, int b)
{
	BMessage Msg(m);
	Msg.AddInt32("a", a);
	Msg.AddInt32("b", b);
	return Msg;
}

char *NewStr(char *s)
{
	if (s)
	{
		char *c = NEW(char[strlen(s)+1]);
		if (c)
		{
			strcpy(c, s);
			return c;
		}
	}

	return 0;	
}

void DrawBox(GdcBasePrimitives *pDC, GdcRegion &r, bool Sunken, bool Fill)
{
	if (Fill)
	{
		pDC->Colour(CBit(pDC->GetBits(), TOOL_MED));
		pDC->Rectangle(r.x1+1, r.y1+1, r.x2-1, r.y2-1);
	}

	pDC->Colour(CBit(pDC->GetBits(), (Sunken) ? TOOL_HIGH : TOOL_LOW));
	pDC->Line(r.x2, r.y2, r.x2, r.y1);
	pDC->Line(r.x2, r.y2, r.x1, r.y2);

	pDC->Colour(CBit(pDC->GetBits(), (Sunken) ? TOOL_LOW : TOOL_HIGH));
	pDC->Line(r.x1, r.y1, r.x1, r.y2);
	pDC->Line(r.x1, r.y1, r.x2, r.y1);
}

void DrawMoveDeepEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type)
{
	COLOUR VLow = CBit(pDC->GetBits(), TOOL_VLOW);
	COLOUR Low = CBit(pDC->GetBits(), TOOL_LOW);
	COLOUR High = CBit(pDC->GetBits(), TOOL_HIGH);
	COLOUR VHigh = CBit(pDC->GetBits(), TOOL_VHIGH);

	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(Low);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(VLow);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(High);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VHigh);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case RAISED:
		{
			pDC->Colour(VHigh);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(High);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(Low);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VLow);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case CHISEL:
		{
			pDC->Colour(Low);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(High);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(Low);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(High);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
	}

	r.Size(2, 2);
}

void DrawMoveShallowEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type)
{
	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(CBit(pDC->GetBits(), TOOL_HIGH));
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(CBit(pDC->GetBits(), TOOL_LOW));
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
		case RAISED:
		{
			pDC->Colour(CBit(pDC->GetBits(), TOOL_LOW));
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(CBit(pDC->GetBits(), TOOL_HIGH));
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
	}
}

void DrawMoveFill(GdcBasePrimitives *pDC, GdcRegion &r, int Width)
{
	pDC->Colour(CBit(pDC->GetBits(), TOOL_MED));
	if (Width < 1 OR r.X() < (2 * Width) OR r.Y() < (2 * Width))
	{
		pDC->Rectangle(&r);
		r.ZOff(-1, -1);
	}
	else
	{
		pDC->Rectangle(r.x1, r.y1, r.x2, r.y1+Width-1);
		pDC->Rectangle(r.x1, r.y2-Width+1, r.x2, r.y2);
		pDC->Rectangle(r.x1, r.y1+Width, r.x1+Width-1, r.y2-Width);
		pDC->Rectangle(r.x2-Width+1, r.y1+Width, r.x2, r.y2-Width);
		r.Size(Width, Width);
	}
}

////////////////////////////////////////////////////////////////////////////
GMessage CreateMsg(int m, int a, int b)
{
	GMessage Msg(m);
	Msg.AddInt32("a", a);
	Msg.AddInt32("b", b);
	return Msg;
}

////////////////////////////////////////////////////////////////////////////
bool LgiExecute(char *File, char *Args, char *Dir)
{
	char f[256];
	
	if (Dir)
	{
		sprintf(f, "%s/%s", Dir, File);
	}
	else
	{
		strcpy(f, File);
	}

	BRoster Roster;
	BEntry Entry(f);
	entry_ref Ref;
	if (Entry.GetRef(&Ref) == B_OK)
	{
		return Roster.Launch(&Ref) == B_OK;
	}
	return false;
}

int LgiMsg(GWindow *Parent, char *Str, char *Title, int Type, ...)
{
	int Res = 0;
	if (Str)
	{
		char Buffer[1024];
		va_list Arg;

		va_start(Arg, Type);
		vsprintf(Buffer, Str, Arg);
		va_end(Arg);

		BAlert *Dlg = 0;
		
		switch (Type)
		{
			case MB_OK:
			{
				Dlg = NEW(BAlert((Title) ? Title : "Message", Buffer, "Ok"));
				break;
			}
			case MB_YESNO:
			{
				Dlg = NEW(BAlert((Title) ? Title : "Message", Buffer, "Yes", "No", NULL, B_WIDTH_AS_USUAL, B_IDEA_ALERT));
				break;
			}
		}
		
		if (Dlg)
		{
			Res = Dlg->Go();
		}
		
		if (Res >= 0 AND Type == MB_YESNO)
		{
			if (Res == 0)
				return IDYES;
			if (Res == 1)
				return IDNO;
		}
	}
	return Res;
}

bool LgiGetCursorPos(GdcPt2 *Pt)
{
	bool Status = false;
	if (Pt)
	{
		BWindow *Wnd = new BWindow(
			BRect(-20, -20, -10, -10), 
			"temp", 
			B_BORDERED_WINDOW, 
			0);
		if (Wnd)
		{
			BView *View = new BView(BRect(0, 0, 10, 10), "the view", B_FOLLOW_ALL_SIDES, B_WILL_DRAW);
			if (View)
			{
				Wnd->AddChild(View);
				BPoint Cursor;
				uint32 Buttons;
				View->GetMouse(&Cursor, &Buttons);
				Pt->x = Cursor.x;
				Pt->y = Cursor.y;
				Status = TRUE;
			}
	
			Wnd->Quit();
		}
	}

	return Status;
}

#include "resource.h"

char *LgiLoadString(int Id)
{
	class StringObj {
	public:
		int Id;
		char *Str;
	} Objs[] = {
		{IDS_APPNAME,			"i.Scribe"},
		{IDS_ERROR,				"<Error>"},
		{IDS_ERROR2,			"<Error>"},
		{IDS_ERROR3, 			"<Error>"},
		{IDS_SURNAME,			"Surname"},
		{IDS_EMAIL,				"Email"},
		{IDS_OK, 				"Ok"},
		{IDS_CANCEL,			"Cancel"},
		{IDS_SEND,				"Send"},
		{IDS_SAVE,				"Save"},
		{IDS_REPLY,				"Reply"},
		{IDS_REPLYALL,			"Reply All"},
		{IDS_FORWARD, 			"Forward"},
		{IDS_REMOVE,			"Remove"},
		{IDS_NONREFLOWWORDWRAP,	"Non-reflow word wrap"},
		{IDS_TO,				"To"},
		{IDS_CC,				"Cc"},
		{IDS_RECIPIENTS,		"Recipients"},
		{IDS_TEXT,				"Text"},
		{IDS_ATTACHMENTS,		"Attachments"},
		{IDS_INTERNETHEADER,	"Internet Header"},
		{IDS_FOLDER_PROPERTIES_DLG,	"%s Folder\n\n   Mail: %i\n   Contacts: %i\n   Folders: %i\n   Size: %s"},
		{IDS_DELETE_FOLDER_DLG,	"Do you want to delete the '%s' folder?"},
		{IDS_CANT_FIND_FOLDERS_DLG, "Could not find or open the mail folders file.\nWould you like to create one?"},
		{IDS_COMPACT_TESTING_DLG, "Do you want to compact your folders?"},
		{IDS_FOLDER_COMPACT_ERR_DLG, "An error occured compacting the mail folders."},
		{IDS_COULDNT_OPEN,		"Couldn't open '%s'"},
		{IDS_ATTACH_FILE,		"&Attach File"},
		{IDS_EDIT,				"&Edit"},
		{IDS_DELETE,			"&Delete"},
		{IDS_ADD_CONTACTS,		"&Add to contacts"},
		{IDS_ATTACH_WARNING_DLG, "A file by this name already exists, do you still want to attach it?"},
		{IDS_MAIL_PROPS_DLG,	"Mail Message\n\n   Size: %s\n   Date: %s\n   Flags: %X"},
		{IDS_MAIL_PROPS,		"Mail Properties"},
		{IDS_ADD,				"&Add"},
		{IDS_OPEN,				"&Open"},
		{IDS_SET_READ,			"Set Read"},
		{IDS_SET_UNREAD,		"Set Unread"},
		{IDS_PROPERTIES,		"&Properties"},
		{IDS_SUBJECT,			"Subject"},
		{IDS_MAIL_MESSAGE,		"Mail Message"},
		{IDS_FROM,				"From"},
		{IDS_IDENTITY,			"Identity"},
		{IDS_OPTIONS,			"Options"},
		{IDS_SMTP,				"Smtp"},
		{IDS_POP3,				"Pop3"},
		{IDS_SIZE,				"Size"},
		{IDS_DATE,				"Date"},
		{IDS_SAVEAS,			"Save &As"},
		{IDS_EMPTY,				"Empty"},
		{IDS_NEW,				"New"},
		{IDS_MAIL,				"Mail"},
		{IDS_CONTACT,			"Contact"},
		{IDS_DELETEFOLDER,		"Delete Folder"},
		{IDS_CREATESUBFOLDER,	"Create Sub Folder"},
		{IDS_RENAMEFOLDER,		"Rename Folder"},
		{IDS_FIRST,				"First"},
		{IDS_MOVE_ERROR,		"Couldn't move %i of %i items to folder.\nThe most common cause is that the item is a different\ntype to the folder."},
		{0, 0}};

	StringObj *o = Objs;
	while (o->Str)
	{
		if (o->Id == Id)
		{
			return o->Str;
		}
		o++;
	}
	
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
/*
LRESULT CALLBACK MouseHook(int nCode, WPARAM wParam, LPARAM lParam)
{
	// check whether the mouse has moved off the last window
	MOUSEHOOKSTRUCT *Ms = (MOUSEHOOKSTRUCT*) lParam;
	if (Ms)
	{
		RECT rc;
		GetClientRect(Ms->hwnd, &rc);
		ScreenToClient(Ms->hwnd, &Ms->pt);

		// is the mouse inside the client area?
		bool Inside = NOT      (Ms->pt.x < 0 OR
					Ms->pt.y < 0 OR
					Ms->pt.x >= rc.right-rc.left OR
					Ms->pt.y >= rc.bottom-rc.top);

		HWND hWnd = (Inside) ? Ms->hwnd : 0;

		if (hWnd != hMouseOver)
		{
			// window has changed
			if (hMouseOver AND IsWindow(hMouseOver))
			{
				// window is LOSING mouse
				SendMessage(hMouseOver, WM_MOUSEEXIT, 0, MAKELONG((short) Ms->pt.x, (short) Ms->pt.y));
			}

			// set current window
			hMouseOver = hWnd;

			if (hMouseOver AND IsWindow(hMouseOver))
			{
				// window is GETTING mouse
				SendMessage(hMouseOver, WM_MOUSEENTER, Inside, MAKELONG((short) Ms->pt.x, (short) Ms->pt.y));
			}
		}
	}

	return CallNextHookEx(hMouseHook, nCode, wParam, lParam);
}
*/

////////////////////////////////////////////////////////////////////////////////////////////////////
GWindow::GWindow(BView *view)
{
	Sys_View = view;
	// Sys_View->SetViewColor(B_TRANSPARENT_32_BIT);
	WndFlags = GWF_ENABLE | GWF_VISIBLE;
	Pos.ZOff(0, 0);
	Parent = 0;
	Notify = 0;
	Mark = FALSE;
}

GWindow::~GWindow()
{
/*
	if (Parent)
	{
		Parent->Children.Delete(this);
	}

	GWindow *c;
	while (c = Children.First())
	{
		DeleteObj(c);
	}
*/

	WndFlags = 0;
}

bool GWindow::IsOver(VMouse &m)
{
	return	(m.x >= 0) AND
		(m.y >= 0) AND
		(m.x < Pos.X()) AND
		(m.y < Pos.Y());
}

GWindow *GWindow::WindowFromPoint(int x, int y)
{
	for (GWindow *c = Children.First(); c; c = Children.Next())
	{
		if (c->Pos.Overlap(x, y))
		{
			return c;
		}
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		return this;
	}

	return NULL;
}

bool GWindow::Create(char *name, GdcRegion &pos)
{
	GObject::Name(name);
	Pos = pos;
	return TRUE;
}

void GWindow::OnPosChange()
{
}

bool GWindow::SetPos(GdcRegion &p, bool Repaint)
{
	if (Sys_View AND Sys_View->LockLooper())
	{
		Sys_View->MoveTo(p.x1, p.y1);
		Sys_View->ResizeTo(p.X()-1, p.Y()-1);
		Sys_View->UnlockLooper();
	}
	Pos = p;
	OnPosChange();
	
	return TRUE;
}

bool GWindow::Invalidate(GdcRegion *r, bool Repaint)
{
	if (Sys_View->LockLooper())
	{
		if (r)
		{
			BRect Rc = *r;
			Sys_View->Invalidate(Rc);
		}
		else
		{
			Sys_View->Invalidate();
		}
		Sys_View->UnlockLooper();
	}
	
	return FALSE;
}

void GWindow::SetPulse(int Length)
{
	if (Sys_View->LockLooper())
	{
		if (Length > 0)
		{
			BWindow *AppWnd = Sys_View->Window();
			if (AppWnd)
			{
				bigtime_t Rate = (bigtime_t) Length * 1000;
				AppWnd->SetPulseRate(Rate);
			}

			Sys_View->SetFlags(Sys_View->Flags() | B_PULSE_NEEDED);
		}
		else
		{
			Sys_View->SetFlags(Sys_View->Flags() & ~B_PULSE_NEEDED);
		}
		
		Sys_View->UnlockLooper();
	}
}

int GWindow::OnEvent(BMessage *Msg)
{
	/*
	if (hWnd)
	{
		switch (Msg)
		{
			case WM_NCDESTROY:
			{
				hWnd = 0;
				break;
			}
			case WM_DESTROY:
	 		case WM_CLOSE:
			{
				break;
			}
			case WM_CREATE:
			{
				OnCreate();
				break;
			}
			case WM_SETFOCUS:
			{
				OnFocus(TRUE);
				break;
			}
			case WM_KILLFOCUS:
			{
				OnFocus(FALSE);
				break;
			}
			case WM_WANT_DIALOG_PROC:
			{
				return (WndFlags & GWF_DIALOG_PROCESSING) != 0;
			}
			case WM_WINDOWPOSCHANGED:
			{
				WINDOWPOS *Info = (LPWINDOWPOS) b;
				if (Info)
				{
					Pos.ZOff(Info->cx-1, Info->cy-1);
					Pos.Offset(Info->x, Info->y);
				}
				break;
			}
			case WM_CAPTURECHANGED:
			{
				Capturing = 0;
				break;
			}
			case WM_MOUSEENTER:
			{
				if (a AND NOT TestFlag(WndFlags, GWF_OVER))
				{
					VMouse Ms;
					Ms.x = (short) (b&0xFFFF);
					Ms.y = (short) (b>>16);
					Ms.Flags = 0;
					Ms.Down = FALSE;
					Ms.Double = FALSE;

					GWindow *MouseOver = WindowFromPoint(Ms.x, Ms.y);
					if (Capturing)
					{
						if (MouseOver == Capturing)
						{
							Capturing->OnMouseEnter(AdjustMouseInfoForWindow(Ms, Capturing));
						}
					}
					else
					{
						if (Over != MouseOver)
						{
							if (Over)
							{
								Over->OnMouseExit(AdjustMouseInfoForWindow(Ms, Over));
							}

							Over = MouseOver;

							if (Over)
							{
								Over->OnMouseEnter(AdjustMouseInfoForWindow(Ms, Over));
							}
						}
					}

					SetFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case WM_MOUSEEXIT:
			{
				if (NOT a AND TestFlag(WndFlags, GWF_OVER))
				{
					VMouse Ms;
					Ms.x = (short) (b&0xFFFF);
					Ms.y = (short) (b>>16);
					Ms.Flags = 0;
					Ms.Down = FALSE;
					Ms.Double = FALSE;

					if (Capturing)
					{
						Capturing->OnMouseExit(AdjustMouseInfoForWindow(Ms, Capturing));
					}
					else
					{
						if (Over)
						{
							Over->OnMouseExit(AdjustMouseInfoForWindow(Ms, Over));
							Over = 0;
						}
						else
						{
							OnMouseExit(Ms);
						}
					}

					ClearFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case WM_MOUSEMOVE:
			{
				VMouse Ms;
				Ms.x = (short) (b&0xFFFF);
				Ms.y = (short) (b>>16);
				Ms.Flags = a;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down = ((a & (MK_LBUTTON|MK_MBUTTON|MK_RBUTTON)) != 0);
				Ms.Double = FALSE;

				GWindow *MouseOver = WindowFromPoint(Ms.x, Ms.y);
				if (Over != MouseOver)
				{
					if (Over)
					{
						Over->OnMouseExit(AdjustMouseInfoForWindow(Ms, Over));
					}

					Over = MouseOver;

					if (Over)
					{
						Over->OnMouseEnter(AdjustMouseInfoForWindow(Ms, Over));
					}
				}

				if (Capturing)
				{
					Capturing->OnMouseMove(AdjustMouseInfoForWindow(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseMove(AdjustMouseInfoForWindow(Ms, Over));
				}
				break;
			}
			case WM_LBUTTONDBLCLK:
			case WM_LBUTTONDOWN:
			case WM_LBUTTONUP:
			{
				VMouse Ms;
				Ms.x = (short) (b&0xFFFF);
				Ms.y = (short) (b>>16);
				Ms.Flags = a | MK_LEFT;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down = Msg != WM_LBUTTONUP;
				Ms.Double = Msg == WM_LBUTTONDBLCLK;

				if (Capturing)
				{
					Capturing->OnMouseClick(AdjustMouseInfoForWindow(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(AdjustMouseInfoForWindow(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_RBUTTONDBLCLK:
			case WM_RBUTTONDOWN:
			case WM_RBUTTONUP:
			{
				VMouse Ms;
				Ms.x = (short) (b&0xFFFF);
				Ms.y = (short) (b>>16);
				Ms.Flags = a | MK_RIGHT;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down = Msg != WM_RBUTTONUP;
				Ms.Double = Msg == WM_RBUTTONDBLCLK;

				if (Capturing)
				{
					Capturing->OnMouseClick(AdjustMouseInfoForWindow(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(AdjustMouseInfoForWindow(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_MBUTTONDBLCLK:
			case WM_MBUTTONDOWN:
			case WM_MBUTTONUP:
			{
				VMouse Ms;
				Ms.x = (short) (b&0xFFFF);
				Ms.y = (short) (b>>16);
				Ms.Flags = a | MK_MIDDLE;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
 				Ms.Down = Msg != WM_MBUTTONUP;
				Ms.Double = Msg == WM_MBUTTONDBLCLK;

				if (Capturing)
				{
					Capturing->OnMouseClick(AdjustMouseInfoForWindow(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(AdjustMouseInfoForWindow(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_KEYDOWN:
			case WM_KEYUP:
			{
				VKey Key;
				Key.c = a;		// virtual-key code 
				Key.Flags = 0;
				SetKeyFlag(Key.Flags, VK_MENU, MK_ALT);
				SetKeyFlag(Key.Flags, VK_SHIFT, MK_SHIFT);
				SetKeyFlag(Key.Flags, VK_CONTROL, MK_CTRL);
				Key.Data = b;
				Key.Down = Msg == WM_KEYDOWN;
				Key.IsChar = FALSE;

				OnKey(Key);
				break;
			}
			case WM_CHAR:
			{
				VKey Key;
				Key.c = a;		// virtual-key code 
				Key.Flags = 0;
				SetKeyFlag(Key.Flags, VK_MENU, MK_ALT);
				SetKeyFlag(Key.Flags, VK_SHIFT, MK_SHIFT);
				SetKeyFlag(Key.Flags, VK_CONTROL, MK_CTRL);
				Key.Data = b;
				Key.Down = TRUE;
				Key.IsChar = TRUE;

				OnKey(Key);
				break;
			}
			case WM_PAINT:
			{
				GdcGdiScreen DeviceContext(hWnd);
				OnPaint(&DeviceContext);
				break;
			}
			default:
			{
				return DefWindowProc(hWnd, Msg, a, b);
			}
		}
	}
*/

	return 0;
}

bool GWindow::Name(char *n)
{
	bool Status = GObject::Name(n);
	Sys_View->SetName(n);
	return TRUE;
}

char *GWindow::Name()
{
	return GObject::Name();
}

bool GWindow::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLE);
}

void GWindow::Enabled(bool e)
{
	if (e) SetFlag(WndFlags, GWF_ENABLE);
	else ClearFlag(WndFlags, GWF_ENABLE);
}

bool GWindow::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GWindow::Visible(bool v)
{
	if (v) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);

	if (v)
	{
		Sys_View->Show();
	}
	else
	{
		Sys_View->Hide();
	}
}

bool GWindow::Focus()
{
	bool f = false;
	BWindow *Wnd = GetBWindow();

	printf("Wnd = %p, Sys_View = %p\n", Wnd, Sys_View);
	if (Wnd AND Sys_View AND Sys_View->LockLooper())
	{
		BView *Cur = Wnd->CurrentFocus();
		f = (Sys_View == Cur);
		
		printf("GetFocus %p - %p\n", Sys_View, Cur);
		
		if (f) SetFlag(WndFlags, GWF_FOCUS);
		else ClearFlag(WndFlags, GWF_FOCUS);

		Sys_View->UnlockLooper();
	}

	return TestFlag(WndFlags, GWF_FOCUS);
}

void GWindow::Focus(bool f)
{
	if (f) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);

	if (Sys_View AND Sys_View->LockLooper())
	{
		Sys_View->MakeFocus(f);
		Sys_View->UnlockLooper();
	}
}

bool GWindow::Capture(bool c)
{
	return FALSE;
}

void GWindow::Sys_KeyDown(const char *bytes, int32 numBytes)
{
	// printf("GWindow::Sys_KeyDown(%p, %i)\n", bytes, numBytes);

	key_info Info;
	get_key_info(&Info);

	for (int i=0; i<numBytes; i++)
	{
		VKey k;
		
		k.c = bytes[i];
		k.Flags = 0;
		k.Data = 0;
		k.Down = TRUE;
		k.IsChar = (k.c < 128) AND (k.c >= ' ');
		
		OnKey(k);
	}
}

void GWindow::Sys_KeyUp(const char *bytes, int32 numBytes)
{
	key_info Info;
	get_key_info(&Info);

	for (int i=0; i<numBytes; i++)
	{
		VKey k;
		
		k.c = bytes[i];
		k.Flags = 0;
		k.Data = 0;
		k.Down = FALSE;
		k.IsChar = 0;
		
		OnKey(k);
	}
}

long MouseThread(void *Ptr)
{
	GWindow *Wnd = (GWindow*) Ptr;
	if (Wnd)
	{
		VMouse m;
		BPoint Point;
		uint32 Buttons;
	
		BWindow *Window = Wnd->Sys_View->Window();
		if (Window->Lock())
		{
			int32 Clicks = 1;
			
			Wnd->Sys_View->GetMouse(&Point, &Buttons, FALSE);
			BWindow *BWnd = Wnd->Sys_View->Window();
			if (BWnd)
			{
				BMessage *Msg = BWnd->CurrentMessage();
				if (Msg)
				{
					Msg->FindInt32("clicks", &Clicks);
				}
			}
	
			m.x = Point.x;
			m.y = Point.y;
			m.Flags = Buttons;
			m.Down = true;
			m.Double = Clicks > 1;
			
			Wnd->OnMouseClick(m);
			Window->Unlock();
		}
	
		while (Buttons)
		{
			if (Window->Lock())
			{
				Wnd->Sys_View->GetMouse(&Point, &Buttons, FALSE);
				if (m.x != Point.x OR m.y != Point.y)
				{	
					m.x = Point.x;
					m.y = Point.y;
					m.Flags = Buttons;
					m.Double = FALSE;
					
					Wnd->OnMouseMove(m);
				}
				
				Window->Unlock();
			}
			
			snooze(20 * 1000);
		}
		
		if (Window->Lock())
		{
			Wnd->Sys_View->GetMouse(&Point, &Buttons, FALSE);
			m.x = Point.x;
			m.y = Point.y;
			m.Flags = Buttons;
			m.Down = FALSE;
			m.Double = FALSE;

			Wnd->OnMouseClick(m);
			Window->Unlock();
		}
	}
	
	return 0;
}

void GWindow::Sys_MouseDown(BPoint Point)
{
	thread_id Thread = spawn_thread(MouseThread, "LGI_MouseWatcher", B_NORMAL_PRIORITY, this);
	if (Thread > 0)
	{
		resume_thread(Thread);
	}
}

void GWindow::Sys_MouseMoved(BPoint Point, uint32 transit, const BMessage *message)
{
	VMouse m;
	ulong Buttons;

	Sys_View->GetMouse(&Point, &Buttons, FALSE);
	
	m.x = Point.x;
	m.y = Point.y;
	m.Flags = Buttons;
	
	OnMouseMove(m);
}

bool GWindow::Attach(GWindow *Wnd)
{
	bool Status = FALSE;
	Parent = Wnd;
	if (Parent)
	{
		Parent->Sys_View->AddChild(Sys_View);
		Parent->OnChildrenChanged(this, TRUE);

		Status = TRUE;
	}
	return Status;
}

bool GWindow::Detach()
{
	bool Status = FALSE;
	if (Parent)
	{
		Visible(FALSE);
		Parent->Sys_View->RemoveChild(Sys_View);
		Parent->OnChildrenChanged(this, FALSE);

		Parent = 0;
		Status = TRUE;
	}
	return Status;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GWnd::GWnd(GWindow *notify) : BWindow(BRect(10, 10, 100, 100), "TestWnd", B_TITLED_WINDOW, 0, B_CURRENT_WORKSPACE)
{
	Notify = notify;
}

GWnd::GWnd(	GWindow *notify,
			BRect frame,
			char *title,
			window_type type,
			uint32 flags,
			uint32 workspaces)
	: BWindow(frame, title, type, flags, workspaces)		
{
	Notify = notify;
}

bool GWnd::QuitRequested()
{
	if (Notify)
	{
		return Notify->QuitRequested();
	}
	
	return TRUE;
}

GAppWindow::GAppWindow()
	: BView(BRect(0, 0, 1, 1), NULL, B_FOLLOW_NONE, B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE),
	GWindow(this)
{
	Menu = 0;

	int F = ResizingMode();
	F &= ~B_FOLLOW_NONE;
	F |= B_FOLLOW_ALL_SIDES;
	SetResizingMode(F);

	rgb_color Rgb;
	Rgb.red = 216;
	Rgb.green = 216;
	Rgb.blue = 216;
	SetViewColor(Rgb); // B_TRANSPARENT_32_BIT);	

	Wnd = NEW(GWnd(this));
	if (Wnd)
	{
		BRect r = Wnd->Bounds();
		ResizeTo(r.Width(), r.Height());
		Wnd->AddChild(this);
	}
}

GAppWindow::~GAppWindow()
{
	DeleteObj(Menu);

	for (GView *v = Views.First(); v; v = Views.Next())
	{
		DeleteObj(v);
	}
}

bool GAppWindow::QuitRequested()
{
	// be_app->PostMessage(B_QUIT_REQUESTED);
	return OnRequestClose();
}

void GAppWindow::OnPaint(GdcBasePrimitives *pDC)
{
/*
	pDC->Colour(Rgb24(255, 0, 0), 24);
	pDC->Box(10, 10, pDC->X() - 10, pDC->Y() - 10);
*/
}

void GAppWindow::OnChildrenChanged(GWindow *Wnd, bool Attaching)
{
}

void GAppWindow::OnPosChange()
{
	GWindow::OnPosChange();
	Pour();
}

void GAppWindow::OnCreate()
{
	if (Wnd)
	{
		BRect f = Wnd->Frame();
		ResizeTo(f.Width(), f.Height());
	}
}

bool GAppWindow::Visible()
{
	bool Status = FALSE;
	if (Wnd)
	{
		Status = NOT Wnd->IsHidden();
	}
	return Status;
}

void GAppWindow::Visible(bool v)
{
	if (Wnd)
	{
		if (v) Wnd->Show();
		else Wnd->Hide();
		Wnd->Sync();
	}

	if (v)
	{
		if (Wnd->Lock())
		{
			Pour();
			Wnd->Unlock();
		}
	}
}

bool GAppWindow::Name(char *n)
{
	bool Status = GObject::Name();
	if (Wnd)
	{
		Wnd->SetTitle(n);
	}
	return Status;
}

char *GAppWindow::Name()
{
	return GObject::Name();
}

GdcRegion &GAppWindow::GetPos()
{
	GdcRegion r(0, 0, 1, 1);
	if (Wnd)
	{
		BRect frame = Wnd->Frame();
		r.x1 = frame.left;
		r.y1 = frame.top;
		r.x2 = frame.right;
		r.y2 = frame.bottom;
	}
	return r;
}

bool GAppWindow::SetPos(GdcRegion &p, bool Repaint)
{
	if (Wnd)
	{
		Wnd->MoveTo(p.x1, p.y1);
		Wnd->ResizeTo(p.X(), p.Y());
	}
	return TRUE;
}

void GAppWindow::Pour()
{
	GdcRegion r;
	r = GetPos();
	r.Offset(-r.x1, -r.y1);
	GdcComplexRegion Client(r);
	GdcComplexRegion Update;

	if (Menu)
	{
		BRect Rc = Menu->Frame();
		GdcRegion Mp = &Rc;
		Mp.x2 = 10000;
		
		Client.Subtract(&Mp);
	}

	for (GView *v = Views.First(); v; v = Views.Next())
	{
		GdcRegion OldPos = v->Pos;
		Update.Union(&OldPos);

		if (v->Pour(Client))
		{
			if (NOT v->Visible())
			{
				v->Visible(TRUE);
			}

			if (OldPos != v->Pos)
			{
				// position has changed update...
				v->GWindow::Invalidate();
			}

			Client.Subtract(&v->Pos);
			Update.Subtract(&v->Pos);
		}
		else
		{
			// make the view not visible
			v->Visible(FALSE);
		}
	}

	for (int i=0; i<Update.GetSize(); i++)
	{
		GWindow::Invalidate(Update[i]);
	}
}

int GAppWindow::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
		case WM_MENU:
		{
			int32 Cmd = 0;
			int32 Event = 0;
			
			Msg->FindInt32("Cmd", &Cmd);
			Msg->FindInt32("Event", &Event);
			OnCommand(Cmd, Event, 0);
			break;
		}
	}

	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
char *GApp::CmdLine = 0;
GApp *GApp::TheApp = 0;

GApp::GApp(char *Mime, char *lpCmdLine) : BApplication(Mime)
{
	TheApp = this;
	CmdLine = lpCmdLine;
	AppWnd = 0;
	SystemFont = NEW(GdcBeTtf);
	if (SystemFont)
	{
		SystemFont->CreateFont(	10,
								0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
								"Courier10 BT");
	}
}

GApp::~GApp()
{
	// DeleteObj(AppWnd);	
	// DeleteObj(SystemFont);
	TheApp = 0;
}

void GApp::OnCmdLine(char *Cmd)
{
	if (Cmd AND strlen(Cmd) > 0)
	{
		int Args = 0;
		for (char *c = Cmd; c; )
		{
			Args++;
			c = strchr(c, ' ');
			if (c) c++;
		}

		char **Arg = NEW(char*[Args]);
		if (Arg)
		{
			char *c = Cmd;
			for (int i=0; i<Args; i++)
			{
				Arg[i] = c;
				c = strchr(c, ' ');
				if (c)
				{
					*c = 0;
					c++;
				}
			}

			OnArgs(Args, Arg);
			DeleteArray(Arg);
		}
	}
}

void GApp::OnArgs(int Args, char **Arg)
{
	if (Arg AND AppWnd)
	{
		List<char> Files;
		char File[256] = "";

		for (int i=0; i<Args; i++)
		{
			if (strlen(File) == 0)
			{
				strcpy(File, Arg[i]);
				if (FileExists(File))
				{
					Files.Insert(NewStr(File));
					File[0] = 0;
				}
			}
			else
			{
				strcat(File, " ");
				strcat(File, Arg[i]);
				if (FileExists(File))
				{
					Files.Insert(NewStr(File));
					File[0] = 0;
				}
			}
		}

		AppWnd->OnReceiveFiles(Files);

		for (char *s = Files.First(); s; s = Files.Next())
		{
			DeleteArray(s);
		}
	}
}

void GApp::Exit(int Code)
{
	// DeleteObj(AppWnd);
	be_app->PostMessage(B_QUIT_REQUESTED);
}

////////////////////////////////////////////////////////////////////////////
GThread::GThread(bool Sleep = FALSE, int StackSize = 8 << 10)
{
	Exited = FALSE;
	ExitCodeVal = -1;
	ThreadId = spawn_thread(EntryPoint, 
							"worker", 
							B_NORMAL_PRIORITY, 
							this);
	if (ThreadId == B_NO_ERROR)
	{
		if (NOT Sleep)
		{
			resume_thread(ThreadId);
		}
	}
	else
	{
		// error
	}
}

GThread::~GThread()
{
	if (NOT IsExited())
	{
		Terminate();
	}
}

long GThread::EntryPoint(void *Ptr)
{
	if (Ptr)
	{
		GThread *Thread = (GThread*) Ptr;
		
		Thread->ExitCodeVal = Thread->Main();
		Thread->Exited = true;
		return Thread->ExitCodeVal;
	}
	return 0;
}

int GThread::ExitCode()
{
	return ExitCodeVal;
}

bool GThread::IsExited()
{
	return Exited;
}

void GThread::Run(bool b)
{
	if (b)
	{
		resume_thread(ThreadId);
	}
	else
	{
		suspend_thread(ThreadId);
	}
}

void GThread::Terminate()
{
	kill_thread(ThreadId);
}

int GThread::Main()
{
	return 0;
}

////////////////////////////////////////////////////////////////////////////
int GTrayIcon::Id = 1;

GTrayIcon::GTrayIcon(int icons)
{
	Val = 0;
	Msg = 0;
	Notify = 0;
	MyId = 0;
	Icons = max(icons, 1);
}

GTrayIcon::~GTrayIcon()
{
	Visible(FALSE);
}

bool GTrayIcon::LoadIcon(char *Str, int Index)
{
	return FALSE;
}

bool GTrayIcon::Visible()
{
	return 0;
}

void GTrayIcon::Visible(bool v)
{
	if (Visible() != v)
	{
		if (v)
		{
		}
		else
		{
		}
	}
}

int GTrayIcon::Value()
{
	return Val;
}

void GTrayIcon::Value(int v)
{
	Val = v;
}

//////////////////////////////////////////////////////////////////////////////
