/*hdr
**	FILE:		Text.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		2/7/97
**	DESCRIPTION:	Generic text document handler
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

/****************************** Includes ************************************************************************************/

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "Text.h"

/****************************** Defines *************************************************************************************/

// Cursor type
// if the cursor goes between charaters (like in graphics mode)
#define CURSOR_BETWEEN
// else the cursor is at or under a charater (like in text mode)
//#define CURSOR_UNDER

/****************************** Locals **************************************************************************************/

// char Alpha[] = "_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTTUVWXYZ";
//  char Digit[] = "0123456789";
char WhiteSpace[] = " \t\r\n";
char Delimiters[] = "!@#$%^&*()'\":;,.<>/?[]{}-=+\\|`~";

bool bAutoIndent = TRUE;

/****************************** Helper Functions ****************************************************************************/

int StrLen(char *c)
{
	int i = 0;
	if (c) for (i=0; *c AND *c != '\n'; i++, c++);
	return i;
}

int RevStrLen(char *c)
{
	int i = 0;
	if (c) for (i=0; *c AND *c != '\n'; i++, c--);
	return i;	
}

void CountLinesAndLen(char *c, int Len, int &EndLen, int &Lines)
{
	EndLen = 0;
	Lines = 0;
	if (c)
	{
		for (; Len>0; Len--, c++)
		{
			if (*c == '\n')
			{
				EndLen = 0;
				Lines++;
			}
			else
			{
				EndLen++;
			}
		}
	}
}

/****************************** Document Class ******************************************************************************/

Document::Document()
{
	File = 0;
}

Document::~Document()
{
	DeleteArray(File);
}

bool Document::Open(char *FileName, int Attrib)
{
	bool Status = FALSE;
	if (FileName)
	{
		char *FName = NewStr(FileName);
		DeleteObj(File);
		File = FName;
		Status = F.Open(File, Attrib);
	}
	return Status;
}

/****************************** Cursor Class ********************************************************************************/

Cursor::Cursor()
{
	x = 0;
	y = 0;
	Parent = 0;
	Offset = 0;
}

Cursor::~Cursor()
{
}

Cursor::operator char*()
{
	return Parent->Data + (Offset + x);
}

char *Cursor::GetStr()
{
	if (Parent)
	{
		return Parent->Data + Offset;
	}
	return NULL;
}

void Cursor::GoUpLine()
{
	if (y > 0)
	{
		char *c = Parent->Data + Offset;

		if (y > 1)
		{
			y--;
			c -= 2;
			x = Length = RevStrLen(c);
			Offset = ((long) c - (long) Parent->Data) - x + 1;
		}
		else
		{
			Offset = 0;
			x = Length = StrLen(Parent->Data);
			y = 0;
		}
	}
	else
	{
		x = 0;
	}
}

void Cursor::GoDownLine()
{
	if (y < Parent->Lines-1)
	{
		char *c = Parent->Data + (Offset + Length);

		y++;
		x = 0;
		c++;
		Length = StrLen(c);
		Offset = (long) c - (long) Parent->Data;
	}
	else
	{
		x = Length;
	}
}

bool Cursor::AtEndOfLine()
{
	return x == Length;
}

bool Cursor::AtBeginningOfLine()
{
	return x == 0;
}

void Cursor::SetX(int X)
{
	x = limit(X, 0, Length);
}

void Cursor::SetY(int Y)
{
	Y = limit(Y, 0, Parent->Lines-1);
	char *c = Parent->FindLine(Y);
	if (c)
	{
		Offset = (long) c - (long) Parent->Data;
		Length = StrLen(c);
		y = Y;
	}
}

void Cursor::MoveX(int Dx)
{
	char *c = Parent->Data + Offset;
	while (Dx)
	{
		if (Dx < 0)
		{
			if (x)
			{
				x--;
			}
			else
			{
				GoUpLine();
			}
		}
		else
		{
			if (c[x])
			{
				if (c[x] != '\n')
				{
					x++;
				}
				else
				{
					GoDownLine();
				}
			}
			else
			{
				return;
			}
		}
		
		Dx += (Dx > 0) ? -1 : 1;
	}
}

void Cursor::MoveY(int Dy)
{
	while (Dy)
	{
		if (Dy < 0)
		{
			GoUpLine();
			Dy++;
		}
		else
		{
			GoDownLine();
			Dy--;
		}
	}
}

int Cursor::CursorX(int TabSize)
{
	char *c = Parent->Data + Offset;
	int cx = 0;

	for (int i=0; i<x; i++, c++)
	{
		if (*c == 9)
		{
			cx += TabSize - (cx % TabSize);
		}
		else
		{
			cx++;
		}
	}

	return cx;
}

/****************************** TextLock Class ******************************************************************************/

TextLock::TextLock()
{
	StartLine = 0;
	Lines = 0;
	Line = 0;
}

TextLock::~TextLock()
{
	DeleteArray(Line);
}

char *TextLock::operator [](int i)
{
	if (i >= StartLine AND i < StartLine + Lines)
	{
		return Line[i - StartLine];
	}
	return 0;
}

/****************************** TextDocument Class **************************************************************************/

TextDocument::TextDocument()
{
	Flags = 0;
	LockCount = 0;
	Dirty = FALSE;
	Lines = 1;
	Length = 0;
	Alloc = 1;
	Data = NEW(char[1]);
	if (Data) *Data = 0;
}

TextDocument::~TextDocument()
{
	DeleteArray(Data);
}

int TextDocument::CountLines(char *c, int Len)
{
	int L = 0;
	if (Length > 0)
	{
		if (Len >= 0)
		{
			while (*c AND Len)
			{
				L += (*c++ == '\n');
				Len--;
			}
			L += (Len > 0 AND *c == 0) ? 1 : 0;
		}
		else
		{
			while (*c)
			{
				L += (*c++ == '\n');
			}
			L++;
		}
	}
	return L;
}

char *TextDocument::FindLine(int i)
{
	char *c = 0;
	if (Length > 0)
	{
		c = Data;
		while (i)
		{
			for (; *c AND *c != '\n'; c++);
			if (*c) c++;
			else return 0;
			i--;
		}
	}
	return c;
}

bool TextDocument::SetLength(int Len)
{
	bool Status = TRUE;

	if (Len + 1 > Alloc)
	{
		int NewAlloc = (Len + TEXT_BLOCK) & (~TEXT_MASK);
		char *Temp = NEW(char[NewAlloc]);
		if (Temp)
		{
			memset(Temp + Length, 0, NewAlloc - Length);
			memcpy(Temp, Data, Length);
			DeleteArray(Data);
			Data = Temp;
			Length = Len;
			Alloc = NewAlloc;
		}
		else
		{
			Status = FALSE;
		}
	}
	else
	{
		// TODO: resize Data to be smaller if theres a
		// considerable size change
		Length = Len;
	}

	return Status;
}

bool TextDocument::Load(char *File)
{
	bool Status = FALSE;

	if (Open(File, O_READ))
	{
		int FileSize = F.GetSize();
		if (SetLength(FileSize))
		{
			F.Read(Data, FileSize);
			
			CrLf = FALSE;

			char *In = Data, *Out = Data;
			int Size = 0;
			for (int i=0; i<FileSize; i++, In++)
			{
				if (*In != '\r')
				{
					*Out++ = *In;
					Size++;
				}
				else
				{
					CrLf = TRUE;
				}
			}

			Status = SetLength(Size);
			Data[Length] = 0;
			Lines = max(CountLines(Data), 1);
		}

		F.Close();
		Dirty = FALSE;
	}

	return Status;
}

bool TextDocument::Save(char *FileName)
{
	bool Status = FALSE;
	
	if (FileName)
	{
		if (Open(FileName, O_WRITE))
		{
			if (CrLf)
			{
				char pCrLf[2] = { 0x0D, 0x0A };

				for (int i=0; i<Length;)
				{
					int Start = i;
					for (char *Wr=Data+i; *Wr AND *Wr != '\n' AND i<Length; i++, Wr++);
					F.Write(Data+Start, i-Start);
					if (*Wr == '\n')
					{
						F.Write(pCrLf, 2);
						i++;
					}
				}

				Status = NOT F.GetStatus();
			}
			else
			{
				// no '\r's so just write the whole thing
				Status = F.Write(Data, Length);
			}

			F.SetSize(F.GetPosition());
			F.Close();
			Dirty = FALSE;
		}
	}

	return Status;
}

bool TextDocument::Import(char *s, int Length)
{
	bool Status = FALSE;
	if (s)
	{
		if (Length < 0)
		{
			Length = strlen(s);
		}

		if (SetLength(Length))
		{
			CrLf = FALSE;

			char *In = s, *Out = Data;
			int Size = 0;
			for (int i=0; i<Length; i++, In++)
			{
				if (*In != '\r')
				{
					*Out++ = *In;
					Size++;
				}
				else
				{
					CrLf = TRUE;
				}
			}

			Status = SetLength(Size + 1);
			Data[Size] = 0;
			Lines = max(CountLines(Data), 1);
		}

		Dirty = FALSE;
	}
	return Status;	
}

bool TextDocument::Export(char *&s, int &size)
{
	return FALSE;
}

bool TextDocument::Lock(TextLock *Lock, int StartLine, int Ls)
{
	bool Status = FALSE;
	if (Lock)
	{
		Lock->StartLine = limit(StartLine, 0, Lines);
		Lock->Lines = min(Ls, Lines - StartLine);
		Lock->Line = NEW(char*[Lock->Lines]);

		char *c = FindLine(StartLine);
		if (Lock->Line)
		{
			for (int i=0; i<Lock->Lines; i++)
			{
				if (c)
				{
					Lock->Line[i] = (*c) ? c : 0;
					for (; *c AND *c != '\n'; c++);
					if (*c) c++;
				}
				else
				{
					Lock->Line[i] = 0;
				}
			}

			LockCount++;
			Status = TRUE;
		}
	}
	return Status;
}

void TextDocument::UnLock(TextLock *Lock)
{
	LockCount = max(LockCount-1, 0);
}

bool TextDocument::MoveLock(TextLock *Lock, int Dy)
{
	bool Status = FALSE;
	if (Lock)
	{
	}
	return Status;
}

bool TextDocument::CursorCreate(Cursor *c, int X, int Y)
{
	bool Status = FALSE;
	if (c)
	{
		char *StartOfLine = FindLine(Y);

		c->Parent = this;
		c->y = Y;
		c->Length = StrLen(StartOfLine);
		c->Offset = (StartOfLine) ? ((long) StartOfLine - (long) Data) : 0;
		c->x = min(X, c->Length);

		Status = TRUE;
	}
	return Status;
}

bool TextDocument::Insert(Cursor *At, char *Text, int Len)
{
	bool Status = FALSE;

	if (Text)
	{
		if (Len < 0)
		{
			Len = strlen(Text);
		}

		int OldLength = Length;
		if (SetLength(Length + Len))
		{
			int Offset = (At) ? At->Offset + At->x : 0;

			memmove(Data + Offset + Len, Data + Offset, OldLength - Offset + 1);
			memcpy(Data + Offset, Text, Len);
			if (At)
			{
				At->Length = StrLen(Data + At->Offset);
			}
			Lines += CountLines(Text, Len);

			Status = TRUE;
			Dirty = TRUE;
		}
	}

	return Status;
}

bool TextDocument::Delete(Cursor *From, int Len, char *Buffer)
{
	bool Status = FALSE;

	if (From)
	{
		int Offset = From->Offset + From->x;

		Len = min(Length - Offset, Len);
		if (Len > 0)
		{
			char *c = Data + Offset;
			if (Buffer)
			{
				memcpy(Buffer, c, Len);
			}
			
			Lines -= CountLines(c, Len);
			memmove(c, c + Len, Length - Offset - Len + 1);
	
			Status = SetLength(Length - Len);
			From->Length = StrLen(Data + From->Offset);
			Dirty = TRUE;
		}
	}

	return Status;
}

/****************************** TextView Class ******************************************************************************/

TextView::TextView()
{
	Flags = 0;
	HiddenLines = 0;
	DisplayLines = 1;
	ClipData = NULL;
	StatusMsg = NULL;
	WrapType = TEXTED_WRAP_NONE;
	Doc.CursorCreate(&User, 0, 0);
}

TextView::~TextView()
{
	DeleteArray(ClipData);
	DeleteArray(StatusMsg);
}

void TextView::SetStatus(char *Msg)
{
	if (Msg)
	{
		DeleteArray(StatusMsg);
		StatusMsg = NEW(char[strlen(Msg)+1]);
		if (StatusMsg)
		{
			strcpy(StatusMsg, Msg);
		}
	}
}

bool TextView::ClipText(char *Str, int Len)
{
	bool Status = FALSE;
	if (Str)
	{
		if (Len < 0)
		{
			Len = strlen(Str);
		}

		DeleteArray(ClipData);
		ClipData = NEW(char[Len+1]);
		if (ClipData)
		{
			memcpy(ClipData, Str, Len);
			ClipData[Len] = 0;
			Status = TRUE;
		}
	}

	return Status;
}

char *TextView::ClipText()
{
	return ClipData;
}

int TextView::ProcessKey(VKey &K)
{
	Flags &= ~(TVF_GOTO_START | TVF_GOTO_END | TVF_EAT_MOVE);

	if (NOT K.IsChar)
	{
		if (K.c == VK_SHIFT)
		{
			if (K.Down)
			{
				Flags |= TVF_SHIFT;
			}
			else
			{
				Flags &= ~(TVF_SHIFT);
			}
		}

		if (K.Down)
		{
			switch (K.c)
			{
				case VK_ESCAPE:
				{
					OnEscape(K);
					break;
				}
				case VK_RETURN:
				{
					OnEnter(K);
					break;
				}
				case VK_BACK:
				{
					if (Flags & TVF_SELECTION)
					{
						OnDeleteSelection(FALSE);
						UpdateHiddenCheck();
					}
					else if (User.Y() > 0 OR User.X() > 0)
					{
						Flags &= ~(TVF_SHIFT);
						OnMoveCursor(-1, 0);
						OnDeleteText(&User, 1, FALSE);
					}
					break;
				}
				case VK_RIGHT:
				case VK_LEFT:
				{
					bool bLeft = K.c == VK_LEFT;
					int Inc = (bLeft) ? -1 : 1;

					if (bLeft)
					{
						Flags |= TVF_GOTO_START;
					}
					else
					{
						Flags |= TVF_GOTO_END;
					}

					Flags |= TVF_EAT_MOVE;

					if (K.Flags & MK_CTRL)
					{
						char *c = User;

						if (bLeft)
						{
							OnMoveCursor(Inc, 0);
							c = User;
							while (	strchr(WhiteSpace, *c) AND
								OnMoveCursor(Inc, 0))
							{
								c = User;
							}
						}

						if (strchr(Delimiters, *c))
						{
							while (	strchr(Delimiters, *c) AND
								OnMoveCursor(Inc, 0))
							{
								c = User;
							}
						}
						else
						{
							while (	IsText(*c) AND
								OnMoveCursor(Inc, 0))
							{
								c = User;
							}
						}

						if (bLeft)
						{
							if (User.Y() > 0 OR User.X() > 0)
							{
								OnMoveCursor(-Inc, 0);
							}
						}
						else
						{
							while (	strchr(WhiteSpace, *c) AND
								OnMoveCursor(Inc, 0))
							{
								c = User;
							}
						}
					}
					else
					{
						OnMoveCursor(Inc, 0);
					}
					break;
				}
				case VK_UP:
				{
					Flags |= TVF_GOTO_START;
					OnMoveCursor(0, -1);
					break;
				}
				case VK_DOWN:
				{
					Flags |= TVF_GOTO_END;
					OnMoveCursor(0, 1);
					break;
				}
				case VK_PRIOR:
				{
					int Move = 1-DisplayLines;
					OnSetHidden(HiddenLines+Move);
					OnMoveCursor(0, Move);
					break;
				}
				case VK_NEXT:
				{
					int Move = DisplayLines-1;
					OnSetHidden(HiddenLines+Move);
					OnMoveCursor(0, Move);
					break;
				}
				case VK_HOME:
				{
					if (K.Flags & MK_CTRL)
					{
						OnSetCursor(0, 0);
					}
					else
					{
						OnMoveCursor(-User.X(), 0);
					}
					break;
				}
				case VK_END:
				{
					if (K.Flags & MK_CTRL)
					{
						OnSetCursor(1024, Doc.GetLines());
					}
					else
					{
						OnMoveCursor(User.LineLength()-User.X(), 0);
					}
					break;
				}
				case VK_DELETE:
				{
					if (	(K.Flags & MK_SHIFT) AND
						(Flags & TVF_SELECTION))
					{
						Cut();
					}
					else
					{
						if (Flags & TVF_SELECTION)
						{
							OnDeleteSelection(FALSE);
							UpdateHiddenCheck();
						}
						else if (	User.Y() < Doc.GetLines() OR
								User.X() < User.LineLength())
						{
							OnDeleteText(&User, 1, FALSE);
						}
					}
					break;
				}
				case VK_INSERT:
				{
					if (Flags & TVF_SELECTION)
					{
						if (K.Flags & MK_CTRL)
						{
							Copy();
						}
					}

					if (K.Flags & MK_SHIFT)
					{
						Paste();
					}
					break;
				}
			}
		}
	}
	else
	{
		// IsChar
		#define EXTENDED (1<<24)

		// if (	NOT (K.Data & EXTENDED))
		{
			bool InsertChar = K.c >= ' '; //  AND K.c < 128

			if (K.c == 9)
			{
				if ((Flags & TVF_SELECTION) AND Start.Y() != End.Y())
				{
					if (Flags & TVF_SHIFT)
					{
						InsertChar = NOT OnMultiLineTab(FALSE);
					}
					else
					{
						InsertChar = NOT OnMultiLineTab(TRUE);
					}
				}
				else
				{
					InsertChar = TRUE;
				}
			}
			
			if (InsertChar)
			{
				char Char[2] = " ";
				Char[0] = K.c;
				OnInsertText(Char);
				OnMoveCursor(1, 0, TRUE);
			}
		}
	}

	return TRUE;
}

bool TextView::IsDirty()
{
	return Flags & TVF_DIRTY_MASK;
}

void TextView::Dirty(int Type)
{
	Flags |= Type;
}

void TextView::Clean()
{
	Flags &= ~TVF_DIRTY_MASK;
}

void TextView::OnInsertText(char *Text, int Len)
{
	OnDeleteSelection(FALSE);

	int OldLines = Doc.GetLines();
	Doc.Insert(&User, Text, Len);

	if (OldLines != Doc.GetLines())
	{
		Dirty(TVF_DIRTY_TO_EOP);
	}
	else
	{
		Dirty(TVF_DIRTY_TO_EOL);
	}

	AfterInsertText(Text, Len);
}

void TextView::OnDeleteText(Cursor *c, int Len, bool Clip)
{
	if (Clip)
	{
		ClipText((char*) *c, Len);
	}
	
	int OldLines = Doc.GetLines();
	Doc.Delete(c, Len, (Clip) ? ClipText() : 0);
	if (Clip)
	{
		Dirty(TVF_DIRTY_ALL);
	}
	else
	{
		if (OldLines != Doc.GetLines())
		{
			Dirty(TVF_DIRTY_TO_EOP);
		}
		else
		{
			Dirty(TVF_DIRTY_TO_EOL);
		}
	}
}

bool TextView::UpdateHiddenCheck()
{
	bool Status = FALSE;

	if (User.Y() < HiddenLines)
	{
		HiddenLines = User.Y();
		Dirty(TVF_DIRTY_ALL);
		Status = TRUE;
	}

	if (User.Y() >= HiddenLines + DisplayLines)
	{
		HiddenLines = User.Y() - DisplayLines + 1;
		Dirty(TVF_DIRTY_ALL);
		Status = TRUE;
	}

	return Status;
}

bool TextView::OnMoveCursor(int Dx, int Dy, bool NoSelect)
{
	bool Status = FALSE;
	bool StartSelect = (Flags & TVF_SHIFT) AND NOT (Flags & TVF_SELECTION);
	bool EndSelect = NOT (Flags & TVF_SHIFT) AND (Flags & TVF_SELECTION);
	Cursor Old = User;

	if (EndSelect)
	{
		Cursor S, E;
		if (Start < End)
		{
			S = Start;
			E = End;
		}
		else
		{
			S = End;
			E = Start;
		}

		if (Flags & TVF_GOTO_START)
		{
			User = S;
		}
		
		if (Flags & TVF_GOTO_END)
		{
			User = E;
		}

		if (Flags & TVF_EAT_MOVE)
		{
			goto AfterMove;
		}
	}

	if (Dy)
	{
		int OldX = User.X();
		User.MoveY(Dy);
		User.SetX(OldX);
	}
	else
	{
		User.MoveX(Dx);
	}

	if (Old != User OR EndSelect)
	{
		AfterMove:
		if (NOT NoSelect)
		{
			if (StartSelect)
			{
				OnStartSelection(&Old);
			}

			if (EndSelect)
			{
				OnEndSelection();
			}
		}

		Dirty(TVF_DIRTY_CURSOR);
		UpdateHiddenCheck();

		if (Flags & TVF_SELECTION)
		{
			Dirty(TVF_DIRTY_SELECTION);
			End = User;
		}

		Status = TRUE;
	}

	return Status;
}

void TextView::OnSetHidden(int Hidden)
{
	int NewHidden = limit(Hidden, 0, Doc.GetLines()-1);
	if (NewHidden != HiddenLines)
	{
		HiddenLines = NewHidden;
		Dirty(TVF_DIRTY_ALL);
	}
}

void TextView::OnSetCursor(int X, int Y)
{
	Cursor Old = User;

	User.SetY(Y);
	User.SetX(X);

	if (Old != User)
	{
		Dirty(TVF_DIRTY_CURSOR);
		UpdateHiddenCheck();

		if (Flags & TVF_SELECTION)
		{
			Dirty(TVF_DIRTY_SELECTION);
			End = User;
		}
	}
}

void TextView::OnStartSelection(Cursor *c)
{
	Start = *c;
	End = *c;
	Flags |= TVF_SELECTION;
	Dirty(TVF_DIRTY_SELECTION);
}

void TextView::OnEndSelection()
{
	Flags &= ~TVF_SELECTION;
	Dirty(TVF_DIRTY_ALL);
}

void TextView::OnDeleteSelection(bool Clip)
{
	if (Flags & TVF_SELECTION)
	{
		Cursor S;
		int Len;
		
		if (Start < End)
		{
			S = Start;
			Len = End - Start;
		}
		else
		{
			S = End;
			#ifdef CURSOR_UNDER
			S.MoveX(1);
			#endif
			Len = Start - End;
		}

		if (Len > 0)
		{
			User = S;
			OnDeleteText(&User, Len, Clip);
		}

		OnEndSelection();
	}
}

bool TextView::Open(char *Name)
{
	Doc.Load(Name);
	return Doc.CursorCreate(&User, 0, 0);
}

bool TextView::Save(char *Name)
{
	return Doc.Save(Name);
}

bool TextView::Cut()
{
	if (Flags & TVF_SELECTION)
	{
		OnDeleteSelection(TRUE);
		UpdateHiddenCheck();
		return TRUE;
	}
	return FALSE;
}

bool TextView::Copy()
{
	if (Flags & TVF_SELECTION)
	{
		Cursor S;
		int Len;

		if (Start < End)
		{
			S = Start;
			Len = End - Start;
		}
		else
		{
			S = End;
			Len = Start - End;
		}
		
		ClipText((char*) S, Len);
		return TRUE;
	}
	return FALSE;
}

bool TextView::Paste()
{
	if (ClipText())
	{
		char *Text = ClipText();
		
		// Insert text into document
		OnInsertText(Text);

		// Calculate the end point of the inserted text
		int X, Y;
		CountLinesAndLen(Text, strlen(Text), X, Y);

		// Make the cursor point to the end of the inserted text
		if (Y > 0)
		{
			User.MoveY(Y);
			User.SetX(X);
		}
		else
		{
			User.MoveX(X);
		}
	}

	OnEndSelection();
	UpdateHiddenCheck();
	return TRUE;
}

bool TextView::ClearDirty(bool Ask)
{
	return FALSE;
}

void TextView::OnSave()
{
	ClearDirty(FALSE);
}

void TextView::OnGotoLine()
{
}

void TextView::OnEscape(VKey &K)
{
	ExitApp();
}

void TextView::OnEnter(VKey &K)
{
	OnInsertText("\n");
	OnMoveCursor(1, 0, TRUE);
	
	if (bAutoIndent)
	{
		Cursor S = User;
		S.MoveY(-1);
		S.SetX(0);

		char *c = S;
		int Len = 0;
		while (IsWhite(*c) AND Len < S.LineLength())
		{
			Len++;
			c++;
		}

		if (Len > 0)
		{
			OnInsertText(S, Len);
			OnMoveCursor(Len, 0, TRUE);
		}
	}
}

