/*hdr
**      FILE:           TextView.h
**      AUTHOR:         Matthew Allen
**      DATE:           6/11/98
**      DESCRIPTION:    TextView header
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __TEXTVIEW_H
#define __TEXTVIEW_H

#include "Text.h"

////////////////////////////////////////////////////////////////////////////////
#define TVF_DIRTY_BORDER		0x40000000
#define TVF_FOCUS				0x00002000
#define TABSTOPS				16

#define WM_STATUS_UPDATE		(WM_USER+0x200)
#define WM_TITLE				(WM_USER+0x201)

////////////////////////////////////////////////////////////////////////////////
extern char *Untitled;
extern char *FindMiss;

extern int CountLines(char *Str, int Len);
extern void StrCpyToCrLf(char *d, char *s, int Len);
extern void StrCpyToLf(char *d, char *s, int Len);

////////////////////////////////////////////////////////////////////////////////
#define LGI_TEXT			"LGI_TextView"

#ifdef WIN32
class GTextView : public GControl, public TextView, public GFindReplace {
#else if BEOS
class GTextView : public GControl, public BControl, public TextView, public GFindReplace {
#endif

protected:
	#ifdef WIN32

	static GWin32Class WndClass;
	char *Class();
	DWORD Style();

	#else if BEOS

	void AttachedToWindow() { OnCreate(); }
	void Draw(BRect UpdateRect) { GdcViewDC DC(Sys_View); OnPaint(&DC); }
	void FrameMoved(BPoint Point) { Pos.Offset(Point.x-Pos.x1, Point.y-Pos.y1); OnPosChange(); }
	void FrameResized(float width, float height) { Pos.Dimension(width+1, height+1); OnPosChange(); }
	void Pulse() { OnPulse(); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	void MakeFocus(bool focused = true) { OnFocus(focused); }
	void KeyDown(const char *bytes, int32 numBytes) { Sys_KeyDown(bytes, numBytes); }
	void KeyUp(const char *bytes, int32 numBytes) { Sys_KeyUp(bytes, numBytes); }
	void MouseDown(BPoint point) { Sys_MouseDown(point); }
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) { Sys_MouseMoved(point, transit, message); }

	#endif

	char	*AppName;
	int		Border;
	#ifdef WIN32
	GdcWin32Ttf Font;
	#else if BEOS
	GdcBeTtf Font;
	#endif

	Cursor	Old;
	int		TopMargin;
	int		LeftMargin;
	int		LineHt;
	int		Spacer;

	int		TabSize;
	int		*TabStops;
	int		WrapAtCol;
	bool	UrlDetect;

	int		ExtraFlags;
	GScrollBar *VScroll;
	GScrollBar *HScroll;
	bool	SetScrollBars(bool x, bool y);

	bool	MouseSelect;
	void	Title(char *s);
	void	SetStatus(int i, char *s);

public:
	// Data access
	TextDocument *GetDoc() { return &Doc; }

	// Construction
	#ifdef WIN32
	GTextView(GWindow *notify = 0, char *appname = 0, LOGFONT *LogFont = 0);
	#else if BEOS
	GTextView(GWindow *notify = 0, char *appname = 0, BFont *LogFont = 0);
	#endif
	~GTextView();

	// API
	char *Name();
	bool Name(char *n);

	// Options
	void SetBorder(int b) { Border = b; }
	void SetWrapAtCol(int i) { WrapAtCol = i; }
	int GetWrapAtCol() { return WrapAtCol; }
	void SetURLDetect(bool b) { UrlDetect = b; }
	bool GetURLDetect() { return UrlDetect; }

	// Clipboard IO
	bool Cut();
	bool Copy();
	bool Paste();

	// Actions
	void Dirty(int Type);
	bool ClearDirty(bool Ask);
	void CalcCursorPos(Cursor c, int &x, int &y);
	void CalcMousePos(Cursor &c, VMouse m);
	void DrawCursor(GdcBasePrimitives *pDC, Cursor &c, COLOUR col);
	bool ClipText(char *Str, int Len = -1);
	char *ClipText();
	bool Open(char *Name);
	bool UpdateHiddenCheck();
	void UpdateScrollBars();
	void AfterInsertText(char *Text, int Len);
	bool DoFind();
	void SelectWord(VMouse &m);

	// Object Events
	bool OnMoveCursor(int Dx, int Dy, bool NoSelect = FALSE);
	void OnSetCursor(int X, int Y);
	bool OnFind();
	bool OnReplace();
	bool OnMultiLineTab(bool In);
	void OnSetHidden(int Hidden);
	void OnPosChange();
	void OnCreate();
	virtual void OnUrl(char *Url);

	// Window Events
	void OnFocus(bool f);
	void OnMouseClick(VMouse &m);
	void OnMouseMove(VMouse &m);
	void OnKey(VKey &k);
	void OnPaint(GdcBasePrimitives *pDC);
	int OnEvent(GMessage *Msg);
	void OnPulse();
};


#endif