/*hdr
**      FILE:           Mail.h
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Mail app header
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __MAIL_H
#define __MAIL_H

#include "Inet.h"
#include "base64.h"

#define SMTP_PORT			25
#define POP3_PORT			110
#define MAX_LINE_SIZE			1024
#define MAX_NAME_SIZE			64
#define EMAIL_LINE_SIZE			76

// MIME content types
#define CONTENT_NONE			0
#define CONTENT_BASE64			1
#define CONTENT_QUOTED_PRINTABLE	2
#define CONTENT_OCTET_STREAM		3

// Helper functions
extern GdcTtf Font;
extern void TokeniseStrList(char *Str, List<char> &Output, char *Delim);
extern char ConvHexToBin(char c);
extern char *GetField(char *s);
extern void DecodeAddrName(char *Start, char *&Name, char *&Addr, char *DefaultDomain);

class FileDescriptor : public GObject {
protected:
	// Global to read/write
	int Size;

	// Read from file
	VFile File;
	VFile *Embeded;
	int Offset;

	// Write to memory
	uchar *Data;

public:
	FileDescriptor(VFile *embed, int Offset, int Size, char *Name);
	FileDescriptor(char *name);
	FileDescriptor();
	~FileDescriptor();

	// Access functions
	VFile *GotoObject();		// Get data to read
	uchar *GetData();		// Get data from write
	int Sizeof();

	// Decode MIME data to memory
	bool Decode(	char *ContentType,
			char *ContentTransferEncoding,
			char *MimeData,
			int MimeDataLength);
};

class AddressDescriptor : public GObject {
protected:
public:
	bool Status;
	bool CC;
	char *Name;
	char *Addr;

	AddressDescriptor();
	~AddressDescriptor();

	int Sizeof();
	bool Serialize(VFile &f, bool Write);
};

class MailMessage {
public:
	List<AddressDescriptor>		To;
	AddressDescriptor			From;
	AddressDescriptor			Reply;
	char*						Subject;
	char*						Text;
	char*						MessageID;
	List<FileDescriptor>		Attachments;
	char*						InternetHeader;

	MailMessage();
	virtual ~MailMessage();
	void Empty();
	void Decode(char *DefaultDomain = 0);
};

class MailProtocol {
protected:
	char Buffer[MAX_LINE_SIZE];
	IConnection *Socket;

	bool Read();
	bool Write(char *Buf = NULL);
	virtual void OnUserMessage(char *Str) {}

public:
	#ifdef MAIL_GUI
	ProgressList Info;
	#endif

	MailProtocol();
	virtual ~MailProtocol();
};

class MailSmtp : public MailProtocol {
protected:
	bool ReadReply(char *Str);
	bool WriteText(char *Str);

public:
	char *ProgramName;

	MailSmtp();
	~MailSmtp();

	bool Open(IConnection *S, char *RemoteHost, char *UserName, int Port = SMTP_PORT);
	bool Send(MailMessage *Msg, bool Mime = FALSE);
	bool Close();
};

typedef List<MailMessage> MailList;

class MailPop3 : public MailProtocol {
protected:
	bool ReadReply();
	int GetInt();
	bool MailIsEnd(char *Ptr, int Len);

	char LastLine[1024];
	int Messages;

public:
	MailPop3();
	~MailPop3();

	int GetMessages() { return Messages; }
	bool Open(IConnection *S, char *RemoteHost, int Port, char *User, char *Password);
	bool Receive(int Message, MailMessage *Msg, bool Delete = TRUE);
	bool Delete(int Message);
	int Sizeof(int Message);
	bool Close();
};

class ListItem : public GObject, public GEvent {

	COLOUR c;
	int Lines;

public:
	ListItem(char *Str, COLOUR Col);

	int X();
	int Y();
	bool Name(char *s);
	int GetLines() { return Lines; }
	char *Name() { return GObject::Name(); }
	void OnPaint(GdcBasePrimitives *pDC);
};

typedef List<ListItem> ItemList;

class MailTest : public GAppWindow, public IConnection {

	#ifdef WIN32
	static GWin32Class	Class;
	#endif

	COLOUR		Red;
	COLOUR		Blue;
	COLOUR		Green;
	COLOUR		White;
	COLOUR		Black;
	COLOUR		DkGrey;
	COLOUR		LtGrey;

	int		VisibleLines;
	int		HiddenLines;
	int		TotalLines;
	ItemList	Items;

public:
	MailTest(ISockets *Winsock);
	~MailTest();

	bool Print(COLOUR c, char *s, ...);

	void UpdateScrollBars();

	void OnCmdLine(int Args, char **Arg);
	void OnMouseClick(VMouse &m);
	void OnMouseMove(VMouse &m);
	void OnKey(VKey &k);
	void OnPaint(GdcBasePrimitives *pDC);
	int OnEvent(int Msg, int a, int b);

	void SendTest();
	void OnRead(char *Data, int Len);
	void OnWrite(char *Data, int Len);
	void OnError(int ErrorCode, char *ErrorDescription);
	void OnInfomation(char *Str);
};


#endif