/*hdr
**      FILE:           Mail.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Mail app
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <stdlib.h>
#include <sys/timeb.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "gui.h"

#include "Mail.h"

//////////////////////////////////////////////////////////////////////////////////////////////////
extern void MDStringToDigest(unsigned char digest[16], char *Str);

char ConvHexToBin(char c)
{
	if (c >= '0' AND c <= '9')
		return c - '0';
	if (c >= 'a' AND c <= 'f')
		return c + 10 - 'a';
	if (c >= 'A' AND c <= 'F')
		return c + 10 - 'A';
	return 0;
}

char *GetField(char *s)
{
	char *e = s;

	while (*e)
	{
		if (*e == '\n')
		{
			if (*(e+1) != 9)
			{
				break;
			}
		}

		e++;
	}

	int Size = (int) e - (int) s;
	char *Str = NEW(char[Size]);
	if (Str)
	{
		char *In = s;
		char *Out = Str;

		while (In <= e)
		{
			if (*In == '\r')
			{
			}
			else if (*In == '\n')
			{
				if (*(In + 1) == 9)
				{
					In++;
				}
			}
			else
			{
				*Out++ = *In;
			}

			In++;
		}

		*Out++ = 0;
	}

	return Str;
}

bool ValidNonWSStr(char *s)
{
	if (s AND *s)
	{
		while (*s AND strchr(" \r\t\n", *s))
		{
			s++;
		}

		if (*s)
		{
			return TRUE;
		}
	}

	return FALSE;
}

void TokeniseStrList(char *Str, List<char> &Output, char *Delim)
{
	if (Str AND Delim)
	{
		char *s = Str;
		while (*s)
		{
			char *e = s;
			for (; *e AND NOT strchr(Delim, *e); e++);
			int Len = (int)e - (int)s;
			if (Len > 0)
			{
				char *Temp = NEW(char[Len+1]);
				if (Temp)
				{
					memcpy(Temp, s, Len);
					Temp[Len] = 0;
					Output.Insert(Temp);
				}
			}

			s = e;
			for (; *s AND strchr(Delim, *s); s++);
		}
	}
}

void DecodeAddrName(char *Start, char *&Name, char *&Addr, char *DefaultDomain)
{
	if (Start)
	{
		char StartChar = '(';
		char EndChar = ')';

		if (strchr(Start, '<') AND strchr(Start, '>'))
		{
			StartChar = '<';
			EndChar = '>';
		}

		char *Str1 = 0;
		char *Str2 = 0;

		char *StartName = strchr(Start, StartChar);
		if (StartName)
		{
			// has name too
			*StartName = 0;

			char *EndName = strchr(StartName + 1, EndChar);
			if (EndName)
			{
				*EndName = 0;
				Str1 = NewStr(Start);
				Str2 = NewStr(StartName + 1);
			}
			else
			{
				// error, but recover addr anyway
				Str1 = NewStr(Start);
			}
		}
		else
		{
			if (strchr(Start, '@'))
			{
				// has no name, but has a complete email address
				Str1 = NewStr(Start);
			}
			else if (ValidNonWSStr(Start))
			{
				// has just a name, so we assume the email address has the same
				// host as our address.

				if (DefaultDomain)
				{
					char Addr[256];
					strcpy(Addr, Start);
					char *At = strchr(DefaultDomain, '@');
					if (At)
					{
						strcat(Addr, At);
					}
					else
					{
						strcat(Addr, "@");
						strcat(Addr, DefaultDomain);
					}
					Str1 = NewStr(Addr);
				}
			}
		}

		char *Delim = " \r\n\t<>()\"'";
		if (Str1)
		{
			if (strchr(Str1, '@'))
			{
				Addr = TrimStr(Str1, Delim);

				if (Str2 AND strchr(Str2, '@'))
				{
					DeleteArray(Str2);
				}
			}
			else
			{
				Name = TrimStr(Str1, Delim);
			}
		}

		if (Str2)
		{
			if (strchr(Str2, '@'))
			{
				Addr = TrimStr(Str2, Delim);
			}
			else
			{
				Name = TrimStr(Str2, Delim);
			}
		}

		DeleteArray(Str1);
		DeleteArray(Str2);
	}
}

void StrCopyToEOL(char *d, char *s)
{
	if (d AND s)
	{
		while (*s AND *s != '\r' AND *s != '\n')
		{
			*d++ = *s++;
		}

		*d = 0;
	}
}

/*
		int Found = 0;
		int Ints[8];
		char Num[32];
		char *In = Buffer;
		char *Out = 0;

		while (*In)
		{
			if (Out)
			{
				if (IsDigit(*In))
				{
					*Out++ = *In;
				}
				else
				{
					*Out++ = 0;
					Ints[Found++] = atoi(Num);
					Out = NULL;
				}
			}
			else
			{
				if (IsDigit(*In))
				{
					Out = Num;
					*Out++ = *In;
				}
			}

			In++;
		}

		int Size = 0;
		for (int i=0; i<Found; i++)
		{
			if (Ints[i] != Message)
			{
				// probably the size
				Size = Ints[i];
			}
		}

*/
//////////////////////////////////////////////////////////////////////////////////////////////////
FileDescriptor::FileDescriptor()
{
	Embeded = 0;
	Offset = 0;
	Size = 0;
	Data = 0;
}

FileDescriptor::FileDescriptor(VFile *embed, int offset, int size, char *name)
{
	Embeded = embed;
	Offset = offset;
	Size = size;
	Data = 0;

	if (name)
	{
		Name(name);
	}
}

FileDescriptor::FileDescriptor(char *name)
{
	Embeded = 0;
	Offset = 0;
	Size = 0;
	Data = 0;

	if (name)
	{
		Name(name);
		if (File.Open(name, O_READ))
		{
			Size = File.GetSize();
			File.Close();
		}
	}
}

FileDescriptor::~FileDescriptor()
{
	DeleteArray(Data);
}

VFile *FileDescriptor::GotoObject()
{
	if (Embeded)
	{
		Embeded->Seek(Offset, SEEK_SET);
		return Embeded;
	}
	else if (Name() AND File.Open(Name(), O_READ))
	{
		return &File;
	}

	return NULL;
}

int FileDescriptor::Sizeof()
{
	return Size;
}

uchar *FileDescriptor::GetData()
{
	return Data;
}

bool FileDescriptor::Decode(	char *ContentType,
				char *ContentTransferEncoding,
				char *MimeData,
				int MimeDataLen)
{
	bool Status = FALSE;
	int Content = CONTENT_NONE;

	if (ContentType AND ContentTransferEncoding)
	{
		// Content-Type: application/octet-stream; name="Scribe.opt"
		Content = CONTENT_OCTET_STREAM;

		if (strnistr(ContentTransferEncoding, "base64", 1000))
		{
			Content = CONTENT_BASE64;
		}

		if (strnistr(ContentTransferEncoding, "quoted-printable", 1000))
		{
			Content = CONTENT_QUOTED_PRINTABLE;
		}

		if (Content != CONTENT_NONE);
		{
			char *NameKey = "name";
			char *n = strnistr(ContentType, NameKey, 1000);
			if (n)
			{
				char *Equal = strchr(n, '=');
				if (Equal)
				{
					Equal++;
					while (*Equal AND *Equal == '\"')
					{
						Equal++;
					}

					char *End = strchr(Equal, '\"');
					if (End)
					{
						*End = 0;
					}

					Name(Equal);
					Status = TRUE;
				}
			}
		}
	}

	if (Status AND MimeData AND MimeDataLen > 0 AND Content != CONTENT_NONE)
	{
		Status = FALSE;
		char *Base64 = NEW(char[MimeDataLen]);

		switch (Content)
		{
			case CONTENT_OCTET_STREAM:
			{
				Size = 0;
				DeleteObj(Data);
				Data = NEW(uchar[MimeDataLen]);
				if (Data)
				{
					Size = MimeDataLen;
					memcpy(Data, MimeData, Size);
					Status = TRUE;
				}
				break;
			}
			case CONTENT_QUOTED_PRINTABLE:
			{
				Size = 0;
				DeleteObj(Data);
				Data = NEW(uchar[MimeDataLen+1]);
				if (Data)
				{
					char *Out = (char*) Data;
					for (int i=0; i<MimeDataLen; i++)
					{
						if (MimeData[i] == '=')
						{
							if (MimeData[i+1] == '\r' OR MimeData[i+1] == '\n')
							{
								if (MimeData[i+1] == '\r') i++;
								if (MimeData[i+1] == '\n') i++;
							}
							else
							{
								i++;
								char c = ConvHexToBin(MimeData[i]) << 4;
								i++;
								c |= ConvHexToBin(MimeData[i]) & 0xF;

								Size++;
								*Out++ = c;
							}
						}
						else
						{
							Size++;
							*Out++ = MimeData[i];
						}
					}

					*Out++ = 0;
					Status = TRUE;
				}
				break;
			}
			case CONTENT_BASE64:
			{
				int InCount = 0;
				int OutCount = 0;
				char *In = MimeData;
				char *Out = Base64;
				while (InCount < MimeDataLen)
				{
					if (*In != '\r' AND *In != '\n')
					{
						*Out++ = *In;
						OutCount++;
					}
					InCount++;
					In++;
				}

				Size = BufferLen_64ToBin(OutCount);
				Data = NEW(uchar[Size]);
				if (Data)
				{
					int Converted = ConvertBase64ToBinary(Data, Size, Base64, OutCount);
					Status = Converted <= Size AND Converted >= Size - 3;
					if (Status)
					{
						Size = Converted;
					}
					else
					{
						DeleteArray(Data);
						Size = 0;
					}
				}
				break;
			}
		}
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
AddressDescriptor::AddressDescriptor()
{
	Status = FALSE;
	CC = FALSE;
	Addr = Name = 0;
}

AddressDescriptor::~AddressDescriptor()
{
	DeleteArray(Name);
	DeleteArray(Addr);
}

int AddressDescriptor::Sizeof()
{
	return	SizeofStr(Name) +
		SizeofStr(Addr);
}

bool AddressDescriptor::Serialize(VFile &f, bool Write)
{
	bool Status = TRUE;
	if (Write)
	{
		WriteStr(f, Name);
		WriteStr(f, Addr);
	}
	else
	{
		DeleteArray(Name);
		Name = ReadStr(f);
		DeleteArray(Addr);
		Addr = ReadStr(f);
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////

MailProtocol::MailProtocol()
{
	Socket = 0;
	Buffer[0] = 0;
}

MailProtocol::~MailProtocol()
{
	#ifdef MAIL_GUI
	if (Info.Lock())
	{
		for (Progress *p = Info.First(); p; p = Info.Next())
		{
			DeleteObj(p);
		}
		Info.Unlock();
	}
	#endif
}

bool MailProtocol::Read()
{
	bool Status = FALSE;
	if (Socket)
	{
		Status = Socket->Read(Buffer, sizeof(Buffer), 0) > 0;
	}
	return Status;
}

bool MailProtocol::Write(char *Buf)
{
	bool Status = FALSE;
	if (Socket)
	{
		if (NOT Buf) Buf = Buffer;
		Status = Socket->Write(Buf, strlen(Buf), 0);
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
#define	VERIFY_RET_VAL(Arg) { if (NOT Arg) { if (Socket) { Socket->Close(); Socket = NULL; } return FALSE; } }

MailSmtp::MailSmtp()
{
	ProgramName = "MailSmtp Object";
}

MailSmtp::~MailSmtp()
{
}

bool MailSmtp::Open(IConnection *S, char *RemoteHost, char *UserName, int Port)
{
	Socket = S;
	if (Socket->Open(RemoteHost, Port))
	{
		// receive signon message
		VERIFY_RET_VAL(ReadReply("220"));

		// send HELO message
		sprintf(Buffer, "HELO %s\r\n", (UserName) ? UserName : "default");
		VERIFY_RET_VAL(Write());
		VERIFY_RET_VAL(ReadReply("250"));

		return TRUE;
	}

	return FALSE;
}

bool MailSmtp::WriteText(char *Str)
{
	// we have to convert all strings to CRLF in here
	bool Status = FALSE;
	if (Str)
	{
		BytePipe Temp;

		char *Start = Str;
		while (*Str)
		{
			if (*Str == '\n')
			{
				// send a string
				int Size = (int)Str-(int)Start;
				if (Str[-1] == '\r') Size--;
				Temp.Push((uchar*) Start, Size);
				Temp.Push((uchar*) "\r\n", 2);
				Start = Str + 1;
			}

			Str++;
		}

		// send the final string
		int Size = (int)Str-(int)Start;
		if (Str[-1] == '\r') Size--;
		Temp.Push((uchar*) Start, Size);

		Size = Temp.Sizeof();
		char *Data = NEW(char[Size]);
		if (Data)
		{
			Temp.Pop((uchar*) Data, Size);
			Status = Socket->Write(Data, Size, 0);
			DeleteArray(Data);
		}
	}

	return Status;
}

bool MailSmtp::Send(MailMessage *Msg, bool Mime)
{
	time_t Time = time(NULL);
	timeb tbTime;
	tm *ptm = localtime(&Time);
	unsigned long Size = 0;
	char User[MAX_NAME_SIZE];
	char LocalHost[MAX_NAME_SIZE];
	
	// get user name and local host name
	Size = sizeof(User);
	#ifdef WIN32
	GetUserName(User, &Size);
	#else if BEOS
	strcpy(User, "localhost");
	#endif
	gethostname(LocalHost, sizeof(LocalHost));

	if (Socket AND Msg)
	{
		// send MAIL message
		if (Msg->From.Addr)
		{
			sprintf(Buffer, "MAIL FROM: <%s", Msg->From.Addr);
			if (strchr(Msg->From.Addr, '@' ))
			{
				strcat(Buffer, ">\r\n");
			}
			else
			{
				sprintf(Buffer + strlen(Buffer), "@%s>\r\n", LocalHost);
			}
		}
		else
		{
			sprintf(Buffer, "MAIL FROM:<%s@%s>\r\n", User, LocalHost);
		}

		VERIFY_RET_VAL(Write());
		VERIFY_RET_VAL(ReadReply("250"));

		// send RCPT message
		for (AddressDescriptor *a = Msg->To.First(); a; a = Msg->To.Next())
		{
			sprintf(Buffer, "RCPT TO: <%s>\r\n", a->Addr);
			VERIFY_RET_VAL(Write());
			a->Status = ReadReply("25");
		}

		// send DATA message
		sprintf(Buffer, "DATA\r\n");
		VERIFY_RET_VAL(Write());
		VERIFY_RET_VAL(ReadReply("354"));

		// construct date string
		char TimeStr[MAX_NAME_SIZE+1];
		strftime(TimeStr, MAX_NAME_SIZE, "%a, %d %b %Y %H:%M:%S %Z", ptm);

		// find time zone offset and correct for DST
		ftime(&tbTime);
		if (tbTime.dstflag)
		{
			tbTime.timezone -= 60;
		}

		sprintf(TimeStr + strlen(TimeStr),
			" %2.2d%2.2d",
			-tbTime.timezone / 60,
			tbTime.timezone % 60);

		// send mail headers
		// Date:
		sprintf(Buffer, "Date: %s\r\n", TimeStr);
		VERIFY_RET_VAL(Write());

		// X-Mailer:
		sprintf(Buffer, "X-Mailer: %s\r\n", ProgramName);
		VERIFY_RET_VAL(Write());

		// Message-ID:
		if (Msg->MessageID)
		{
			sprintf(Buffer, "Message-ID: %s\r\n", Msg->MessageID);
			VERIFY_RET_VAL(Write());
		}

		// To:
		strcpy(Buffer, "To: ");
		for (a = Msg->To.First(); a; )
		{
			AddressDescriptor *NextA = Msg->To.Next();
			if (a->Name)
			{
				sprintf(Buffer + strlen(Buffer), "%s (%s)", a->Addr, a->Name);
			}
			else if (a->Addr)
			{
				sprintf(Buffer + strlen(Buffer), "%s", a->Addr);
			}

			if (NextA) strcat(Buffer, ",\r\n\t");
			a = NextA;
		}
		strcat(Buffer, "\r\n");
		VERIFY_RET_VAL(Write());

		// From:
		if (Msg->From.Addr)
		{
			sprintf(Buffer, "From: %s", Msg->From.Addr);
			if (Msg->From.Name)
			{
				sprintf(Buffer + strlen(Buffer), " (%s)\r\n", Msg->From.Name);
			}
			else
			{
				strcat(Buffer, "\r\n");
			}
		}
		else
		{
			sprintf(Buffer, "From: %s@%s\r\n", User, LocalHost);
		}
		VERIFY_RET_VAL(Write());

		// Reply-To:
		if (Msg->Reply.Addr)
		{
			sprintf(Buffer, "Reply-To: %s", Msg->Reply.Addr);
			if (Msg->Reply.Name)
			{
				sprintf(Buffer + strlen(Buffer), " (%s)\r\n", Msg->Reply.Name);
			}
			else
			{
				strcat(Buffer, "\r\n");
			}
			VERIFY_RET_VAL(Write());
		}

		// Subject:
		sprintf(Buffer, "Subject: %s\r\n", (Msg->Subject) ? Msg->Subject : "");
		VERIFY_RET_VAL(Write());

		if (Mime)
		{
			bool MultiPart = (Msg->Text) AND (Msg->Attachments.GetItems() > 0);
			char Separator[256];
			#ifdef WIN32
			sprintf(Separator, "---- =_NextPart_000_%08.8X.%08.8X", time(NULL), GetCurrentProcessId());
			#else if BEOS
			sprintf(Separator, "---- =_NextPart_000_%08.8X", time(NULL));
			#endif

			sprintf(Buffer, "MIME-Version: 1.0\r\n");
			VERIFY_RET_VAL(Write());
			
			if (MultiPart)
			{
				sprintf(Buffer, "Content-Type: multipart/mixed; boundary=\"%s\"\r\n", Separator);
				VERIFY_RET_VAL(Write());
				VERIFY_RET_VAL(Write("Status: RO\r\n\r\n"));
			}
			
			if (Msg->Text)
			{
				if (MultiPart)
				{
					sprintf(Buffer, "\r\n--%s\r\n", Separator);
					VERIFY_RET_VAL(Write());
				}

				strcpy(Buffer, "Content-Type: text/plain; charset=\"us-ascii\"\r\n");
				VERIFY_RET_VAL(Write());
				strcpy(Buffer, "Content-Transfer-Encoding: 7bit\r\n\r\n");
				VERIFY_RET_VAL(Write());
				VERIFY_RET_VAL(WriteText(Msg->Text));
				VERIFY_RET_VAL(Write("\r\n\r\n"));
			}

			FileDescriptor *FileDes = Msg->Attachments.First();
			while (FileDes)
			{
				VFile *F = FileDes->GotoObject();
				if (F)
				{
					int InLen = FileDes->Sizeof();
					uchar *InBuf = NEW(uchar[InLen]);
					if (InBuf)
					{
						F->Read(InBuf, InLen);
						int OutLen = BufferLen_BinTo64(InLen);
						char *OutBuf = NEW(char[OutLen]);
						if (OutBuf)
						{
							int Len = ConvertBinaryToBase64(OutBuf, OutLen, InBuf, InLen);
							if (Len > 0)
							{
								// successfully encoded attachment
								// so write a MIME segment for this

								if (MultiPart)
								{
									sprintf(Buffer, "\r\n--%s\r\n", Separator);
									VERIFY_RET_VAL(Write());
								}

								char FileName[256];
								char *s = FileDes->Name(), *d = FileName;
								while (*s)
								{
									if (*s != '\\')
									{
										*d++ = *s++;
									}
									else
									{
										*d++ = '\\';
										*d++ = '\\';
										s++;
									}
								}
								*d = 0;

								/*
								char *pName = strchr(FileName, '\\');
								while (pName)
								{
									FileName = pName;
									pName = strchr(FileName+1, '\\');
								}
								if (*FileName == '\\') FileName++;
								*/

								sprintf(Buffer, "Content-Type: application/x-zip-compressed; name=\"%s\"\r\n", FileName);
								VERIFY_RET_VAL(Write());
								sprintf(Buffer, "Content-Transfer-Encoding: base64\r\n\r\n");
								VERIFY_RET_VAL(Write());

								#ifdef MAIL_GUI
								Progress *Prog = 0;
								if (OutLen > (20<<10))
								{
									int Lines = OutLen / EMAIL_LINE_SIZE;
									int ActualSize = OutLen + (Lines * 2);

									Prog = NEW(Progress);
									if (Prog)
									{
										Prog->SetLimits(0, ActualSize);
										Prog->SetScale(1.0 / 1024.0);
										Prog->SetType("K");
										Prog->SetDescription("Sending attachment...");
										Prog->Value(0);
										Socket->SetParameter(5150, (int) Prog);

										if (Info.Lock())
										{
											Info.Insert(Prog);
											Info.Unlock();
										}
									}
								}
								#endif

								int Pos = 0;
								while (Pos < Len)
								{
									int Size = Len - Pos;
									int WriteLen = min(Size, EMAIL_LINE_SIZE);

									memcpy(Buffer, OutBuf + Pos, WriteLen);
									strcpy(Buffer + WriteLen, "\r\n");
									VERIFY_RET_VAL(Write());
									Pos += WriteLen;
								}

								#ifdef MAIL_GUI
								if (Prog)
								{
									Socket->SetParameter(5150, 0);
									
									if (Info.Lock())
									{
										Info.Delete(Prog);
										DeleteObj(Prog);
										Info.Unlock();
									}
								}
								#endif
							}

							DeleteArray(OutBuf);
						}

						DeleteArray(InBuf);
					}
				}

				FileDes = Msg->Attachments.Next();
			}

			if (MultiPart)
			{
				// write final separator
				sprintf(Buffer, "\r\n--%s--\r\n", Separator);
				VERIFY_RET_VAL(Write());
			}

			// send message terminator and receive reply
			strcpy(Buffer, ".\r\n");
			VERIFY_RET_VAL(Write());
			VERIFY_RET_VAL(ReadReply("250"));
		}
		else
		{
			// send a blank line
			VERIFY_RET_VAL(Write("\r\n"));
			
			// send message text
			VERIFY_RET_VAL(WriteText(Msg->Text));

			// send message terminator and receive reply
			strcpy(Buffer, "\r\n.\r\n");
			VERIFY_RET_VAL(Write());
			VERIFY_RET_VAL(ReadReply("250"));
		}

		return TRUE;
	}

	return FALSE;
}

bool MailSmtp::Close()
{
	if (Socket)
	{
		// send QUIT message
		sprintf(Buffer, "QUIT\r\n");
		VERIFY_RET_VAL(Write());
		VERIFY_RET_VAL(ReadReply("221"));

		Socket->Close();
		Socket = NULL;

		return TRUE;
	}

	return FALSE;
}

bool MailSmtp::ReadReply(char *Str)
{
	bool Status = FALSE;
	if (Socket AND Str)
	{
		if (Socket->Read(Buffer, sizeof(Buffer), 0) > 0)
		{
			char *p = strtok(Buffer, "\n");
			while (p)
			{
				if (*(p + 3) == ' ')
				{
					if (NOT strncmp(p, Str, strlen(Str)))
					{
						Status = TRUE;
					}

					p = 0;
				}
				else
				{
					p = strtok(NULL, "\n");
				}
			}
    		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
MailPop3::MailPop3()
{
	Messages = 0;
}

MailPop3::~MailPop3()
{
}

int MailPop3::GetInt()
{
	char Buf[32];
	char *Start = strchr(Buffer, ' ');
	if (Start)
	{
		Start++;
		char *End = strchr(Start, ' ');
		if (End)
		{
			int Len = (int) (End - Start);
			memcpy(Buf, Start, Len);
			Buf[Len] = 0;
			return atoi(Buf);
		}
	}
	return 0;
}

bool MailPop3::ReadReply()
{
	bool Status = FALSE;
	if (Socket)
	{
		if (Socket->Read(Buffer, sizeof(Buffer), 0) > 0)
		{
			Status = (Buffer[0] == '+');
    		}
	}

	return Status;
}

bool MailPop3::Open(IConnection *S, char *RemoteHost, int Port, char *User, char *Password)
{
	bool Status = FALSE;
	bool RemoveMail = FALSE;

	if (S AND
		RemoteHost AND
		User AND
		Password)
	{
		Socket = S;
		if (Socket->Open(RemoteHost, Port))
		{
			char Key[256];
			bool GotKey = FALSE;

			// receive signon message
			VERIFY_RET_VAL(ReadReply());

			// get key here
			char *Start = strchr(Buffer, '<');
			if (Start)
			{
				char *End = strchr(Start, '>');
				if (End)
				{
					int Len = ((int) (End - Start)) + 1;
					memcpy(Key, Start, Len);
					Key[Len] = 0;

					GotKey = TRUE;
				}
			}

			// login
			if (FALSE) // GotKey
			{
				// using encrypted password
				unsigned char Digest[16];
				char HexDigest[33];

				// append password
				strcat(Key, Password);

				ZeroObj(Digest);
				MDStringToDigest(Digest, Key);
				for (int i = 0; i < 16; i++) sprintf(HexDigest + (i*2), "%02.2x", Digest[i]);
				HexDigest[32] = 0;

				sprintf(Buffer, "APOP %s %s\r\n", User, HexDigest);
				VERIFY_RET_VAL(Write());
				VERIFY_RET_VAL(ReadReply());
			}
			else
			{
				// have to use non-key method
				sprintf(Buffer, "USER %s\r\n", User);
				VERIFY_RET_VAL(Write());
				VERIFY_RET_VAL(ReadReply());

				sprintf(Buffer, "PASS %s\r\n", Password);
				VERIFY_RET_VAL(Write());
				if (NOT ReadReply())
				{
					if (Socket)
					{
						Socket->Close();
						Socket = NULL;
					}
				}
				else
				{
					Status = TRUE;

					// see how many messages there are
					VERIFY_RET_VAL(Write("STAT\r\n"));
					VERIFY_RET_VAL(ReadReply());
					Messages = GetInt();
				}
			}
		}
	}

	return Status;
}

bool MailPop3::MailIsEnd(char *Ptr, int Len)
{
	char *c = LastLine + strlen(LastLine);

	while (Len--)
	{
		*c++ = *Ptr;
		if (*Ptr == '\n')
		{
			*c = 0;

			if (stricmp(LastLine, ".\r\n") == 0)
			{
				return TRUE;
			}

			c = LastLine;
		}

		Ptr++;
	}

	*c = 0;
	return FALSE;
}

#define IsDigit(c) ((c) >= '0' AND (c) <= '9')

bool MailPop3::Receive(int Message, MailMessage *Msg, bool Delete)
{
	bool Status = FALSE;

	if (Msg AND Socket AND Message >= 0 AND Message < Messages)
	{
		// get messages
		sprintf(Buffer, "LIST %i\r\n", Message + 1);
		VERIFY_RET_VAL(Write());
		VERIFY_RET_VAL(ReadReply());

		LastLine[0] = 0;
		int Size = 0;
		char *s = strchr(Buffer, ' ');
		if (s)
		{
			s = strchr(s+1, ' ');
			if (s)
			{
				Size = atoi(s);

				if (Size > 0)
				{
					BytePipe Temp;

					sprintf(Buffer, "RETR %i\r\n", Message + 1);
					VERIFY_RET_VAL(Write());

					int ReadLen = Socket->Read(Buffer, sizeof(Buffer), 0);
					if (ReadLen > 0 AND Buffer[0] == '+')
					{
						char *Eol = strchr(Buffer, '\n');
						if (Eol)
						{
							char *Ptr = Eol + 1;
							ReadLen -= ((int)Ptr-(int)Buffer);

							Temp.Push((uchar*) Ptr, ReadLen);

							#ifdef MAIL_GUI
							Progress *Prog = 0;
							if (Size > (20<<10))
							{
								Prog = NEW(Progress);
								if (Prog)
								{
									Prog->SetLimits(0, Size);
									Prog->SetScale(1.0 / 1024.0);
									Prog->SetType("K");
									Prog->SetDescription("Receiving email...");
									Prog->Value(0);
									Socket->SetParameter(5150, (int) Prog);

									if (Info.Lock())
									{
										Info.Insert(Prog);
										Info.Unlock();
									}
								}
							}
							#endif

							while (NOT MailIsEnd(Ptr, ReadLen))
							{
								ReadLen = Socket->Read(Buffer, sizeof(Buffer), 0);
								Temp.Push((uchar*) Buffer, ReadLen);
								Ptr = Buffer;
							}

							Status = TRUE;
							Size = max(Temp.Sizeof() - 3, 0);
							Msg->Text = NEW(char[Size+1]);

							if (Msg->Text)
							{
								Temp.Pop((uchar*) Msg->Text, Size);
								Msg->Text[Size] = 0;
							}

							#ifdef MAIL_GUI
							if (Prog)
							{
								Socket->SetParameter(5150, 0);
								
								if (Info.Lock())
								{
									Info.Delete(Prog);
									DeleteObj(Prog);
									Info.Unlock();
								}
							}
							#endif

							if (Delete)
							{
								sprintf(Buffer, "DELE %i\r\n", Message + 1);
								VERIFY_RET_VAL(Write());
								VERIFY_RET_VAL(ReadReply());
							}
						}
					}
				}
			}
		}
	}

	return Status;
}

int MailPop3::Sizeof(int Message)
{
	int Size = 0;
	if (Socket)
	{
		sprintf(Buffer, "LIST %i\r\n", Message + 1);
		VERIFY_RET_VAL(Write());
		VERIFY_RET_VAL(ReadReply());

		char *s = strchr(Buffer, ' ');
		if (s)
		{
			s = strchr(s+1, ' ');
			if (s)
			{
				Size = atoi(s);
			}
		}
	}

	return Size;
}

bool MailPop3::Delete(int Message)
{
	if (Socket)
	{
		sprintf(Buffer, "DELE %i\r\n", Message + 1);
		VERIFY_RET_VAL(Write());
		VERIFY_RET_VAL(ReadReply());

		return TRUE;
	}

	return FALSE;
}

bool MailPop3::Close()
{
	if (Socket)
	{
		// logout
		VERIFY_RET_VAL(Write("QUIT\r\n"));
		VERIFY_RET_VAL(ReadReply());

		Socket->Close();
		Socket = 0;
		Messages = 0;
		return TRUE;
	}
	return FALSE;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
MailMessage::MailMessage()
{
	Subject = NULL;
	Text = NULL;
	MessageID = NULL;
}

MailMessage::~MailMessage()
{
	Empty();
}

void MailMessage::Empty()
{
	DeleteArray(Subject);
	DeleteArray(Text);
	DeleteArray(MessageID);

	for (AddressDescriptor *a = To.First(); a; a = To.Next())
	{
		DeleteObj(a);
	}
	To.Empty();

	for (FileDescriptor *f = Attachments.First(); f; f = Attachments.Next())
	{
		DeleteObj(f);
	}
	Attachments.Empty();
}

#define MODE_FIELDS			1
#define MODE_WAIT_MSG			2
#define MODE_TEXT_MIME			3
#define MODE_BIN_MIME			4
#define MODE_WAIT_TYPE			5
#define MODE_WAIT_BOUNDRY		6

#define MF_UNKNOWN			1
#define MF_SUBJECT			2
#define MF_TO				3
#define MF_FROM				4
#define MF_REPLY_TO			5
#define MF_CONTENT_TYPE			6
#define MF_CC				7

void MailMessage::Decode(char *DefaultDomain)
{
	if (Text)
	{
		char Line[1024];
		bool Done = FALSE;
		char *Current = Text;
		char *Next = Text;
		bool Mime = FALSE;
		char *Boundry = 0;
		char *ContentType = 0;
		char *ContentTransferEncoding = 0;
		char *ContentDisposition = 0;
		char *StartText = 0;
		char *StartBin = 0;
		char *End = Text + strlen(Text);
		BytePipe NewText;

		int Mode = MODE_FIELDS;
		int Field = 0;
		char *FieldStart = 0;
		char *FieldEnd = 0;

		while (NOT Done)
		{
			bool MimeBoundry = FALSE;
			int Size = 0;

			if (End < Next)
			{
				Done = TRUE;
				Next = Current;
				Line[0] = 0;
				Size = 0;
			}
			else
			{
				Next = strchr(Current, '\n');
				if (NOT Next)
				{
					Next = Current + strlen(Current);
				}

				if (Next)
				{
					Size = (int)Next-(int)Current;
					if (Size < sizeof(Line)-1)
					{
						memcpy(Line, Current, Size);
						Line[Size] = 0;
						Next++;
					}
					else
					{
						// trucate invalid line
						Line[0] = 0;
						Size = 0;
						break;
					}
				}
			}

			if (Size > 2)
			{
				if (Mime AND Boundry AND strncmp(Line, "--", 2) == 0)
				{
					int BLen = strlen(Boundry);
					if (strncmp(Boundry, Line+2, BLen) == 0)
					{
						// found segment
						MimeBoundry = TRUE;

						if (strncmp(Line + 2 + BLen, "--", 2) == 0)
						{
							// end mime
							Done = TRUE;
						}
						else
						{
							// what type is it?
							// Mode = MODE_WAIT_TYPE;
						}
					}
				}
			}

			switch (Mode)
			{
				case MODE_FIELDS:
				{
					int NewField = 0;
					char *NewFieldStart = 0;
					char *Colon = strchr(Line, ':');
					if (Colon)
					{
						*Colon = 0;
						NewFieldStart = strchr(Current, ':') + 1;

						NewField = MF_UNKNOWN;
						if (stricmp(Line, "Subject") == 0)	NewField = MF_SUBJECT;
						if (stricmp(Line, "To") == 0)		NewField = MF_TO;
						if (stricmp(Line, "From") == 0)		NewField = MF_FROM;
						if (stricmp(Line, "Reply-To") == 0)	NewField = MF_REPLY_TO;
						if (stricmp(Line, "Content-Type") == 0) NewField = MF_CONTENT_TYPE;
						if (stricmp(Line, "CC") == 0)		NewField = MF_CC;
					}

					if ((NewField > 0) OR (Size < 2))
					{
						if (Field > MF_UNKNOWN AND FieldStart)
						{
							int Len = max(((int)Current - (int)FieldStart), 0);
							char *FeildStr = NEW(char[Len+1]);
							if (FeildStr)
							{
								char *In = FieldStart, *Out = FeildStr;
								while (In < Current)
								{
									if (*In != '\r' AND *In != '\n')
									{
										*Out++ = *In;
									}
									In++;
								}
								*Out++ = 0;

								switch (Field)
								{
									case MF_SUBJECT:
									{
										Subject = TrimStr(FeildStr);
										break;
									}
									case MF_TO:
									{
										List<char> Addr;
										TokeniseStrList(FeildStr, Addr, ",");

										for (char *RawAddr = Addr.First(); RawAddr; RawAddr = Addr.Next())
										{
											AddressDescriptor *a = NEW(AddressDescriptor);
											if (a)
											{
												DecodeAddrName(RawAddr, a->Name, a->Addr, DefaultDomain);
												if (ValidStr(a->Addr))
												{
													To.Insert(a);
												}
											}

											DeleteArray(RawAddr);
										}
										break;
									}
									case MF_CC:
									{
										List<char> Addr;
										TokeniseStrList(FeildStr, Addr, ",");

										for (char *RawAddr = Addr.First(); RawAddr; RawAddr = Addr.Next())
										{
											AddressDescriptor *a = NEW(AddressDescriptor);
											if (a)
											{
												DecodeAddrName(RawAddr, a->Name, a->Addr, DefaultDomain);
												if (ValidStr(a->Addr))
												{
													a->CC = TRUE;
													To.Insert(a);
												}
											}

											DeleteArray(RawAddr);
										}
										break;
									}
									case MF_FROM:
									{
										DecodeAddrName(FeildStr, From.Name, From.Addr, DefaultDomain);
										break;
									}
									case MF_REPLY_TO:
									{
										DecodeAddrName(FeildStr, Reply.Name, Reply.Addr, DefaultDomain);
										break;
									}
									case MF_CONTENT_TYPE:
									{
										char *Sep = strchr(FeildStr, ';');
										if (Sep)
										{
											*Sep = 0;
											char *Type = TrimStr(FeildStr);
											if (strnicmp(Type, "multipart/", 10) == 0)
											{
												// ok, it's a MIME of some sort... lets go into MIME 
												// decode mode...
												Mime = TRUE;
											}

											char *Start = stristr(Sep + 1, "boundary=");
											if (Start)
											{
												// skip over the id
												Start += 9;
												if (*Start == '\"')
												{
													// attempt to read the start and end of the boundry key
													Start++;
													char *End = strchr(Start, '\"');
													if (End)
													{
														*End = 0;
														Boundry = NewStr(Start);
													}
												}
												else
												{
													// not delimited.... hmmm... grrrr!
													char *End = Start;
													while (	*End AND
															NOT strchr(" \t\r\n", *End) AND
															*End != ';')
													{
														End++;
													}

													*End = 0;
													Boundry = NewStr(Start);
												}
											}

											DeleteArray(Type);
										}
										break;
									}
								}

								DeleteArray(FeildStr);
							}
						}

						Field = NewField;
						FieldStart = NewFieldStart;
					}

					if (Size < 2)
					{
						int Size = (int)Current - (int)Text;
						InternetHeader = NEW(char[Size+256]);
						if (InternetHeader)
						{
							char *In = Text, *Out = InternetHeader;
							while (Size--)
							{
								if (*In == '\r');
								else if (*In == '\n')
								{
									*Out++ = '\r';
									*Out++ = '\n';
								}
								else
								{
									*Out++ = *In;
								}
								In++;
							}
							*Out++ = 0;
						}

						if (Mime)
						{
							Mode = MODE_WAIT_BOUNDRY;
						}
						else
						{
							Mode = MODE_WAIT_MSG;
						}
					}
					break;
				}
				case MODE_WAIT_MSG:
				{
					if (Size > 1)
					{
						NewText.Push((uchar*)Current, strlen(Current));
						Done = TRUE;
					}
					break;
				}
				case MODE_WAIT_BOUNDRY:
				{
					if (MimeBoundry)
					{
						Mode = MODE_WAIT_TYPE;
					}
					break;
				}
				case MODE_WAIT_TYPE:
				{
					if (Size > 2)
					{
						char *Colon = strchr(Line, ':');
						if (Colon)
						{
							*Colon = 0;
							if (stricmp(Line, "Content-Type") == 0)
							{
								ContentType = GetField(strchr(Current, ':') + 1);
							}

							if (stricmp(Line, "Content-Transfer-Encoding") == 0)
							{
								ContentTransferEncoding = GetField(strchr(Current, ':') + 1);
							}

							if (stricmp(Line, "Content-Disposition") == 0)
							{
								ContentDisposition = GetField(strchr(Current, ':') + 1);
							}
						}
					}
					else
					{
						if (ContentType)
						{
							if (strnistr(ContentType, "text/plain", 1000))
							{
								Mode = MODE_TEXT_MIME;
							}
							else
							{
								Mode = MODE_BIN_MIME;
							}
						}
					}
					break;
				}
				case MODE_TEXT_MIME:
				{
					if (NOT StartText)
					{
						StartText = Current;
					}
					else if (MimeBoundry)
					{
						NewText.Push((uchar*)StartText, (int) Current - (int) StartText);
						Mode = MODE_WAIT_TYPE;
						StartText = 0;
					}
					break;
				}
				case MODE_BIN_MIME:
				{
					if (NOT StartBin)
					{
						StartBin = Current;
					}
					else if (MimeBoundry)
					{
						int Size = (int) Current - (int) StartBin;
						FileDescriptor *File = NEW(FileDescriptor);
						if (	File AND
							File->Decode(	ContentType,
									ContentTransferEncoding,
									StartBin,
									Size))
						{
							Attachments.Insert(File);
						}
						else
						{
							// save the attachment as text
							// at least it doesn't get lost totally
							NewText.Push((uchar*) StartBin, Size);
						}

						Mode = MODE_WAIT_TYPE;
						StartBin = 0;
					}
					break;
				}
			}

			Current = Next;
		}

		// delete temp stuff
		DeleteArray(Boundry);
		DeleteArray(ContentType);
		DeleteArray(ContentTransferEncoding);
		DeleteArray(ContentDisposition);

		// Set text
		int TextSize = NewText.Sizeof();
		DeleteArray(Text);
		Text = NEW(char[TextSize+1]);
		if (Text)
		{
			NewText.Pop((uchar*) Text, TextSize);
			Text[TextSize] = 0;
		}
	}
}