/*hdr
**      FILE:           Filter2.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           25/3/97
**      DESCRIPTION:    Graphics file filters
**
**      Copyright (C) 1997, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////

int FindHeader(int Offset, char *Str, VFile &f)
{
	int i = 0;
	
	if (Offset >= 0)
	{
		f.Seek(Offset, SEEK_SET);
	}
	
	while (NOT f.Eof())
	{
		char c;
		
		f >> c;
		if (Str[i] == c || Str[i] == '?')
		{
			i++;
			if (!Str[i])
			{
				return TRUE;
			}
		}
		else
		{
			i = 0;
		}
	}

	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Windows BMP file format
#define BMP_ID			0x4D42

#pragma pack(push, before_pack)
#pragma pack(1)

class BMP_FILE {
public:
	ushort		Type;
	ulong		Size;
	ushort		Reserved1;
	ushort		Reserved2;
	ulong		OffsetToData;
};

class BMP_WININFO {
public:
	long		Size;
	long		Sx;
	long		Sy;
	ushort		Planes;
	ushort		Bits;
	ulong		Compression;
	ulong		DataSize;
	long		XPels;
	long		YPels;
	ulong		ColoursUsed;
	ulong		ColourImportant;
};

class BMP_OS2INFO {
public:
	long		Size;
	ushort		Sx;
	ushort		Sy;
	ushort		Planes;
	ushort		Bits;
};

#pragma pack(pop, before_pack)

bool GdcBmp::SetSize(long *Info, GdcDCBase *pDC)
{
	bool Status = FALSE;

	switch (*Info)
	{
		case sizeof(BMP_WININFO):
		{
			BMP_WININFO *i = (BMP_WININFO*) Info;
			ActualBits = i->Bits;
			ScanSize = BMPWIDTH(i->Sx * i->Bits);
			Status = pDC->Create(i->Sx, i->Sy, max(i->Bits, 8), ScanSize);
			break;
		}

		case sizeof(BMP_OS2INFO):
		{
			BMP_OS2INFO *i = (BMP_OS2INFO*) Info;
			ActualBits = i->Bits;
			ScanSize = BMPWIDTH(i->Sx * i->Bits);
			Status = pDC->Create(i->Sx, i->Sy, max(i->Bits, 8), ScanSize);
			Seek(sizeof(BMP_OS2INFO)-sizeof(BMP_WININFO), SEEK_CUR);
			break;
		}
	}

	return Status;
}

bool GdcBmp::ReadImage(GdcBasePrimitives *pDC)
{
	bool Status = FALSE;
	ActualBits = 0;
	ScanSize = 0;

	if (pDC)
	{
		BMP_FILE File;
		BMP_WININFO Info;
		GdcRegion Cr;

		if (	Read(&File, sizeof(File)) AND
			Read(&Info, sizeof(Info)) AND
			File.Type == BMP_ID AND
			SetSize((long*) &Info, pDC))
		{
			if (pDC->GetBits() <= 8)
			{
				int Colours = 1 << ActualBits;
				GdcPalette *Palette = NEW(GdcPalette);
				if (Palette)
				{
					Palette->SetSize(Colours);
					GdcRGB *Start = (*Palette)[0];
					if (Start)
					{
						Read(Start, sizeof(GdcRGB)*Colours);
						Palette->SwapRAndB();
					}

					Palette->Update();
					pDC->Palette(Palette);
				}
			}

			GdcSurface *pMem = GetSurface(pDC);
			Seek(File.OffsetToData, SEEK_SET);

			Status = TRUE;

			if (Info.Compression == BI_RLE8)
			{
				int Remaining = GetSize() - GetPosition();
				uchar *Data = NEW(uchar[Remaining]);
				if (Data)
				{
					if (Read(Data, Remaining) == Remaining)
					{
						int x=0, y=pDC->Y()-1;
						uchar *p = Data;
						bool Done = FALSE;

						while (NOT Done AND p < Data + Remaining)
						{
							uchar Length = *p++;
							uchar Colour = *p++;

							if (Length == 0)
							{
								switch (Colour)
								{
									case 0:
									{
										x = 0;
										y--;
										break;
									}
									case 1:
									{
										Done = TRUE;
										break;
									}
									case 2:
									{
										x += *p++;
										y -= *p++;
										break;
									}
									default:
									{
										// absolute mode
										uchar *Pixel = (*pDC)[y];
										if (Pixel AND y >= 0 AND y < pDC->Y())
										{
											int Len = min(Colour, pDC->X() - x);
											if (Len > 0)
											{
												memcpy(Pixel + x, p, Len);
												x += Colour;
												p += Colour;
											}
										}
										else
										{
											p += Colour;
										}
										if ((int) p & 1) p++;
										break;
									}
								}
							}
							else
							{
								// encoded mode
								uchar *Pixel = (*pDC)[y];
								if (Pixel AND y >= 0 AND y < pDC->Y())
								{
									int Len = min(Length, pDC->X() - x);
									if (Len > 0)
									{
										memset(Pixel + x, Colour, Len);
										x += Length;
									}
								}
							}
						}
					}
				}
			}
			else if (Info.Compression == BI_RLE4)
			{
			}
			else
			{
				// straight
				switch (ActualBits)
				{
					case 1:
					{
						uchar *Buffer = NEW(uchar[ScanSize]);
						if (Buffer)
						{
							for (int y=pMem->y-1; y>=0; y--)
							{
								Status &= Read(Buffer, ScanSize) == ScanSize;
								if (Status)
								{
									uchar Mask = 0x80;
									uchar *d = Buffer;
									for (int x=0; x<pDC->X(); x++)
									{
										pDC->Colour((*d & Mask) ? 1 : 0);
										pDC->Set(x, y);

										Mask >>= 1;
										if (NOT Mask)
										{
											Mask = 0x80;
											d++;
										}
									}
								}
								else
								{
									int k=0;
								}
							}

							DeleteArray(Buffer);
						}
						break;
					}
					case 4:
					{
						uchar *Buffer = NEW(uchar[ScanSize]);
						if (Buffer)
						{
							for (int y=pMem->y-1; y>=0; y--)
							{
								Status &= Read(Buffer, ScanSize) == ScanSize;

								uchar *d = Buffer;
								for (int x=0; x<pDC->X(); x++)
								{
									if (x & 1)
									{
										pDC->Colour(*d & 0xf);
										d++;
									}
									else
									{
										pDC->Colour(*d >> 4);
									}
									pDC->Set(x, y);
								}
							}

							DeleteArray(Buffer);
						}
						break;
					}
					case 8:
					case 16:
					case 24:
					case 32:
					{
						for (int i=pMem->y-1; i>=0; i--)
						{
							Status &= Read(pMem->Base + (pMem->Line * i), ScanSize) == ScanSize;
						}
						break;
					}
					default:
					{
						Status = FALSE;
						break;
					}
				}
			}

			Cr.ZOff(pDC->X()-1, pDC->Y()-1);
			pDC->Update(GDC_BITS_CHANGE|GDC_PAL_CHANGE);
		}
		else
		{
			Cr.ZOff(-1, -1);
		}

		pDC->ClipRgn(&Cr);
	}
	
	return Status;
}

bool GdcBmp::WriteImage(GdcBasePrimitives *pDC)
{
	bool Status = FALSE;
	
	if (pDC)
	{
		BMP_FILE File;
		BMP_WININFO Info;
		GdcSurface *pMem = GetSurface(pDC);
		int Colours = (pMem->Bits <= 8) ? 1 << pMem->Bits : 0;

		if (	NOT pMem OR
			NOT pMem->x OR
			NOT pMem->y)
		{
			return FALSE;
		}

		File.Type = BMP_ID;
		File.OffsetToData = sizeof(BMP_FILE) + sizeof(BMP_WININFO) + (Colours * 4);
		File.Size = File.OffsetToData + (pMem->Line * pMem->y);
		File.Reserved1 = 0;
		File.Reserved2 = 0;

		Info.Size = sizeof(BMP_WININFO);
		Info.Sx = pMem->x;
		Info.Sy = pMem->y;
		Info.Planes = 1;
		Info.Bits = pMem->Bits;
		Info.Compression = 0;
		Info.DataSize = 0;
		Info.XPels = 3000;
		Info.YPels = 3000;
		Info.ColoursUsed = Colours;
		Info.ColourImportant = Colours;

		if (	Write(&File, sizeof(File)) AND
			Write(&Info, sizeof(Info)))
		{
			if (pMem->Bits <= 8)
			{
				int Done = 0;
				GdcPalette *Palette = pDC->Palette();

				if (Palette)
				{
					GdcRGB *Start = (*Palette)[0];
					if (Start)
					{
						Palette->SwapRAndB();
						Write(Start, sizeof(GdcRGB)*Palette->GetSize());
						Palette->SwapRAndB();
						Done += Palette->GetSize();
					}
				}

				char Grey[4];
				Grey[3] = 0;

				while (Done < Colours)
				{
					int C = Done * 256 / Colours;

					Grey[0] = C;
					Grey[1] = C;
					Grey[2] = C;

					Write(Grey, 4);
					Done++;
				}
			}

			int Bytes = BMPWIDTH(pMem->x * pMem->Bits);
			Status = TRUE;
			for (int i=pMem->y-1; i>=0; i--)
			{
				Status &= Write(pMem->Base + (i * pMem->Line), Bytes) == Bytes;
			}
		}
	}
	
	return Status;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Run length encoded DC
GdcRleDC::GdcRleDC()
{
	Flags = GDC_RLE_COLOUR | GDC_RLE_READONLY;
	Length = 0;
	Alloc = 0;
	Data = 0;
	ScanLine = 0;
	Key = 0;

	pMem = NEW(GdcSurface);
	if (pMem)
	{
		pMem->Base = 0;
		pMem->x = 0;
		pMem->y = 0;
		pMem->Bits = 0;
		pMem->Line = 0;
		pMem->Flags = 0;
	}
}

GdcRleDC::~GdcRleDC()
{
	Empty();
}

void GdcRleDC::Empty()
{
	Length = 0;
	Alloc = 0;
	DeleteArray(Data);
	DeleteArray(ScanLine);
}

bool GdcRleDC::Create(int x, int y, int bits, int LineLen)
{
	bool Status = GdcDibSection::Create(x, y, bits, LineLen);
	if (Status)
	{
		Flags &= ~GDC_RLE_READONLY;
	}
	return Status;
}

bool GdcRleDC::CreateInfo(int x, int y, int bits)
{
	bool Status = FALSE;

	DeleteObj(pMem);
	pMem = NEW(GdcSurface);
	if (pMem)
	{
		pMem->x = x;
		pMem->y = y;
		pMem->Bits = bits;
		pMem->Base = 0;
		pMem->Line = 0;
		pMem->Flags = 0;

		Flags |= GDC_RLE_READONLY;

		if (bits == 8)
		{
			Flags |= GDC_RLE_MONO;
		}
	}

	return Status;
}

void GdcRleDC::ReadOnly(bool Read)
{
	if (Read)
	{
		Flags |= GDC_RLE_READONLY;
	}
	else
	{
		if (pMem)
		{
			pMem->Base = NEW(uchar[pMem->Line * pMem->y]);
		}
		else
		{
			// just in case... shouldn't ever happen
			// am I paranoid?
			// 'course :)
			Create(1, 1, 24);
		}

		Flags &= ~GDC_RLE_READONLY;
	}
}

bool GdcRleDC::ReadOnly()
{
	return (Flags & GDC_RLE_READONLY) != 0;
}

void GdcRleDC::Mono(bool Mono)
{
	if (Mono)
	{
		Flags |= GDC_RLE_MONO;
		Flags &= ~GDC_RLE_COLOUR;
	}
	else
	{
		Flags &= ~GDC_RLE_MONO;
		Flags |= GDC_RLE_COLOUR;
	}
}

bool GdcRleDC::Mono()
{
	return (Flags & GDC_RLE_MONO) != 0;
}

bool GdcRleDC::SetLength(int Len)
{
	bool Status = TRUE;

	if (Len + 1 > Alloc)
	{
		int NewAlloc = (Len + 0x4000) & (~0x3FFF);
		uchar *Temp = NEW(uchar[NewAlloc]);
		if (Temp)
		{
			if (NewAlloc > Length)
			{
				memset(Temp + Length, 0, NewAlloc - Length);
			}

			memcpy(Temp, Data, Length);
			DeleteArray(Data);
			DeleteArray(ScanLine);
			Data = Temp;
			Length = Len;
			Alloc = NewAlloc;
		}
		else
		{
			Status = FALSE;
		}
	}
	else
	{
		// TODO: resize Data to be smaller if theres a
		// considerable size change
		Length = Len;
	}

	return Status;
}

void GdcRleDC::Update(int UpdateFlags)
{
	bool Error = FALSE;

	if (	(UpdateFlags & GDC_BITS_CHANGE) AND
		NOT (Flags & GDC_RLE_READONLY))
	{
		Empty();

		COLOUR Key = Get(0, 0);
		ulong Pos = 0;
		ulong PixelSize = GetBits() / 8;
		
		for (ulong y=0; NOT Error AND y<Y(); y++)
		{
			for (ulong x=0; x<X(); )
			{
				ulong Skip = 0;
				ulong Pixels = 0;
				uchar *Bits = 0;

				while (Get(x, y) == Key AND x < X())
				{
					Skip++;
					x++;
				}

				Bits = (*this)[y] + (x * PixelSize);
				while (Get(x, y) != Key AND x < X())
				{
					Pixels++;
					x++;
				}

				if (Flags & GDC_RLE_COLOUR)
				{
					if (SetLength(	Length +
							(2 * sizeof(ulong)) +
							(Pixels * PixelSize)))
					{
						*((ulong*)(Data+Pos)) = Skip;
						Pos += sizeof(Skip);
						*((ulong*)(Data+Pos)) = Pixels;
						Pos += sizeof(Pixels);
						
						int l = Pixels * PixelSize;
						memcpy(Data+Pos, Bits, l);
						Pos += l;
					}
					else
					{
						Error = TRUE;
						break;
					}
				}
				else if (Flags & GDC_RLE_MONO)
				{
					if (SetLength(	Length + (2 * sizeof(ulong))))
					{
						*((ulong*)(Data+Pos)) = Skip;
						Pos += sizeof(Skip);
						*((ulong*)(Data+Pos)) = Pixels;
						Pos += sizeof(Pixels);
					}
					else
					{
						Error = TRUE;
						break;
					}
				}
			}
		}
	}

	if (Error)
	{
		Empty();
	}
	else
	{
		FindScanLines();
	}
}

bool GdcRleDC::FindScanLines()
{
	bool Status = FALSE;
	bool Error = FALSE;
	DeleteArray(ScanLine);

	if (Data)
	{
		ScanLine = NEW(uchar*[Y()]);
		if (ScanLine)
		{
			ulong Pos = 0;
			ulong PixelSize = GetBits() / 8;
			
			for (ulong y=0; NOT Error AND y<Y(); y++)
			{
				ScanLine[y] = Data + Pos;
				for (ulong x=0; x<X();)
				{
					ulong Skip = 0;
					ulong Pixels = 0;

					Skip = *((ulong*)(Data + Pos));
					Pos += sizeof(Skip);
					Pixels = *((ulong*)(Data + Pos));
					Pos += sizeof(Pixels);

					if (Flags & GDC_RLE_COLOUR)
					{
						Pos += Pixels * PixelSize;
					}

					x += Pixels + Skip;
				}

				if (x != X() OR Pos > Length)
				{
					Error = TRUE;
				}
			}
		}

		if (Error)
		{
			DeleteArray(ScanLine);
		}
		else
		{
			Status = TRUE;
		}
	}

	return Status;
}

void GdcRleDC::Draw(GdcBasePrimitives *Dest, int Ox, int Oy)
{
	#ifdef ADJUST_ORIGIN
	int OriX, OriY;
	Dest->GetOrigin(OriX, OriY);
	Ox += OriX;
	Oy += OriY;
	#endif

	if (Dest AND Data)
	{
		bool ReBuildScanInfo = FALSE;
		
		if (NOT ScanLine) ReBuildScanInfo = TRUE;
		if (ScanLine AND Data AND ScanLine[0] != Data) ReBuildScanInfo = TRUE;

		if (ReBuildScanInfo)
		{
			// need to rebuild scan line info
			if (NOT FindScanLines())
			{
				// couldn't create scan line info so quit
				return;
			}
		}

		GdcRegion S;
		S.ZOff(X()-1, Y()-1);
		GdcRegion D;
		D = S;
		D.Offset(Ox, Oy);
		GdcRegion DClip = D;
		GdcRegion Temp = Dest->ClipRgn();
		D.Bound(&Temp);

		if (DClip == D AND (*Dest)[0])
		{
			// no clipping needed
			int PixLen = Dest->GetBits() / 8;

			if (Flags & GDC_RLE_COLOUR)
			{
				if (Dest->GetBits() == GetBits())
				{
					for (int y=0; y<Y(); y++)
					{
						uchar *s = ScanLine[y];
						uchar *d = (*Dest)[Oy+y];
						for (int x=0; x<X(); )
						{
							x += *((ulong*)s); s+=4;
							int Pixels = *((ulong*)s); s+=4;
							int Len = Pixels*PixLen;
							memcpy(d + ((Ox+x)*PixLen), s, Len);
							x += Pixels;
							s += Len;
						}
					}
				}
			}
			else if (Flags & GDC_RLE_MONO)
			{
				GdcApplicator *pDApp = Dest->Applicator();

				for (int y=0; y<Y(); y++)
				{
					uchar *s = ScanLine[y];
					for (int x=0; x<X(); )
					{
						x += *((ulong*)s); s+=4;
						int Pixels = *((ulong*)s); s+=4;
						pDApp->SetPtr(Ox+x, Oy+y);
						pDApp->Rectangle(Pixels, 1);
						x += Pixels;
					}
				}
			}
		}
		else if (DClip.Valid())
		{
			// clip
			int PixLen = Dest->GetBits() / 8;

			if (Flags & GDC_RLE_COLOUR)
			{
				if (Dest->GetBits() == GetBits())
				{
					for (int y=0; y<Y(); y++)
					{
						int Fy = Oy+y;
						if (Fy >= Temp.y1 AND Fy < Temp.y2) // clip y
						{
							uchar *s = ScanLine[y];
							uchar *d = (*Dest)[Fy];
							for (int x=0; x<X(); )
							{
								x += *((ulong*)s); s+=4;
								int Pixels = *((ulong*)s); s+=4;

								int Fx = Ox+x;
								int PreClipPixels = max(0, Temp.x1 - Fx);
								int PostClipPixels = max(0, (Pixels + Fx) - Temp.x2 - 1);
								int PixelsLeft = Pixels - (PreClipPixels + PostClipPixels);
								if (PixelsLeft > 0) // clip x
								{
									memcpy(	d + ((Fx - PreClipPixels) * PixLen),
										s + (PreClipPixels * PixLen),
										PixelsLeft * PixLen);
								}

								x += Pixels;
								s += Pixels * PixLen;
							}
						}
					}
				}
			}
			/*
			else if (Flags & GDC_RLE_MONO)
			{
				GdcApplicator *pDApp = Dest->Applicator();

				for (int y=0; y<Y(); y++)
				{
					uchar *s = ScanLine[y];
					for (int x=0; x<X(); )
					{
						x += *((ulong*)s); s+=4;
						int Pixels = *((ulong*)s); s+=4;
						pDApp->SetPtr(Ox+x, Oy+y);
						pDApp->Rectangle(Pixels, 1);
						x += Pixels;
					}
				}
			}
			*/
		}
	}
}

int GdcRleDC::SizeOf()
{
	return (sizeof(int) * 5) + Length;
}

bool GdcRleDC::Read(VFile &F)
{
	Empty();
	
	int Len = 0, x = 0, y = 0, bits = 0;
	char Monochrome;
	F >> Len;
	if (SetLength(Len))
	{
		F >> Key;
		F >> x;
		F >> y;
		F >> bits;
		F >> Monochrome;
		Mono(Monochrome);
		CreateInfo(x, y, bits);
		F.Read(Data, Len);
		return NOT F.GetStatus();
	}

	return FALSE;
}

bool GdcRleDC::Write(VFile &F)
{
	if (Data)
	{
		char Monochrome = Mono();
		
		F << Length;
		F << Key;
		F << X();
		F << Y();
		F << GetBits();
		F << Monochrome;

		F.Write(Data, Length);

		return NOT F.GetStatus();
	}

	return FALSE;
}

void GdcRleDC::Move(GdcRleDC *pDC)
{
	if (pDC)
	{
		Key = pDC->Key;
		Flags = pDC->Flags;
		Length = pDC->Length;
		Alloc = pDC->Alloc;
		Data = pDC->Data;
		ScanLine = pDC->ScanLine;
		pMem = pDC->pMem;
		pPalette = pDC->pPalette;

		pDC->pPalette = 0;
		pDC->pMem = 0;
		pDC->Length = 0;
		pDC->Alloc = 0;
		pDC->Data = 0;
		pDC->ScanLine = 0;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////
// MBM
#define MBM_HEADER				"MBMG"

// ver 1.0 defines
#define MBM_EOF					0
#define MBM_INFO				1
#define MBM_PALETTE				2
#define MBM_BITMAP				3

// ver 2.0 defines
#define MBM_INFO2				4
#define MBM_RLE					5
#define MBM_COMMENT				6
#define MBM_FREE_SPACE				7

class MbmObject {
public:
	MbmObject() {}
	virtual ~MbmObject() {}
	virtual int SizeOf() { return 0; }
};

class MbmString : public MbmObject {

	char *Str;

public:
	MbmString()
	{
		Str = 0;
	}

	~MbmString()
	{
		DeleteArray(Str);
	}

	operator char *() const { return Str; }
	char *operator =(char *c) { DeleteArray(Str); Str = NewStr(c); return Str; }
	int SizeOf() { return sizeof(ulong) + (Str) ? strlen(Str) : 0; }

	bool Read(VFile &F)
	{
		ulong Len = 0;
		F >> Len;
		DeleteArray(Str);
		Str = NEW(char[Len+1]);
		if (Str)
		{
			F.Read(Str, Len);
			Str[Len] = 0;
		}

		return NOT F.GetStatus();
	}

	bool Write(VFile &F)
	{
		ulong Len = 0;
		if (Str)
		{
			Len = strlen(Str);
			F << Len;
			F.Write(Str, Len);
		}
		else
		{
			F << Len;
		}

		return NOT F.GetStatus();
	}
};

// MBM_INFO
class MbmSimpleInfo : public MbmObject {
public:
	ushort Sx;
	ushort Sy;
	ushort Bits;
	ushort Gamma; // * 1000

	bool Read(VFile &F)
	{
		F >> Sx;
		F >> Sy;
		F >> Bits;
		F >> Gamma;

		return NOT F.GetStatus();
	}

	bool Write(VFile &F)
	{
		F << Sx;
		F << Sy;
		F << Bits;
		F << Gamma;

		return NOT F.GetStatus();
	}

	int SizeOf() { return sizeof(ushort) * 4; }
};

// MBM_INFO2
class MbmComplexInfo : public MbmObject {
public:
	MbmString	Name;
	ulong		Sx;
	ulong		Sy;
	ulong		Bits;
	ulong		LineLength;
	ushort		Channels[4];

	int SizeOf()
	{
		return	Name.SizeOf() + 
			sizeof(Sx) +
			sizeof(Sy) +
			sizeof(Bits) +
			sizeof(LineLength) +
			sizeof(Channels);
	}

	bool Read(VFile &F)
	{
		Name.Read(F);
		F >> Sx;
		F >> Sy;
		F >> Bits;
		F >> LineLength;
		F.Read(Channels, sizeof(Channels));

		return NOT F.GetStatus();
	}

	bool Write(VFile &F)
	{
		Name.Write(F);
		F << Sx;
		F << Sy;
		F << Bits;
		F << LineLength;
		F.Write(Channels, sizeof(Channels));

		return NOT F.GetStatus();
	}
};

GdcMbm::GdcMbm()
{
	pBmp = 0;
	pRle = 0;
}

GdcMbm::~GdcMbm()
{
}

bool GdcMbm::Open(char *Name, int Attrib)
{
	bool Status = VFile::Open(Name, Attrib);

	if (Status)
	{
		if (IsRead())
		{
			// check for MBM id code
			if (NOT FindHeader(0, MBM_HEADER, *this))
			{
				Status = FALSE;
				VFile::Close();
			}
		}
		else if (IsWrite())
		{
			// write MBM id code
			if (NOT Write(MBM_HEADER, 4))
			{
				Status = FALSE;
				VFile::Close();
			}
		}
	}

	return Status;
}

bool GdcMbm::Close()
{
	if (ValidHandle(Handle) AND IsWrite())
	{
		// write the EOF chunk
		int i = MBM_EOF;
		(*this) << i;
	}

	bool Status = VFile::Close();
	return Status;
}

bool GdcMbm::ReadImage(GdcBasePrimitives *pDC)
{
	pBmp = pDC;
	pRle = NULL;
	return ReadAny();
}

bool GdcMbm::WriteImage(GdcBasePrimitives *pDC)
{
	pBmp = pDC;
	pRle = NULL;
	return WriteAny();
}

bool GdcMbm::ReadRle(GdcRleDC *pDC)
{
	pBmp = NULL;
	pRle = pDC;
	return ReadAny();
}

bool GdcMbm::WriteRle(GdcRleDC *pDC)
{
	pBmp = NULL;
	pRle = pDC;
	return WriteAny();
}

bool GdcMbm::ReadAny()
{
	bool Status = FALSE;
	bool Done = FALSE;

	static MbmSimpleInfo *SInfo = NULL;
	static MbmComplexInfo *CInfo = NULL;
	static GdcPalette *Pal = NULL;

	while (NOT Done)
	{
		ulong ChunkId = 0;
		ulong Size = 0;

		(*this) >> ChunkId;
		if (ChunkId != MBM_EOF)
		{
			(*this) >> Size;
		}

		switch (ChunkId)
		{
			case MBM_EOF:
			{
				Done = TRUE;
				break;
			}
			case MBM_INFO:
			{
				DeleteObj(CInfo);
				if (NOT SInfo) SInfo = NEW(MbmSimpleInfo);

				if (SInfo)
				{
					if (NOT SInfo->Read(*this))
					{
						Done = TRUE;
						Status = FALSE;
					}
				}
				break;
			}
			case MBM_INFO2:
			{
				DeleteObj(SInfo);
				if (NOT CInfo) CInfo = NEW(MbmComplexInfo);

				if (CInfo)
				{
					if (NOT CInfo->Read(*this))
					{
						Done = TRUE;
						Status = FALSE;
					}
					else
					{
						Props.Set("Name", (char*) CInfo->Name);
					}
				}
				break;
			}
			case MBM_PALETTE:
			{
				int Entries;
				(*this) >> Entries;

				uchar *PalData = NEW(uchar[Entries]);
				if (PalData)
				{
					Read(PalData, Entries);

					DeleteObj(Pal);
					Pal = NEW(GdcPalette(PalData));
				}
				break;
			}
			case MBM_BITMAP:
			{
				if (pBmp)
				{
					bool CreateStatus = FALSE;
					int LineLength = 0;

					if (CInfo)
					{
						// Prefer complex info
						LineLength = CInfo->LineLength;
						CreateStatus = pBmp->Create(CInfo->Sx, CInfo->Sy, CInfo->Bits, CInfo->LineLength);
					}
					else if (SInfo)
					{
						// Use simple info
						LineLength = ((SInfo->Sx * SInfo->Bits) + 7) / 8;
						CreateStatus = pBmp->Create(SInfo->Sx, SInfo->Sy, SInfo->Bits, LineLength);
					}

					if (CreateStatus AND (*pBmp)[0])
					{
						Status = Read((*pBmp)[0], LineLength * pBmp->Y());

						Props.Set("X", pBmp->X());
						Props.Set("Y", pBmp->Y());
						Props.Set("Bits", pBmp->GetBits());
					}
					else
					{
						Status = FALSE;
					}

					Done = TRUE;
				}
				break;
			}
			case MBM_RLE:
			{
				if (pRle)
				{
					Status = pRle->Read(*this);
					if (Status)
					{
						Props.Set("X", pRle->X());
						Props.Set("Y", pRle->Y());
						Props.Set("Bits", pRle->GetBits());
					}

					Done = TRUE;
				}
				break;
			}
			case MBM_COMMENT:
			{
				MbmString Str;
				Str.Read(*this);
				Props.Set("Comment", Str);
				break;
			}
			default:
			{
				// Skip unknown block
				Seek(Size, SEEK_CUR);
				break;
			}
		}
	}

	if (NOT Status)
	{
		DeleteObj(SInfo);
		DeleteObj(CInfo);
		DeleteObj(Pal);
	}

	return Status;
}

#define MakeChannel(Bits, Comp)			((Bits<<8)|(Comp))

bool GdcMbm::WriteAny()
{
	bool Status = FALSE;
	bool Error = FALSE;

	static MbmComplexInfo *CInfo = NULL;
	static GdcPalette *Pal = NULL;

	GdcDCBase *pCommon = (GdcDCBase*) ((pBmp) ? pBmp : pRle);

	if (pCommon)
	{
		// Info block
		DeleteObj(CInfo);
		CInfo = NEW(MbmComplexInfo);
		if (CInfo)
		{
			char *Name = 0;
			if (Props.Get("Name", Name))
			{
				CInfo->Name = Name;
			}
			else
			{
				CInfo->Name = "Untitled";
			}

			CInfo->Sx = pCommon->X();
			CInfo->Sy = pCommon->Y();
			CInfo->Bits = pCommon->GetBits();
			CInfo->LineLength = 0;
			
			ZeroObj(CInfo->Channels);
			switch (CInfo->Bits)
			{
				case 8:
				{
					CInfo->Channels[0] = MakeChannel(8, 'I');
					break;
				}
				case 15:
				{
					CInfo->Channels[0] = MakeChannel(5, 'B');
					CInfo->Channels[1] = MakeChannel(5, 'G');
					CInfo->Channels[2] = MakeChannel(5, 'R');
					CInfo->Channels[3] = MakeChannel(1, '?');
					break;
				}
				case 16:
				{
					CInfo->Channels[0] = MakeChannel(5, 'B');
					CInfo->Channels[1] = MakeChannel(6, 'G');
					CInfo->Channels[2] = MakeChannel(5, 'R');
					break;
				}
				case 24:
				{
					CInfo->Channels[0] = MakeChannel(8, 'R');
					CInfo->Channels[1] = MakeChannel(8, 'G');
					CInfo->Channels[2] = MakeChannel(8, 'B');
					break;
				}
				case 32:
				{
					CInfo->Channels[0] = MakeChannel(8, 'B');
					CInfo->Channels[1] = MakeChannel(8, 'G');
					CInfo->Channels[2] = MakeChannel(8, 'R');
					CInfo->Channels[3] = MakeChannel(8, 'A');
					break;
				}
			}

			int ChunkId = MBM_INFO2;
			int Size = CInfo->SizeOf();

			*this << ChunkId;
			*this << Size;
			if (NOT CInfo->Write(*this))
			{
				Error = TRUE;
			}
		}

		// Palette
		if (pCommon->GetBits() <= 8)
		{
			GdcPalette *Pal = pCommon->Palette();
			if (Pal)
			{
				int I = MBM_PALETTE;
				*this << I;
				I = sizeof(ulong) + (Pal->GetSize() * 3);
				*this << I;
				I = Pal->GetSize();
				*this << I;

				for (int i=0; i<I; i++)
				{
					GdcRGB *p = (*Pal)[i];
					*this << p->R;
					*this << p->G;
					*this << p->B;
				}
			}
			else
			{
				// Grayscale?
			}
		}

		if (pBmp)
		{
			if ((*pBmp)[0])
			{
				int ChunkId = MBM_PALETTE;
				int Line = GetLineLength(pBmp);
				int Size = Line * pBmp->Y();

				*this << ChunkId;
				*this << Size;
				if (Write((*pBmp)[0], Size) != Size)
				{
					Error = TRUE;
				}
			}
			else
			{
				Error = TRUE;
			}
		}
		else if (pRle)
		{
			int ChunkId = MBM_RLE;
			int Size = pRle->SizeOf();

			*this << ChunkId;
			*this << Size;

			if (NOT pRle->Write(*this))
			{
				Error = TRUE;
			}
		}
		else
		{
			Error = TRUE;
		}

		char *Comment = NULL;
		if (Props.Get("Comment", Comment))
		{
			MbmString Str;
			Str = Comment;

			int ChunkId = MBM_COMMENT;
			int Size = Str.SizeOf();

			*this << ChunkId;
			*this << Size;

			Str.Write(*this);
		}
	}

	if (NOT Error)
	{
		Status = TRUE;
	}

	return Status;
}

