/*hdr
**      FILE:           Tiff.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           9/10/98
**      DESCRIPTION:    Tiff filter
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"


// Misc
#define TIFF_LZW_BUFFER				(16<<10)  // anyone know how big this should be??

// Tiff tag numbers
#define TAG_ImageX				0x100
#define TAG_ImageY				0x101
#define TAG_Bits				0x102
#define TAG_Palette				0x140

#define TAG_Compression				0x103
#define TAG_RowsPerStrip			0x116
#define TAG_StripOffsets			0x111
#define TAG_StripByteOffsets			0x117
#define TAG_PhotometricInterpretation		0x106
#define TAG_Predictor				0x13D

// Tiff tag data types
#define TYPE_UBYTE				1
#define TYPE_USHORT				3
#define TYPE_ULONG				4

#define TYPE_SBYTE				6
#define TYPE_SSHORT				8
#define TYPE_SLONG				9

#define TYPE_ASCII				2
#define TYPE_URATIONAL				5
#define TYPE_UNDEFINED				7
#define TYPE_SRATIONAL				10
#define TYPE_FLOAT				11
#define TYPE_DOUBLE				12

////////////////////////////////////////////////////////////////////////////////////////////
/*
void Log(char *s, ...)
{
	VFile F;
	if (s AND F.Open("log.txt", O_WRITE))
	{
		char Buffer[512];

		va_list Arg;
		va_start(Arg, s);
		vsprintf(Buffer, s, Arg);
		va_end(Arg);

		F.Seek(0, SEEK_END);
		F.Write(Buffer, strlen(Buffer));
	}
}
*/

////////////////////////////////////////////////////////////////////////////////////////////
IFD::IFD()
{
	Data = 0;
}

IFD::~IFD()
{
	DeleteArray(Data);
}

void *IFD::GetData()
{
	return (Data) ? Data : &Offset;
}

int IFD::ElementSize()
{
	switch (Type)
	{
		case TYPE_UBYTE:
		case TYPE_SBYTE:
		case TYPE_ASCII:
		case TYPE_UNDEFINED:
		{
			return 1;
		}
		case TYPE_USHORT:
		case TYPE_SSHORT:
		{
			return 2;
		}
		case TYPE_ULONG:
		case TYPE_SLONG:
		case TYPE_FLOAT:
		{
			return 4;
		}
		case TYPE_URATIONAL:
		case TYPE_SRATIONAL:
		case TYPE_DOUBLE:
		{
			return 8;
		}
	}

	return 0;
}

char *IFD::Str()
{
	if (Data)
	{
		return (char*) Data;
	}
	return (char*) &Offset;
}

int IFD::Value()
{
	if (Count == 1)
	{
		switch (Type)
		{
			case TYPE_UBYTE:
			{
				return (int) ((uchar) (Offset & 0xFF));
			}
			case TYPE_USHORT:
			{
				return (int) ((ushort) (Offset & 0xFFFF));
			}
			case TYPE_ULONG:
			{
				return (int) ((ulong) Offset);
			}
			case TYPE_SBYTE:
			{
				return (int) ((signed char) (Offset & 0xFF));
			}
			case TYPE_SSHORT:
			{
				return (int) ((short) (Offset & 0xFFFF));
			}
			case TYPE_SLONG:
			{
				return (int) ((long) Offset);
			}
			case TYPE_ASCII:
			{
				char *s = Str();
				return (s) ? atoi(s) : 0;
			}
			case TYPE_URATIONAL:
			case TYPE_UNDEFINED:
			case TYPE_SRATIONAL:
			case TYPE_FLOAT:
			case TYPE_DOUBLE:
			{
				break;
			}
		}
	}
	return 0;
}

int IFD::ArrayValue(int i)
{
	void *p = GetData();
	if (Count > 1 AND p)
	{
		switch (Type)
		{
			case TYPE_ASCII:
			case TYPE_UBYTE:
			{
				return (int) ((uchar*)p)[i];
			}
			case TYPE_USHORT:
			{
				return (int) ((ushort*)p)[i];
			}
			case TYPE_ULONG:
			{
				return (int) ((ulong*)p)[i];
			}
			case TYPE_SBYTE:
			{
				return (int) ((signed char*)p)[i];
			}
			case TYPE_SSHORT:
			{
				return (int) ((short*)p)[i];
			}
			case TYPE_SLONG:
			{
				return (int) ((long*)p)[i];
			}
		}
	}
	return 0;
}

bool IFD::Read(VFile &f)
{
	f >> Tag;
	f >> Type;
	f >> Count;
	f >> Offset;

	int Size = Count * ElementSize();
	if (Size > 4)
	{
		Data = NEW(uchar[Size]);
		if (Data)
		{
			int Pos = f.GetPosition();
			f.Seek(Offset, SEEK_SET);
			f.Read(Data, Size);
			f.Seek(Pos, SEEK_SET);
		}
	}
	return TRUE;
}

bool IFD::Write(VFile &f)
{
	f << Tag;
	f << Type;
	f << Count;
	f << Offset;
	return TRUE;
}

////////////////////////////////////////////////////////////////////////////////////////////

GdcTiff::GdcTiff()
{
	Parent = 0;
	BufSize = 0;
	Buffer = 0;
	CodeBuf = NEW(uchar[TIFF_LZW_BUFFER]);
}

GdcTiff::~GdcTiff()
{
	DeleteObj(Buffer);
	DeleteObj(CodeBuf);
}

#ifdef TIFF_USE_LZW
void GdcTiff::InitializeTable()
{
	CodeSize = 9;
	ClearCode = 1 << (CodeSize - 1);
	EoiCode = ClearCode + 1;
	NextCode = EoiCode + 1;

	ZeroObj(CodeLength);
	ZeroObj(CodePtr);
	NextBuf = CodeBuf;
	for (int i=0; i<256; i++)
	{
		CodeLength[i] = 1;
		CodePtr[i] = NextBuf++;
		*CodePtr[i] = i;
	}
}

int GdcTiff::GetNextCode()
{
	int Code = 0;
	int b = CodeSize;
	uchar Mask[] = { 0, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F, 0xFF };

	while (b > 0)
	{
		if (BitsLeft < 1)
		{
			DeleteArray(Buffer);
			BitsLeft = 0;
			Buffer = NEW(uchar[BufSize]);
			if (Buffer)
			{
				BitsLeft = Read(Buffer, BufSize) * 8;
				BufPos = Buffer;
				BitPos = 0;
				BufSize = 0;

			}

			if (BitsLeft <= 0)
			{
				return EoiCode;
			}
		}

		int ThisBits = min(8-BitPos, b);
		int Temp = Mask[ThisBits] & (*BufPos >> (8 - (ThisBits + BitPos)));
		Code |= Temp << (b - ThisBits);
		
		b -= ThisBits;
		BitsLeft -= ThisBits;
		BitPos += ThisBits;
		if (BitPos > 7)
		{
			BitPos -= 8;
			BufPos++;
		}
	}

	return Code;
}

void GdcTiff::WriteString(uchar *d, int Size)
{
	while (Size--)
	{
		if (NOT Out OR Out >= OutEnd)
		{
			CurrentScanline++;
			if (Out)
			{
				Out = (*pOutDC)[CurrentScanline];
				OutEnd = Out + ScanLength;
			}
			else
			{
				break;
			}
		}

		*Out++ = *d++;
	}

	if (Out >= OutEnd)
	{
		CurrentScanline++;
		if (Out)
		{
			Out = (*pOutDC)[CurrentScanline];
			OutEnd = Out + ScanLength;
		}
	}
}
#endif

IFD *GdcTiff::FindTag(ushort n)
{
	for (IFD *i = Blocks.First(); i; i = Blocks.Next())
	{
		if (i->Tag == n)
		{
			return i;
		}
	}

	return NULL;
}

bool GdcTiff::ProcessRead(GdcBasePrimitives *pDC)
{
	bool Status = FALSE;
	bool Error = FALSE;
	IFD *x = FindTag(TAG_ImageX);
	IFD *y = FindTag(TAG_ImageY);
	IFD *BitTag = FindTag(TAG_Bits);
	int Bits = 1;
	int Pos = GetPosition();

	if (BitTag)
	{
		switch (BitTag->Count)
		{
			case 1:
			{
				Bits = BitTag->Value();
				break;
			}
			case 3:
			{
				ushort *Depth = (ushort*) BitTag->GetData();
				if (	Depth AND
					BitTag->Type == TYPE_USHORT AND
					Depth[0] == 8 AND
					Depth[1] == 8 AND
					Depth[2] == 8)
				{
					// Baseline RGB image
					Bits = 24;
				}
				break;
			}
			case 4:
			{
				// maybe a 32-bit image?
				break;
			}
		}
	}

	if (	x AND
		y AND
		pDC AND
		pDC->Create(x->Value(), y->Value(), max(Bits, 8)))
	{
		// Read Palette
		IFD *Palette = FindTag(TAG_Palette);
		if (Palette)
		{
			ushort *s = (ushort*) Palette->GetData();
			if (s AND Palette->Type == TYPE_USHORT)
			{
				int Colours = Palette->Count / 3;
				GdcPalette *Pal = NEW(GdcPalette(0, Colours));
				if (Pal)
				{
					for (int i=0; i<Colours; i++)
					{
						GdcRGB *d = (*Pal)[i];
						if (d)
						{
							d->R = s[(0*Colours)+i] >> 8;
							d->G = s[(1*Colours)+i] >> 8;
							d->B = s[(2*Colours)+i] >> 8;
						}
					}

					pDC->Palette(Pal);
				}
			}

		}

		// Read Image
		IFD *Compression = FindTag(TAG_Compression);
		IFD *RowsPerStrip = FindTag(TAG_RowsPerStrip);
		IFD *StripOffsets = FindTag(TAG_StripOffsets);
		IFD *StripByteCounts = FindTag(TAG_StripByteOffsets);
		IFD *PhotometricInterpretation = FindTag(TAG_PhotometricInterpretation);
		if (	Compression AND
			RowsPerStrip AND
			StripOffsets AND
			StripByteCounts)
		{
			int Comp = Compression->Value();
			int Rows = RowsPerStrip->Value();
			int Interpretation = PhotometricInterpretation->Value();

			switch (Bits)
			{
				case 8:
				case 24:
				{
					int Strip = 0;
					int Bpp = Bits / 8;
					ScanLength = pDC->X() * Bpp;

					for (	int y=0;
							NOT Error AND
							y<pDC->Y() AND
							Strip < StripOffsets->Count;
							Strip++)
					{
						int Offset = StripOffsets->ArrayValue(Strip);
						BufSize = StripByteCounts->ArrayValue(Strip);
						int Limit = min(pDC->Y() - y, Rows);
						Seek(Offset, SEEK_SET);

						switch (Comp)
						{
							case 1: // uncompressed
							{
								for (int i=0; i<Limit; i++, y++)
								{
									uchar *d = (*pDC)[y];
									if (d)
									{
										Read(d, ScanLength);
									}
								}
								break;
							}
							case 32773: // Pack-bits RLE
							{
								for (int i=0; i<Limit; i++, y++)
								{
									char n;
									uchar c;
									uchar *d = (*pDC)[y];
									if (d)
									{
										// unpack data
										for (uchar *e=d+ScanLength; d<e; )
										{
											(*this) >> n;
											if (n & 0x80)
											{
												// copy run
												(*this) >> c;
												n = -n+1;
												memset(d, c, n);
											}
											else
											{
												// copy data
												n++;
												Read(d, n);
											}
											d += n;
										}
									}
								}
								break;
							}
							case 2: // modified huffman
							{
								MessageBox(	Handle(),
										"This image uses a modified huffman compression,\r\n"
										"Image does not support this.\r\n",
										"Image: Tiff",
										MB_OK);
								Error = TRUE;
								break;
							}
							case 5: // LZW compression
							{
								#ifdef TIFF_USE_LZW
								int Code = 0, OldCode = 0;
								uchar *OutString;

								CodeSize = 9;
								ClearCode = 1 << (CodeSize - 1);
								EoiCode = ClearCode + 1;
								NextCode = EoiCode + 1;
								BitPos = 0;
								BitsLeft = 0;
								pOutDC = pDC;
								CurrentScanline = y;
								Out = (*pDC)[CurrentScanline];
								OutEnd = Out + ScanLength;

								int inits = 0;

								while ((Code = GetNextCode()) != EoiCode AND NextCode < 4096)
								{
									if (Code == ClearCode)
									{
										InitializeTable();
										Code = GetNextCode();
										if (Code == EoiCode)
										{
											break;
										}
										WriteString(CodePtr[Code], CodeLength[Code]);
										OldCode = Code;
										inits++;
									}
									else
									{
										if (Code < NextCode)
										{
											WriteString(CodePtr[Code], CodeLength[Code]);

											uchar *d = CodePtr[OldCode];
											int Length = CodeLength[OldCode];
											CodeLength[NextCode] = Length + 1;
											CodePtr[NextCode] = NextBuf;
											while (Length--)
											{
												*NextBuf++ = *d++;
											}
											*NextBuf++ = *CodePtr[Code];
											NextCode++;

											OldCode = Code;
										}
										else
										{
											if (Code > NextCode)
											{
												break;
											}
											else
											{
												int Length = CodeLength[OldCode];
												CodeLength[NextCode] = Length + 1;
												CodePtr[NextCode] = NextBuf;
												uchar *d = CodePtr[OldCode];
												while (Length--)
												{
													*NextBuf++ = *d++;
												}
												*NextBuf++ = *CodePtr[OldCode];
												WriteString(CodePtr[NextCode], CodeLength[NextCode]);
												NextCode++;

												OldCode = Code;
											}
										}

										switch (NextCode)
										{
											case 511:
											case 1023:
											case 2047:
											{
												CodeSize++;
												break;
											}
											case 4095:
											{
												InitializeTable();
												break;
											}
										}
									}
								}

								y = CurrentScanline;
								#else
								MessageBox(	MainWnd->Handle(),
										"This image uses LZW, which is unsupported due to\r\n"
										"UniSys requiring a royalty for it's patented algorithm.\r\n"
										"\r\n"
										"Use PNG instead.",
										"Image: Tiff filter",
										MB_OK);
								Error = TRUE;
								#endif
								break;
							}
							default:
							{
								char Msg[256];
								sprintf(Msg, "This image uses an unknown compression method: %i", Comp);
								MessageBox(Handle(), Msg, "Image: Tiff", MB_OK);
								Error = TRUE;
								break;
							}
						}
					}

					if (Bits == 24)
					{
						// swap R and B
						for (int y=0; y<pDC->Y(); y++)
						{
							uchar *d = (*pDC)[y];
							uchar n;
							for (int x=0; x<pDC->X(); x++, d+=3)
							{
								n = d[0];
								d[0] = d[2];
								d[2] = n;
							}
						}
					}

					IFD *Predictor = FindTag(TAG_Predictor);
					if (Predictor AND Predictor->Value() == 2)
					{
						switch (Bits)
						{
							case 8:
							{
								for (int y=0; y<pDC->Y(); y++)
								{
									uchar *d = (*pDC)[y];
									for (int x=1; x<pDC->X(); x++, d++)
									{
										d[0] += d[-1];
									}
								}
								break;
							}
							case 24:
							{
								for (int y=0; y<pDC->Y(); y++)
								{
									uchar *d = (*pDC)[y] + 3;
									for (int x=1; x<pDC->X(); x++, d+=3)
									{
										d[0] += d[-3];
										d[1] += d[-2];
										d[2] += d[-1];
									}
								}
								break;
							}
						}
					}
					break;
				}
				default:
				{
					char Msg[256];
					sprintf(Msg, "Image currently doesn't support %i bit TIFF files", Bits);
					MessageBox(Handle(), Msg, "Image: Tiff", MB_OK);
					Error = TRUE;
					break;
				}
			}

			Status = NOT Error;
		}
	}

	Seek(Pos, SEEK_SET);

	return Status;
}

bool GdcTiff::ReadImage(GdcBasePrimitives *pDC)
{
	bool Status = FALSE;
	ushort n16;
	ulong n32;

	(*this) >> n16;
	if (n16 == 0x4D4D)
	{
		// file is big endian so we swap
		// Intel is little endian (i think)
		// TODO_PPC: change this
		SetSwap(TRUE);
	}

	(*this) >> n16;
	if (n16 == 42) // check magic number
	{
		bool Done = FALSE;
		bool Error = FALSE;

		ulong Offset;
		(*this) >> Offset;

		// loop through all the IFD's
		while (	NOT Done AND
			NOT Error AND
			NOT Eof())
		{
			if (Offset == 0)
			{
				Done = TRUE;
			}
			else
			{
				Seek(Offset, SEEK_SET);

				ushort IFDEntries;
				(*this) >> IFDEntries;

				for (int i=0; i<IFDEntries; i++)
				{
					IFD *Ifd = NEW(IFD);
					if (Ifd)
					{
						Blocks.Insert(Ifd);
						Ifd->Read(*this);
					}
					else
					{
						Error = TRUE;
					}
				}
			}

			if (NOT Error)
			{
				// process sub file
				Status = ProcessRead(pDC);
				Error = NOT Status;
			}

			(*this) >> Offset;
		}
	}

	return Status;
}

bool GdcTiff::ProcessWrite(GdcBasePrimitives *pDC)
{
	return FALSE;
}

bool GdcTiff::WriteImage(GdcBasePrimitives *pDC)
{
	return FALSE;
}
