/*hdr
**	FILE:		Intel.h
**	AUTHOR:		Matthew Allen
**	DATE:		11/3/97
**	DESCRIPTION:	Fast core routines for intel based machines
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include "defs.h"
#include "memdev.h"
#include "core.h"

void Add64(ulong *DestH, ulong *DestL, ulong SrcH, ulong SrcL)
{
	_asm {
		mov eax, DestL
		mov ebx, SrcL
		mov ecx, SrcH
		mov edx, DestH
		add [eax], ebx
		adc [edx], ecx
	}
}

void MemSet(void *d, int c, uint s)
{
	if (s)
	{
		c &= 0xFF;
		c = (c << 24) | (c << 16) | (c << 8) | c;

		_asm {
			mov edi, d
			mov eax, c
			mov ecx, s
			cmp ecx, 7
			jb short_scan
			test edi,3
			jz dword_aligned
			mov [edi],eax
			mov ebx,edi
			and ebx,3
			sub ecx,ebx
			add edi,ebx
		dword_aligned: sub ecx,4
		dword_loop: mov [edi],eax
			add edi,4
			sub ecx,4
			jae dword_loop
			and ecx,3
			jz done
		short_scan: mov [edi],al
			inc edi
			dec ecx
			jnz short_scan
		done:
		}
	}
}

void MemCpy(void *d, void *s, uint l)
{
	_asm  {
		mov edi, d
		mov esi, s
		mov ecx, l
		mov ebx, ecx
		and ebx, 3
		shr ecx, 2
		jz l2
		rep movsd
	l2:	or ecx, ebx
		jz l4
		rep movsb
	l4:
	}
}

void MemAnd(void *d, void *s, uint l)
{
	_asm {
		mov edi, d
		mov esi, s
		mov ecx, l
		mov ebx, ecx
		and ebx, 3
		shr ecx, 2
		jz l2
	l1:	mov eax, [esi]
		and [edi], eax
		add esi, 4
		add edi, 4
		dec ecx
		jnz l1
	l2:	or ecx, ebx
		jz l4
	l3:	mov al, [esi]
		and [edi], al
		inc edi
		inc esi
		dec ecx
		jnz l3
	l4:
	}
}

void MemXor(void *d, void *s, uint l)
{
	_asm {
		mov edi, d
		mov esi, s
		mov ecx, l
		mov ebx, ecx
		and ebx, 3
		shr ecx, 2
		jz l2
	l1:	mov eax, [esi]
		xor [edi], eax
		add esi, 4
		add edi, 4
		dec ecx
		jnz l1
	l2:	or ecx, ebx
		jz l4
	l3:	mov al, [esi]
		xor [edi], al
		inc edi
		inc esi
		dec ecx
		jnz l3
	l4:
	}
}

void MemOr(void *d, void *s, uint l)
{
	_asm {
		mov edi, d
		mov esi, s
		mov ecx, l
		mov ebx, ecx
		and ebx, 3
		shr ecx, 2
		jz l2
	l1:	mov eax, [esi]
		or [edi], eax
		add esi, 4
		add edi, 4
		dec ecx
		jnz l1
	l2:	or ecx, ebx
		jz l4
	l3:	mov al, [esi]
		or [edi], al
		inc edi
		inc esi
		dec ecx
		jnz l3
	l4:
	}
}

