/*hdr
**      FILE:           Device.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           10/4/98
**      DESCRIPTION:    Device class
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

/****************************** Includes ********************************************/

#include <stdio.h>

#include "defs.h"
#include "device.h"

/****************************** Classes *********************************************/

bool Device::Verbose = TRUE;
Device *Device::First = NULL;

bool Device::AddDevice(Device *pDev)
{
	if (First)
	{
		Device *Temp = First;
		while (Temp->Next)
		{
			Temp = Temp->Next;
		}
		Temp->Next = pDev;
		pDev->Next = 0;
		pDev->Prev = Temp;
	}
	else
	{
		First = pDev;
		pDev->Next = 0;
		pDev->Prev = 0;
	}
	
	return TRUE;
}

bool Device::RemoveDevice(Device *pDev)
{
	if (First && pDev)
	{
		Device *Temp = First;
		
		while (Temp && Temp != pDev)
		{
			Temp = Temp->Next;
		}
		
		if (Temp == pDev)
		{
			if (pDev == First) First = pDev->Next;
			if (pDev->Next) pDev->Next->Prev = pDev->Prev;
			if (pDev->Prev) pDev->Prev->Next = pDev->Next;
			return TRUE;
		}
	}
	
	return FALSE;
}

Device *Device::FindDevice(char *Name)
{
	Device *Temp = First;
	while (Temp)
	{
		if (stricmp(Name, Temp->Name) == 0)
		{
			return Temp;
		}
		Temp = Temp->Next;
	}
	
	return 0;
}

bool Device::Print(char *str, ...)
{
	if (str)
	{
		if (Verbose)
		{
			// char buffer[256];
			va_list arg;

			va_start(arg ,str);
			// vprintf(str, arg);
			va_end(arg);
		}
		
		return TRUE;
	}
	
	return FALSE;
}

bool Device::SetVerbose(bool v)
{
	bool Prev = Verbose;
	Verbose = v;
	return Prev;
}

void Device::CriticalExit(char *str, ...)
{
	if (str)
	{
		va_list arg;

		va_start(arg ,str);
		vprintf(str, arg);
		va_end(arg);
		
	}

	exit(-1);
}

Device::Device()
{
	AddDevice(this);
}

Device::~Device()
{
	RemoveDevice(this);
}
