#ifndef __DEFS_H
#define __DEFS_H

#include "windows.h"
#include "commctrl.h"

// Stupid mouse wheel defines don't work. hmmm
#define WM_MOUSEWHEEL				0x020A
#define WHEEL_DELTA				120
#define SPI_GETWHEELSCROLLLINES			104

typedef         unsigned char                   uchar;
typedef         unsigned short                  ushort;
typedef         unsigned int                    uint;
typedef         unsigned long                   ulong;
typedef         unsigned short                  HND;
typedef         unsigned short                  MESSAGE;

#define SUCCESS					1
#define FAIL					0
#define TRUE					1
#define FALSE					0
#define YES					1
#define NO					0

#define abs(a)					(((a) > 0) ? (a) : -(a))
#define min(a,b)				(((a) < (b)) ? (a) : (b))
#define max(a,b)				(((a) > (b)) ? (a) : (b))
#define limit(i,l,u)				(((i)<(l)) ? (l) : (((i)>(u)) ? (u) : (i)))
#define swap(a, b)				{ int n = a; a = b; b = n; }

#define makelong(a, b)				((a)<<16 | (b&0xFFFF))
#define loword(a)				(a&0xFFFF)
#define hiword(a)				(a>>16)

#define NOT					!
#define AND					&&
#define OR					||

#ifndef NULL
#define NULL					0
#endif

#include "assert.h"
#define ASSERT					assert

#define ZeroObj(obj)				memset(&obj, 0, sizeof(obj))
#define CountOf(array)				(sizeof(array)/sizeof(array[0]))
#define StrEmpty(s)				(s)[0] = 0
#define IsStrEmpty(s)				((NOT s) OR ((s)[0] == 0))

typedef const char *cpchar;

extern void Add64(ulong *DestH, ulong *DestL, ulong SrcH, ulong SrcL);

class quad {
public:
	ulong	High;
	ulong	Low;

	quad() {}
	quad(quad const &q)
	{
		High = q.High;
		Low = q.Low;
	}
	quad(ulong i)
	{
		High = 0;
		Low = i;
	}
	quad(ulong h, ulong l)
	{
		High = h;
		Low = l;
	}

	operator ulong()
	{
		return Low;
	}

	bool operator <(quad const &q)
	{
		return (High < q.High) OR ((High == q.High) AND (Low < q.Low));
	}

	bool operator <=(quad const &q)
	{
		return (High < q.High) OR ((High == q.High) AND (Low <= q.Low));
	}

	bool operator >(quad const &q)
	{
		return (High > q.High) OR ((High == q.High) AND (Low > q.Low));
	}

	bool operator >=(quad const &q)
	{
		return (High > q.High) OR ((High == q.High) AND (Low >= q.Low));
	}

	bool operator ==(quad const &q)
	{
		return (High == q.High) AND (Low == q.Low);
	}

	bool operator !=(quad const &q)
	{
		return (High != q.High) OR (Low != q.Low);
	}

	const quad &operator ++()
	{
		Low++;
		if (NOT Low) High++;
		return *this;
	}

	const quad &operator --()
	{
		Low--;
		if (Low == 0xFFFFFFFF) High--;
		return *this;
	}

	const quad &operator +=(quad const &q)
	{
		Add64(&High, &Low, q.High, q.Low);
		return *this;
	}

	const quad &operator -=(quad const &q)
	{
		High -= (q.High + ((Low < q.Low) ? 1 : 0));
		Low -= q.Low;
		return *this;
	}

	const quad &operator =(quad const &q)
	{
		High = q.High;
		Low = q.Low;
		return *this;
	}

	operator double()
	{
		return ((double) High * 4294967296.0) + (double) Low;
	}
};

inline quad operator +(quad const &a, quad const &b)
{
	quad Temp = a;
	Add64(&Temp.High, &Temp.Low, b.High, b.Low);
	return Temp;
}

inline quad operator -(quad const &a, quad const &b)
{
	return quad(	a.High - (b.High + ((a.Low < b.Low) ? 1 : 0)),
			a.Low - b.Low);
}

inline quad operator *(quad const &a, quad const &b)
{
	ulong La = a.Low;
	ulong Lb = b.Low;
	ulong h, l;

	_asm {
		mov eax, La
		mov ebx, Lb
		mul ebx
		mov h, edx
		mov l, eax
	}

	return quad(h, l);
}

inline quad operator /(quad const &a, quad const &b)
{
	ulong h = a.High, l = a.Low, rem = b.Low;

	_asm {
		mov edx, h
		mov eax, l
		mov ebx, rem
		div ebx
		mov l, eax
		mov rem, edx
	}

	return quad(l);
}

typedef quad Quad;

class String
{
protected:
	int Length;
	int Alloc;
	char *Data;

public:
// Constructors/Destructors
	String();
	String(char *Src);
	~String();

// Attributes & Operations
	int GetLength();
	bool SetLength(int Len);
	bool IsEmpty();
	void Empty();

	char operator[](int i);
	operator char *() const;

	// overloaded assignment
	String& operator =(String &Str);
	String& operator =(char ch);
	String& operator =(char *Str);

	// string concatenation
	String& operator +=(String &Str);
	String& operator +=(char ch);
	String& operator +=(char *Str);

	friend String operator +(String &Str1, String &Str2);
	friend String operator +(String &Str, char ch);
	friend String operator +(char ch, String &Str);
	friend String operator +(String &Str1, char *Str2);
	friend String operator +(char *Str1, String &Str2);

	// string comparison
	int Compare(char *Str);
	int CompareNoCase(char *Str);

	// simple sub-string extraction
	String Mid(int nFirst, int nCount = -1);
	String Left(int nCount);
	String Right(int nCount);

	// upper/lower/reverse conversion
	void MakeUpper();
	void MakeLower();

	// trimming whitespace (either side)
	void TrimRight();
	void TrimLeft();

	// searching (return starting index, or -1 if not found)
	// look for a single character match
	int Find(char ch);
	int ReverseFind(char ch);

	// look for a specific sub-string
	int Find(char *Sub);

	// simple formatting
	void Format(char *lpszFormat, ...);

	/*
	friend CArchive& AFXAPI operator<<(CArchive& ar, const String& string);
	friend CArchive& AFXAPI operator>>(CArchive& ar, String& string);
	*/
};

bool operator ==(String	&s1, String	&s2);
bool operator ==(String	&s1, char	*s2);
bool operator ==(char	*s1, String	&s2);
bool operator !=(String	&s1, String	&s2);
bool operator !=(String	&s1, char	*s2);
bool operator !=(char	*s1, String	&s2);
bool operator <(String	&s1, String	&s2);
bool operator <(String	&s1, char	*s2);
bool operator <(char	*s1, String	&s2);
bool operator >(String	&s1, String	&s2);
bool operator >(String	&s1, char	*s2);
bool operator >(char	*s1, String	&s2);
bool operator <=(String	&s1, String	&s2);
bool operator <=(String	&s1, char	*s2);
bool operator <=(char	*s1, String	&s2);
bool operator >=(String	&s1, String	&s2);
bool operator >=(String	&s1, char	*s2);
bool operator >=(char	*s1, String	&s2);

String operator +(String	&Str1,	String &Str2);
String operator +(String	&Str,	char   ch);
String operator +(char		ch,	String &Str);
String operator +(String	&Str1,	char   *Str2);
String operator +(char		*Str1,	String &Str2);

class GMem {

	int Size;
	HGLOBAL hMem;
	void *pMem;

public:
	GMem(int size)
	{
		Size = size;
		hMem = GlobalAlloc(GMEM_FIXED, Size);
		pMem = 0;
	}

	GMem(HGLOBAL hmem)
	{
		hMem = hmem;
		Size = GlobalSize(hMem);
		pMem = 0;
	}

	~GMem()
	{
		if (pMem)
		{
			GlobalUnlock(hMem);
			pMem = 0;
		}
		GlobalFree(hMem);
		hMem = 0;
	}

	void *Lock()
	{
		return pMem = GlobalLock(hMem);
	}

	void UnLock()
	{
		if (pMem)
		{
			GlobalUnlock(hMem);
			pMem = 0;
		}
	}

	HGLOBAL Handle() { return hMem; }
};

#endif