/*hdr
**      FILE:           Store.h
**      AUTHOR:         Matthew Allen
**      DATE:           29/10/98
**      DESCRIPTION:    Tree storage
**
**      Copyright (C) 1998 Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include "file.h"
#include "gdc2.h"
#include "gui.h"
#include "contain.h"

class StorageKit;
class StorageItem;
class StorageObj;

class StorageObj {
public:
	bool StoreDirty;
	StorageItem *Store;

	StorageObj()
	{
		StoreDirty = FALSE;
		Store = 0;
	}

	virtual ~StorageObj() {}

	virtual int Type() { return 0; }
	virtual int Sizeof() { return 0; }
	virtual bool Serialize(VFile &f, bool Write) { return FALSE; }
};

class StorageItem : public GObject {

	friend class StorageKit;
	friend class StorageItem;

	bool MoveToLoc(int NewLoc);
	bool WriteHeader(VFile &f);

protected:
public:
	int StoreType;
	int StoreSize;
	int StoreAllocatedSize;
	int StoreLoc;
	int StoreNext;
	int StorePrev;
	int StoreParent;
	int StoreChild;

	StorageKit *Tree;
	StorageItem *Next;
	StorageItem *Prev;
	StorageItem *Parent;
	StorageItem *Child;

	StorageObj *Object;

	StorageItem();
	~StorageItem();

	StorageItem *GetNext();
	StorageItem *GetPrev();
	StorageItem *GetParent();
	StorageItem *GetChild();

	StorageItem *CreateNext(StorageObj *Obj);
	StorageItem *CreateChild(StorageObj *Obj);
	StorageItem *CreateSub(StorageObj *Obj);

	int Sizeof();
	bool SetSize(int NewSize);
	bool Serialize(VFile &f, bool Write);
	VFile *GotoObject();
};

class Block {
public:
	StorageItem *Item;
	int Size;
};

class StorageKit : public GObject {

	friend class StorageItem;

protected:
	int SpaceUsed;
	int SpaceFree;
	int Objects;

	char *FileName;
	VFile File;
	bool Status;
	int StoreLoc;
	StorageItem *Root;

	StorageItem *LoadLocation(int Loc);
	void AddItem(StorageItem *Item, List<Block> &Blocks);

public:
	StorageKit(char *FileName);
	~StorageKit();

	bool GetStatus() { return Status; }
	StorageItem *GetRoot();
	StorageItem *CreateRoot(StorageObj *Obj);
	StorageItem *CreateItem(StorageObj *Obj);
	bool Serialize(VFile &f, bool Write);

	bool Compact();
	bool DeleteItem(StorageItem *Item);
	bool SeparateItem(StorageItem *Item);
	bool ReparentItem(StorageItem *Item, StorageItem *NewParent);
};

