/*hdr
**      FILE:           TextView.h
**      AUTHOR:         Matthew Allen
**      DATE:           6/11/98
**      DESCRIPTION:    TextView header
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __TEXTVIEW_H
#define __TEXTVIEW_H

#include "text.h"

////////////////////////////////////////////////////////////////////////////////
#define TVF_DIRTY_BORDER			0x40000000
#define TVF_FOCUS					0x00002000
#define TABSTOPS					16

#define WM_STATUS_UPDATE			(WM_USER+0x200)
#define WM_TITLE					(WM_USER+0x201)

////////////////////////////////////////////////////////////////////////////////
extern char *Untitled;
extern char *FindMiss;

extern int CountLines(char *Str, int Len);
extern void StrCpyToCrLf(char *d, char *s, int Len);
extern void StrCpyToLf(char *d, char *s, int Len);

////////////////////////////////////////////////////////////////////////////////
#define LGI_TEXT			"LGI_TextView"

class GTextView : public GControl, public TextView, public GFindReplace {
protected:
	static GWin32Class WndClass;
	char *Class();
	DWORD Style();

	char	*AppName;
	int	Border;
	GdcWin32Ttf Font;

	Cursor	Old;
	int	TopMargin;
	int	LeftMargin;
	int	LineHt;
	int	Spacer;

	int	TabSize;
	int	*TabStops;
	int	WrapAtCol;
	bool	UrlDetect;

	int			ExtraFlags;
	GScrollBar		*VScroll;
	GScrollBar		*HScroll;
	bool SetScrollBars(bool x, bool y);

	bool	MouseSelect;
	void	Title(char *s);
	void	SetStatus(int i, char *s);

public:
	// Data access
	TextDocument *GetDoc() { return &Doc; }

	// Construction
	GTextView(GWindow *notify = 0, char *appname = 0, LOGFONT *LogFont = 0);
	~GTextView();

	// API
	char *Name();
	bool Name(char *n);

	// Options
	void SetBorder(int b) { Border = b; }
	void SetWrapAtCol(int i) { WrapAtCol = i; }
	int GetWrapAtCol() { return WrapAtCol; }
	void SetURLDetect(bool b) { UrlDetect = b; }
	bool GetURLDetect() { return UrlDetect; }

	// Clipboard IO
	bool Cut();
	bool Copy();
	bool Paste();

	// Actions
	void Dirty(int Type);
	bool ClearDirty(bool Ask);
	void CalcCursorPos(Cursor c, int &x, int &y);
	void CalcMousePos(Cursor &c, VMouse m);
	void DrawCursor(GdcBasePrimitives *pDC, Cursor &c, COLOUR col);
	bool ClipText(char *Str, int Len = -1);
	char *ClipText();
	bool Open(char *Name);
	bool UpdateHiddenCheck();
	void UpdateScrollBars();
	void AfterInsertText(char *Text, int Len);
	bool DoFind();
	void SelectWord(VMouse &m);

	// Object Events
	bool OnMoveCursor(int Dx, int Dy, bool NoSelect = FALSE);
	void OnSetCursor(int X, int Y);
	bool OnFind();
	bool OnReplace();
	bool OnMultiLineTab(bool In);
	void OnSetHidden(int Hidden);
	virtual void OnUrl(char *Url);

	// Window Events
	void OnFocus(bool f);
	void OnMouseClick(VMouse &m);
	void OnMouseMove(VMouse &m);
	void OnKey(VKey &k);
	void OnPaint(GdcBasePrimitives *pDC);
	int OnEvent(GMessage *Msg);
	void OnPulse();
};


#endif