/*hdr
**      FILE:           GuiCommon.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           18/7/98
**      DESCRIPTION:    Common windows and controls
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"

#include "gdc2.h"
#include "gui.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
GFileSelect::GFileSelect()
{
	Size = 4096;
	FileStr = NEW(char[Size]);
	if (FileStr)
	{
		*FileStr = 0;
	}
	else
	{
		Size = 0;
	}

	ZeroObj(Info);
	Info.lStructSize = sizeof(Info);
	Info.hInstance = hAppInst();
}

GFileSelect::~GFileSelect()
{
	DeleteArray(FileStr);

	for (GFileType *Type = TypeList.First(); Type; Type = TypeList.Next())
	{
		DeleteObj(Type);
	}
}

char *GFileSelect::TypeStr()
{
	char *Str = NEW(char[4096]);
	if (Str)
	{
		char *pstr = Str;

		for (GFileType *Type = TypeList.First(); Type; Type = TypeList.Next())
		{
			strcpy(pstr, Type->Description());
			pstr += strlen(Type->Description()) + 1;
			strcpy(pstr, Type->Extension());
			pstr += strlen(Type->Extension()) + 1;
		}

		*pstr = 0;
		Info.lpstrFilter = Str;
	}

	return Str;
}

char *GFileSelect::Name()
{
	return FileStr;
}

bool GFileSelect::Name(char *n)
{
	bool Status = FALSE;

	if (n AND FileStr)
	{
		strncpy(FileStr, n, Size);
		Info.lpstrFile = FileStr;
		Info.nMaxFile = Size;
	}

	return Status;
}

char *GFileSelect::operator [](int i)
{
	return NULL;
}

int GFileSelect::Names()
{
	return 1;
}

char *GFileSelect::File()
{
	return FileStr + Info.nFileOffset;
}

char *GFileSelect::Extension()
{
	return FileStr + Info.nFileExtension;
}

int GFileSelect::Types()
{
	return TypeList.GetItems();
}

GFileType *GFileSelect::Type(int n)
{
	return (GFileType*) TypeList[n];
}

bool GFileSelect::Type(char *Description, char *Extension)
{
	GFileType *Type = NEW(GFileType);
	if (Type)
	{
		Type->Description(Description);
		Type->Extension(Extension);
		TypeList.Insert(Type);
	}

	return Type != 0;
}

GWindow *GFileSelect::Parent()
{
	return ParentWnd;
}

void GFileSelect::Parent(GWindow *Window)
{
	ParentWnd = Window;
	if (ParentWnd)
	{
		Info.hwndOwner = Window->Handle();
	}
}

bool GFileSelect::MultiSelect()
{
	return (Info.Flags & OFN_ALLOWMULTISELECT) != 0;
}

void GFileSelect::MultiSelect(bool Multi)
{
	if (Multi)
	{
		Info.Flags |= OFN_ALLOWMULTISELECT;
	}
	else
	{
		Info.Flags &= ~OFN_ALLOWMULTISELECT;
	}
}

char *GFileSelect::InitialDir()
{
	return (char*) Info.lpstrInitialDir;
}

void GFileSelect::InitialDir(char *InitDir)
{
	Info.lpstrInitialDir = InitDir;
}

char *GFileSelect::Title()
{
	return (char*) Info.lpstrTitle;
}

void GFileSelect::Title(char *Title)
{
	Info.lpstrTitle = Title;
}

char *GFileSelect::DefaultExtension()
{
	return (char*) Info.lpstrDefExt;
}

void GFileSelect::DefaultExtension(char *DefExt)
{
	Info.lpstrDefExt = DefExt;
}

bool GFileSelect::Open()
{
	bool Status = FALSE;

	char *Str = TypeStr();
	if (Str)
	{
		Info.lpstrFilter = Str;		
		Info.lpstrFile = FileStr;
		Info.nMaxFile = Size;

		Status = GetOpenFileName(&Info);

		DeleteObj(Str);
	}

	return Status;
}

bool GFileSelect::Save()
{
	bool Status = FALSE;

	char *Str = TypeStr();
	if (Str)
	{
		Info.lpstrFilter = Str;		
		Info.lpstrFile = FileStr;
		Info.nMaxFile = Size;

		Status = GetSaveFileName(&Info);

		DeleteObj(Str);
	}

	return Status;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GFindReplace::GFindReplace(HWND hParent)
{
	FindOp = TRUE;
	FindMsgID = RegisterWindowMessage(FINDMSGSTRING);
	Find[0] = 0;
	Replace[0] = 0;
	hDlg = 0;

	ZeroObj(FindInfo);
	FindInfo.lStructSize = sizeof(FindInfo);
	FindInfo.hInstance = hAppInst();
	FindInfo.Flags = FR_HIDEUPDOWN;
	FindInfo.hwndOwner = hParent;
	FindInfo.lpstrFindWhat = Find;
	FindInfo.lpstrReplaceWith = Replace;
	FindInfo.wFindWhatLen = sizeof(Find);
	FindInfo.wReplaceWithLen = sizeof(Replace);
}

GFindReplace::~GFindReplace()
{
	if (hDlg)
	{
		DestroyWindow(hDlg);
		hDlg = 0;
	}
}

void GFindReplace::SetParent(HWND hParent)
{
	FindInfo.hwndOwner = hParent;
}

bool GFindReplace::OpenFind()
{
	FindOp = TRUE;
	hDlg = FindText(&FindInfo);
	if (hDlg)
	{
		MSG Msg;
		while (GetMessage(&Msg, NULL, 0, 0) > 0 AND IsWindow(hDlg))
		{
			if (NOT hDlg OR NOT IsDialogMessage(hDlg, &Msg))
			{
				TranslateMessage(&Msg);
				DispatchMessage(&Msg);
			}
		}

		return TRUE;
	}

	return FALSE;
}

bool GFindReplace::OpenReplace()
{
	FindOp = FALSE;
	hDlg = ReplaceText(&FindInfo);
	if (hDlg)
	{
		MSG Msg;
		while (GetMessage(&Msg, NULL, 0, 0) > 0 AND IsWindow(hDlg))
		{
			if (NOT hDlg OR NOT IsDialogMessage(hDlg, &Msg))
			{
				TranslateMessage(&Msg);
				DispatchMessage(&Msg);
			}
		}

		return TRUE;
	}

	return FALSE;
}

int GFindReplace::OnEvent(GMessage *Msg)
{
	int Status = FALSE;

	if (Msg->Msg == FindMsgID)
	{
		if (FindOp)
		{
			Status = OnFind();
		}
		else
		{
			Status = OnReplace();
		}
	}

	return Status;
}

bool GFindReplace::OnFind()
{
	DestroyWindow(hDlg);
	hDlg = 0;
	return TRUE;
}

bool GFindReplace::OnReplace()
{
	DestroyWindow(hDlg);
	hDlg = 0;
	return TRUE;
}

bool GFindReplace::CaseSensitive()
{
	return (FindInfo.Flags & FR_MATCHCASE) != 0;
}

void GFindReplace::CaseSensitive(bool i)
{
	if (i)
	{
		FindInfo.Flags |= FR_MATCHCASE;
	}
	else
	{
		FindInfo.Flags &= ~FR_MATCHCASE;
	}
}

bool GFindReplace::WholeWord()
{
	return (FindInfo.Flags & FR_WHOLEWORD) != 0;
}

void GFindReplace::WholeWord(bool i)
{
	if (i)
	{
		FindInfo.Flags |= FR_WHOLEWORD;
	}
	else
	{
		FindInfo.Flags &= ~FR_WHOLEWORD;
	}
}

bool GFindReplace::BtnFindNext()
{
	return (FindInfo.Flags & FR_FINDNEXT) != 0;
}

bool GFindReplace::BtnReplace()
{
	return (FindInfo.Flags & FR_REPLACE) != 0;
}

bool GFindReplace::BtnReplaceAll()
{
	return (FindInfo.Flags & FR_REPLACEALL) != 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GClipBoard::GClipBoard(GWindow *o)
{
	Owner = o;
	if (Owner)
	{
		Open = OpenClipboard(Owner->Handle());
	}
	else
	{
		Open = FALSE;
	}
}

GClipBoard::~GClipBoard()
{
	if (Open)
	{
		CloseClipboard();
		Open = FALSE;
	}
}

bool GClipBoard::Empty()
{
	return EmptyClipboard();
}

bool GClipBoard::Text(char *Str)
{
	bool Status = FALSE;

	if (Str)
	{
		HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, strlen(Str)+1);
		if (hMem)
		{
			char *Data = (char*) GlobalLock(hMem);
			if (Data)
			{
				strcpy(Data, Str);
				GlobalUnlock(hMem);

				Status = SetClipboardData(CF_TEXT, hMem) != 0;
			}
			else
			{
				GlobalFree(hMem);
			}
		}
	}

	return Status;
}

char *GClipBoard::Text()
{
	char *Str = NULL;

	HGLOBAL hMem = GetClipboardData(CF_TEXT);
	if (hMem)
	{
		char *Data = (char*) GlobalLock(hMem);
		if (Data)
		{
			Str = NewStr(Data);
			GlobalUnlock(hMem);
		}
	}

	return Str;
}

bool GClipBoard::Bitmap(GdcBasePrimitives *pDC)
{
	bool Status = FALSE;
	if (pDC)
	{
		int Bytes = BMPWIDTH(pDC->X() * pDC->GetBits());
		int Colours = (pDC->GetBits()>8) ? ((pDC->GetBits() != 24) ? 3 : 0) : 1 << pDC->GetBits();
		int HeaderSize = sizeof(BITMAPINFOHEADER);
		int PaletteSize = Colours * sizeof(RGBQUAD);
		int MapSize = Bytes * pDC->Y();

		HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, HeaderSize + PaletteSize + MapSize);
		if (hMem)
		{
			BITMAPINFO *Info = (BITMAPINFO*) GlobalLock(hMem);
			if (Info)
			{
				// Header
				Info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
				Info->bmiHeader.biWidth = pDC->X();
				Info->bmiHeader.biHeight = pDC->Y();
				Info->bmiHeader.biPlanes = 1;
				Info->bmiHeader.biBitCount = pDC->GetBits();
				
				if (pDC->GetBits() == 16 OR
					pDC->GetBits() == 32)
				{
					Info->bmiHeader.biCompression = BI_BITFIELDS;
				}
				else
				{
					Info->bmiHeader.biCompression = BI_RGB;
				}

				Info->bmiHeader.biSizeImage = 0;
				Info->bmiHeader.biXPelsPerMeter = 3000;
				Info->bmiHeader.biYPelsPerMeter = 3000;
				Info->bmiHeader.biClrUsed = 0;
				Info->bmiHeader.biClrImportant = 0;

				if (pDC->GetBits() <= 8)
				{
					// Palette
					GdcPalette *Pal = pDC->Palette();
					RGBQUAD *Rgb = Info->bmiColors;
					if (Pal)
					{
						GdcRGB *p = (*Pal)[0];
						if (p)
						{
							for (int i=0; i<Colours; i++, p++, Rgb++)
							{
								Rgb->rgbRed = p->R;
								Rgb->rgbGreen = p->G;
								Rgb->rgbBlue = p->B;
								Rgb->rgbReserved = p->Flags;
							}
						}
					}
					else
					{
						memset(Rgb, 0, Colours * sizeof(RGBQUAD));
					}
				}
				else
				{
					int *Primaries = (int*) Info->bmiColors;
					
					switch (pDC->GetBits())
					{
						case 16:
						{
							Primaries[0] = Rgb16(255, 0, 0);
							Primaries[1] = Rgb16(0, 255, 0);
							Primaries[2] = Rgb16(0, 0, 255);
							break;
						}
						case 32:
						{
							Primaries[0] = Rgb32(255, 0, 0);
							Primaries[1] = Rgb32(0, 255, 0);
							Primaries[2] = Rgb32(0, 0, 255);
							break;
						}
					}
				}

				// Bits
				uchar *Dest = ((uchar*)Info) + HeaderSize + PaletteSize;
				for (int y=pDC->Y()-1; y>=0; y--)
				{
					uchar *s = (*pDC)[y];
					if (s)
					{
						memcpy(Dest, s, Bytes);
					}
					Dest += Bytes;
				}

				Status = SetClipboardData(CF_DIB, hMem) != 0;
			}
			else
			{
				GlobalFree(hMem);
			}
		}
	}
	return Status;
}

GdcBasePrimitives *GClipBoard::ConvertFromPtr(void *Ptr)
{
	GdcBasePrimitives *pDC = NULL;
	if (Ptr)
	{
		BITMAPINFO *Info = (BITMAPINFO*) Ptr;
		if (Info)
		{
			pDC = NEW(GdcDibSection);
			if (	pDC AND
				(Info->bmiHeader.biCompression == BI_RGB OR
				Info->bmiHeader.biCompression == BI_BITFIELDS))
			{
				if (pDC->Create(Info->bmiHeader.biWidth,
						Info->bmiHeader.biHeight,
						max(Info->bmiHeader.biPlanes * Info->bmiHeader.biBitCount, 8)))
				{
					int Colours = 0;
					char *Source = (char*) Info->bmiColors;

					// do palette
					if (pDC->GetBits() <= 8)
					{
						if (Info->bmiHeader.biClrUsed > 0)
						{
							Colours = Info->bmiHeader.biClrUsed;
						}
						else
						{
							Colours = 1 << pDC->GetBits();
						}

						GdcPalette *Pal = NEW(GdcPalette(NULL, Colours));
						if (Pal)
						{
							GdcRGB *d = (*Pal)[0];
							RGBQUAD *s = (RGBQUAD*) Source;
							if (d)
							{
								for (int i=0; i<Colours; i++, d++, s++)
								{
									d->R = s->rgbRed;
									d->G = s->rgbGreen;
									d->B = s->rgbBlue;
									d->Flags = s->rgbReserved;
								}
							}
							Source = (char*) s;
							pDC->Palette(Pal);
						}
					}

					if (Info->bmiHeader.biCompression == BI_BITFIELDS)
					{
						Source += sizeof(DWORD) * 3;
					}

					// do pixels
					int Bytes = BMPWIDTH(pDC->X() * pDC->GetBits());
					for (int y=pDC->Y()-1; y>=0; y--)
					{
						uchar *d = (*pDC)[y];
						if (d)
						{
							memcpy(d, Source, Bytes);
						}
						Source += Bytes;
					}
				}
			}
		}
	}
	return pDC;
}

GdcBasePrimitives *GClipBoard::Bitmap()
{
	HGLOBAL hMem = GetClipboardData(CF_DIB);
	void *Ptr = GlobalLock(hMem);
	GdcBasePrimitives *pDC = 0;
	if (Ptr)
	{
		pDC = ConvertFromPtr(Ptr);
		GlobalUnlock(hMem);
	}
	return pDC;
}

///////////////////////////////////////////////////////////////////////////////////////
// String resource stuff

char *LgiLoadString(WORD Res)
{
	static char Str[256];

	Str[0] = 0;
	HRSRC hRes = FindResourceEx(NULL, RT_STRING, MAKEINTRESOURCE(Res), LgiGetLanguageId());
	if (hRes)
	{
			HGLOBAL hMem = LoadResource(hAppInst(), hRes);
			if (hMem)
			{
				char *Ptr = (char*) LockResource(hMem);
				if (Ptr)
				{
					char *d = Str;
					short *s = (short*) (Ptr + 2);
					int Len = *s++;

					while (Len--)
					{
						*d++ = *s++;
					}
					*d++ = 0;
				}
			}
	}

	if (strlen(Str) < 1)
	{
		// can't find the language dependent string
		LoadString(hAppInst(), Res, Str, sizeof(Str));
	}

	return Str;
}

