/*hdr
**      FILE:           GuiDlg.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Dialog components
**
**      Copyright (C) 1998 Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "gui.h"

/////////////////////////////////////////////////////////////////////////////////////////
// Classes
short *GDlgFunc::StrCopy(short *A, char *N)
{
	uchar *n = (uchar*) N;
	if (n)
	{
		while (*n)
		{
			*A++ = *n++;
		}
	}
	*A++ = 0;
	return A;
}

short *GDlgFunc::Push(short *A, int n)
{
	int *Ptr = (int*) A;
	*Ptr++ = n;
	return (short*)Ptr;
}

short *GDlgFunc::PadToDWord(short *A, bool Seek)
{
	char *c = (char*) A;
	while ((int) c & 3)
	{
		if (Seek)
		{
			c++;
		}
		else
		{
			*c++ = 0;
		}
	}
	return (short*) c;
}

short *GDlgFunc::StrSkip(short *Ptr)
{
	if (*Ptr == -1)
	{
		Ptr++;
		Ptr++; // skip global atom
	}
	else
	{
		while (*Ptr)
		{
			Ptr++;
		}
		Ptr++;
	}
	return Ptr;
}

short *GDlgFunc::StrExtract(short *Ptr, char *Str, short *Atom)
{
	if (*Ptr == -1)
	{
		Ptr++;
		if (Atom)
		{
			*Atom = *Ptr++;
		}
		else
		{
			Ptr++; // skip global atom
		}
	}
	else
	{
		while (*Ptr)
		{
			*Str++ = (char) *Ptr;
			Ptr++;
		}
		Ptr++;
	}
	*Str++ = 0;
	return Ptr;
}

void GDlgFunc::PixelToDialog(GdcRegion &r)
{
	r.x1 = (double) r.x1 / DIALOG_X;
	r.y1 = (double) r.y1 / DIALOG_Y;
	r.x2 = (double) r.x2 / DIALOG_X;
	r.y2 = (double) r.y2 / DIALOG_Y;
}

void GDlgFunc::DialogToPixel(GdcRegion &r)
{
	r.x1 = (double) r.x1 * DIALOG_X;
	r.y1 = (double) r.y1 * DIALOG_Y;
	r.x2 = (double) r.x2 * DIALOG_X;
	r.y2 = (double) r.y2 * DIALOG_Y;
}

bool GDlgFunc::LoadFromResource(char *Resource, List<GWindow> &Children, GdcRegion *Pos, char *Name)
{
	bool Status = FALSE;
	HRSRC hRsrc = FindResourceEx(NULL, RT_DIALOG, Resource, LgiGetLanguageId());

	if (NOT hRsrc)
	{
		hRsrc = FindResource(NULL, Resource, RT_DIALOG);
	}

	if (hRsrc)
	{
		int Size = SizeofResource(NULL, hRsrc);
		HGLOBAL hMem = LoadResource(NULL, hRsrc);
		if (hMem)
		{
			uchar *Ptr = (uchar*) LockResource(hMem);
			if (Ptr)
			{
				DLGTEMPLATE *Template = (DLGTEMPLATE*) Ptr;

				if (Pos)
				{
					Pos->ZOff((double) Template->cx * DIALOG_X, (double) Template->cy * DIALOG_Y);
					Pos->Offset((double) Template->x * DIALOG_X, (double) Template->y * DIALOG_Y);
				}

				char Temp[256];
				short *A = (short*) (Template+1);

				// menu
				A = StrExtract(A, Temp);
				// class
				A = StrExtract(A, Temp);
				// title
				A = StrExtract(A, Temp);
				if (strlen(Temp) > 0 AND Name)
				{
					strcpy(Name, Temp);
				}

				if (Template->style & DS_SETFONT)
				{
					// font
					int PointSize = *A++;
					// name
					A = StrExtract(A, Temp);
				}

				A = PadToDWord(A, TRUE);

				// loop thru controls
				for (int i=0; i<Template->cdit; i++)
				{
					GControlInfo *CtrlInfo = NEW(GControlInfo);
					if (CtrlInfo)
					{
						DLGITEMTEMPLATE *Item = (DLGITEMTEMPLATE*) A;
						memcpy(&CtrlInfo->Item, Item, sizeof(DLGITEMTEMPLATE));
						A = (short*) (Item+1);

						// class
						A = StrExtract(A, Temp, &CtrlInfo->Atom);
						if (strlen(Temp) > 0)
						{
							CtrlInfo->Class = NewStr(Temp);
						}

						// title
						A = StrExtract(A, Temp);
						if (strlen(Temp) > 0)
						{
							CtrlInfo->Text = NewStr(Temp);
						}

						// creation data
						int Size = *A++;
						if (Size > 0)
						{
							A = PadToDWord(A, TRUE);
							A = (short*) (((uchar*)A)+Size);
						}
						A = PadToDWord(A, TRUE);

						GControl *c = CreateControl(CtrlInfo);
						if (c)
						{
							Children.Insert(c);
						}

						DeleteObj(CtrlInfo);
					}
				}

				Status = TRUE;
				GlobalUnlock(hMem);
			}
		}
	}
	return Status;
}

GControl *GDlgFunc::CreateControl(GControlInfo *Info)
{
	GControl *Ctrl = 0;

	Info->Item.x	= (double) Info->Item.x * CTRL_X;
	Info->Item.cx	= (double) Info->Item.cx * CTRL_X;
	Info->Item.y	= (double) Info->Item.y * CTRL_Y;
	Info->Item.cy	= (double) Info->Item.cy * CTRL_Y;

	switch (Info->Atom)
	{
		case 0x0080: // Button
		{
			int Type = Info->Item.style & 0xF;
			static GRadioGroup *LastGroup = NULL;

			if (Type == BS_CHECKBOX OR Type == BS_AUTOCHECKBOX)
			{
				Ctrl = NEW(GCheckBox(		Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			}
			else if (Type == BS_GROUPBOX)
			{
				Ctrl = LastGroup = NEW(GRadioGroup(
								Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			}
			else if (Type == BS_RADIOBUTTON OR Type == BS_AUTORADIOBUTTON)
			{
				if (LastGroup)
				{	
					GdcRegion p = LastGroup->GetPos();
					LastGroup->Append(	Info->Item.x-p.x1,
								Info->Item.y-p.y1,
								Info->Text);
				}
			}
			else
			{
				Ctrl = NEW(GButton(		Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			}
			break;
		}
		case 0x0081: // Edit
		{
			Ctrl = NEW(GEdit(			Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			if (Ctrl)
			{
				GEdit *Ed = (GEdit*) Ctrl;

				if (Info->Item.style & ES_PASSWORD)
				{
					Ed->Password(TRUE);
				}

				if (Info->Item.style & ES_MULTILINE)
				{
					Ed->MultiLine(TRUE);
				}

				if (Info->Item.style & ES_READONLY)
				{
					Ed->Enabled(FALSE);
				}
			}
			break;
		}
		case 0x0082: // Static
		{
			Ctrl = NEW(GText(			Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			break;
		}
		case 0x0083: // List box
		case 0x0084: // Scroll bar
		{
			// not implemented
			// use list view and slider instead
			break;
		}
		case 0x0085: // Combo box
		{
			Ctrl = NEW(GCombo(			Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
			break;
		}
		default:
		{
			if (Info->Class)
			{
				if (NOT strcmp(Info->Class, TRACKBAR_CLASS))
				{
					Ctrl = NEW(GSlider(	Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text,
								(Info->Item.style & TBS_VERT) ? TRUE : FALSE));
				}
				else if (NOT strcmp(Info->Class, WC_TABCONTROL))
				{
					Ctrl = NEW(GTabControl(	Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
				}
				else if (NOT strcmp(Info->Class, WC_LISTVIEW))
				{
					Ctrl = NEW(GList(	Info->Item.id,
								Info->Item.x,
								Info->Item.y,
								Info->Item.cx,
								Info->Item.cy,
								Info->Text));
				}
			}
			break;
		}
	}

	return Ctrl;
}

///////////////////////////////////////////////////////////////////////////////////////////
GControlInfo::GControlInfo()
{
	ZeroObj(Item);
	Atom = 0;
	Class = 0;
	Text = 0;
	Ctrl = 0;
}

GControlInfo::~GControlInfo()
{
	DeleteArray(Class);
	DeleteArray(Text);
	Ctrl = 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
int GControlGroup::OnEvent(GMessage *Msg)
{
	int Status = 0;
	switch (Msg->Msg)
	{
		case WM_HSCROLL:
		case WM_VSCROLL:
		{
			GControl *Ctrl = GetControl((HWND) Msg->b);
			if (Ctrl)
			{
				OnNotify(Ctrl, 0);
			}
			break;
		}
		case WM_CHANGE:
		{
			GControl *Ctrl = (GControl*) Msg->a;
			if (Ctrl)
			{
				OnNotify(Ctrl, Msg->b);
			}
			break;
		}
		case WM_NOTIFY:
		{
			LPNMHDR Hdr = (LPNMHDR) Msg->b;
			GControl *Ctrl = GetControl(Hdr->hwndFrom);

			if (Ctrl) 
			{
				uint Foo = TVN_SELCHANGED;

				switch (Hdr->code)
				{
					// General notification
					case NM_DBLCLK:
					case NM_CLICK:
					{
						// I'm not even going to try processing these here
						// could be different for every control
						// so I'll pass it straight back to the control
						// maybe it'll know what to do with it
						// SendMessage(Ctrl->Handle(), Msg, a, b);
						break;
					}

					// List View Stuff
					case LVN_GETDISPINFO:
					{
						LV_DISPINFO *Info = (LV_DISPINFO*) Msg->b;
						GListItem *Item = (GListItem*) Info->item.lParam;

						if (Item)
						{
							if (Info->item.mask & LVIF_TEXT)
							{
								if (Info->item.pszText)
								{
									strcpy(Info->item.pszText, Item->GetText(Info->item.iSubItem));
								}
								else
								{
									// what the?
									// don't know what causes windows to call using a NULL
									// pointer. Any ideas ppl?
									Info->item.pszText = Item->GetText(Info->item.iSubItem);
								}
							}

							if (Info->item.mask & LVIF_IMAGE)
							{
								Info->item.iImage = Item->GetImage();
							}
						}
						break;
					}
					case LVN_ENDLABELEDIT:
					{
						LV_DISPINFO *Info = (LV_DISPINFO*) Msg->b;
						GListItem *Item = (GListItem*) Info->item.lParam;
						if (Info AND Item)
						{
							if (Info->item.mask & LVIF_TEXT)
							{
								return Item->SetText(Info->item.pszText);
							}
						}
						break;
					}
					case LVN_BEGINDRAG:
					case LVN_BEGINRDRAG:
					{
						NM_LISTVIEW *Info = (NM_LISTVIEW*) Msg->b;
						LV_ITEM Item;

						ZeroObj(Item);
						Item.mask = LVIF_PARAM;
						Item.iItem = Info->iItem;
						if (ListView_GetItem(Ctrl->Handle(), &Item))
						{
							((GList*) Ctrl)->OnItemBeginDrag((GListItem*) Item.lParam, (Hdr->code == LVN_BEGINDRAG) ? MK_LBUTTON : MK_RBUTTON);
						}
						break;
					}
					case LVN_COLUMNCLICK:
					{
						NM_LISTVIEW *Info = (NM_LISTVIEW*) Msg->b;
						((GList*) Ctrl)->OnColumnClick(Info->iSubItem);
						break;
					}

					// Tab Control Stuff
					case TCN_SELCHANGING:
					{
						Status = Ctrl->SysOnNotify(Hdr->code);
						break;
					}
					case TCN_SELCHANGE:
					{
						Ctrl->SysOnNotify(Hdr->code);
						OnNotify(Ctrl, 0);
						break;
					}

					// Tree View Control
	 				case TVN_BEGINRDRAG:
					case TVN_BEGINDRAG:
					{
						NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
						((GTree*) Ctrl)->OnItemBeginDrag((GTreeItem*) Item->itemNew.lParam, (Hdr->code == TVN_BEGINDRAG) ? MK_LBUTTON : MK_RBUTTON);
						break;
					}
					case TVN_SELCHANGED:
					{
						NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
						((GTree*) Ctrl)->OnItemSelect((GTreeItem*) Item->itemNew.lParam);
						break;
					}
					case TVN_ITEMEXPANDED:
					{
						NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
						((GTree*) Ctrl)->OnItemExpand((GTreeItem*) Item->itemNew.lParam, Item->action);
						break;
					}
					case TVN_GETDISPINFO:
					{
						TV_DISPINFO *Info = (TV_DISPINFO*) Msg->b;
						GTreeItem *Item = (GTreeItem*) Info->item.lParam;
						if (Item)
						{
							if (Info->item.mask & TVIF_TEXT)
							{
								if (Info->item.pszText)
								{
									strcpy(Info->item.pszText, Item->GetText());
								}
								else
								{
									Info->item.pszText = Item->GetText();
								}								
							}

							if (Info->item.mask & TVIF_IMAGE)
							{
								Info->item.iImage = Item->GetImage();
							}
							
							if (Info->item.mask & TVIF_SELECTEDIMAGE)
							{
								Info->item.iSelectedImage = Item->GetImage(1);
							}							
						}
						break;
					}
				}
			}
			break;
		}
		case WM_COMMAND:
		{
			GControl *Ctrl = GetControl((HWND) Msg->b);
			if (Ctrl)
			{
				short Code = HIWORD(Msg->a);
				switch (Code)
				{
					case BN_CLICKED: // BUTTON
					{
						OnNotify(Ctrl, 0);
						break;
					}
					case EN_CHANGE: // EDIT
					{
						Ctrl->SysOnNotify(Code);
						OnNotify(Ctrl, 0);
						break;
					}
					case CBN_CLOSEUP:
					{
						PostMessage(hWnd, WM_COMMAND, MAKELONG(Ctrl->GetId(), CBN_EDITCHANGE), Msg->b);
						break;
					}
					case CBN_EDITCHANGE: // COMBO
					{
						Ctrl->SysOnNotify(Code);
						OnNotify(Ctrl, 0);
						break;
					}
				}
			}
			break;
		}
	}
	return Status;
}

GControl *GControlGroup::FindControl(int Id)
{
	Iterator<GControl> List(&Children);
	for (GControl *c = List.First(); c; c = List.Next())
	{
		GControl *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return NULL;
}

GControl *GControlGroup::GetControl(HWND hCtrl)
{
	GControl *c = 0;
	if (hCtrl)
	{
		Iterator<GControl> List(&Children);

		for (GControl *Ctrl = List.First(); Ctrl AND NOT c; Ctrl = List.Next())
		{
			c = Ctrl->GetControl(hCtrl);
		}
	}

	return c;
}

///////////////////////////////////////////////////////////////////////////////////////////
GDialog::GDialog()
{
	Name("Dialog");
	IsResizable = FALSE;
	Mem = 0;
}

GDialog::~GDialog()
{
	DeleteObj(Mem);
}

int GDialog::SizeOfControls()
{
	int Size = 0;
	GControl *Ctrl = (GControl*) Children.First();
	while (Ctrl)
	{
		Size += Ctrl->SizeOf();
		Ctrl = (GControl*) Children.Next();
	}
	return Size;
}

LRESULT CALLBACK DlgRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	/*
	switch (m)
	{
		case WM_NOTIFY:
		{
			LPNMHDR Hdr = (LPNMHDR) b;
			if (Hdr->code == LVN_GETDISPINFO)
			{
				int i=0;
			}
			break;
		}
	}
	*/

	if (m == WM_INITDIALOG)
	{
		GDialog *NewWnd = (GDialog*) b;
		NewWnd->hWnd = hWnd;
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
		GApp::ActiveDialog = hWnd;
	}

	GDialog *Wnd = (GDialog*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		return Wnd->OnEvent(&Msg);
	}

	return 0;
}

bool GDialog::LoadFromResource(char *Resource)
{
	char n[256];
	bool Status = GDlgFunc::LoadFromResource(Resource, Children, &Pos, n);
	if (Status)
	{
		Name(n);
	}
	return Status;
}
void GDialog::MoveOnScreen()
{
	GdcRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);

	if (Pos.x2 >= Screen.X())
		Pos.Offset(Screen.X() - Pos.x2, 0);
	if (Pos.y2 >= Screen.Y())
		Pos.Offset(0, Screen.Y() - Pos.y2);
	if (Pos.x1 < 0)
		Pos.Offset(-Pos.x1, 0);
	if (Pos.y1 < 0)
		Pos.Offset(0, -Pos.y1);
}

void GDialog::MoveToCenter()
{
	GdcRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	Pos.Offset(-Pos.x1, -Pos.y1);
	Pos.Offset((Screen.X() - Pos.X()) / 2, (Screen.Y() - Pos.Y()) / 2);
}

void GDialog::MoveToMouse()
{
	POINT Pt;
	if (GetCursorPos(&Pt))
	{
		Pos.Offset(-Pos.x1, -Pos.y1);
		Pos.Offset(Pt.x-(Pos.X()/2), Pt.y-(Pos.Y()/2));
	}
}

int GDialog::DoModal()
{
	int Status = -1;
	int Size = 16<<10;
	Size += SizeOfControls();
	Mem = NEW(GMem(Size));
	if (Mem)
	{
		DLGTEMPLATE *Template = (DLGTEMPLATE*) Mem->Lock();
		if (Template)
		{
			GdcRegion r = Pos;
			PixelToDialog(r);

			Template->style =	DS_ABSALIGN |
						WS_SYSMENU |
						WS_CAPTION |
						DS_SETFONT |
						DS_NOFAILCREATE |
						DS_3DLOOK;
			if (IsResizable)
			{
				Template->style |= WS_THICKFRAME;
			}
			else
			{
				Template->style |= DS_MODALFRAME;
			}

			Template->dwExtendedStyle = WS_EX_DLGMODALFRAME;
			Template->cdit = 0;
			Template->x = r.x1;
			Template->y = r.y1;
			Template->cx = r.X();
			Template->cy = r.Y();

			short *A = (short*) (Template+1);
			// menu
			*A++ = 0;
			// class
			*A++ = 0;
			// title
			A = StrCopy(A, Name());
			// font
			*A++ = 8; // point size
			A = StrCopy(A, "MS Sans Serif");
			A = PadToDWord(A);

			GWindow *p = (Parent) ? ((Parent->Handle()) ? Parent : Parent->FindReal()) : 0;
			Status = DialogBoxIndirectParam(hAppInst(),
							Template,
							(p)?p->Handle():0,
							(DLGPROC) DlgRedir, 
							(LPARAM) this);
		}

		DeleteObj(Mem);
	}

	return Status;
}

int GDialog::DoModeless()
{
	int Status = -1;
	int Size = 16<<10;
	Size += SizeOfControls();
	Mem = NEW(GMem(Size));
	if (Mem)
	{
		DLGTEMPLATE *Template = (DLGTEMPLATE*) Mem->Lock();
		if (Template)
		{
			GdcRegion r = Pos;
			PixelToDialog(r);

			Template->style =	WS_VISIBLE |
						DS_ABSALIGN |
						WS_SYSMENU |
						WS_CAPTION |
						DS_SETFONT |
						DS_NOFAILCREATE |
						DS_3DLOOK;
			if (IsResizable)
			{
				Template->style |= WS_THICKFRAME;
			}
			else
			{
				Template->style |= DS_MODALFRAME;
			}

			Template->dwExtendedStyle = WS_EX_DLGMODALFRAME;
			Template->cdit = 0;
			Template->x = r.x1;
			Template->y = r.y1;
			Template->cx = r.X();
			Template->cy = r.Y();

			short *A = (short*) (Template+1);
			// menu
			*A++ = 0;
			// class
			*A++ = 0;
			// title
			A = StrCopy(A, Name());
			// font
			*A++ = 8; // point size
			A = StrCopy(A, "MS Sans Serif");
			A = PadToDWord(A);

			GWindow *p = (Parent) ? ((Parent->Handle()) ? Parent : Parent->FindReal()) : 0;
			CreateDialogIndirectParam(	hAppInst(),
							Template,
							(p)?p->Handle():0,
							(DLGPROC) DlgRedir, 
							(LPARAM) this);
		}
	}

	return Status;
}

int GDialog::OnEvent(GMessage *Msg)
{
	int Status = GControlGroup::OnEvent(Msg);

	switch (Msg->Msg)
	{
		case WM_INITDIALOG:
		{
			HFONT hFont = (HFONT) SendMessage(hWnd, WM_GETFONT, 0, 0);

			// attach all our controls
			Iterator<GControl> List(&Children);
			for (	GControl *Ctrl = List.First();
				Ctrl;
				Ctrl = List.Next())
			{	
				Ctrl->Attach(this);
			}

			Status = TRUE;
			OnCreate();
			break;
		}
		case WM_CLOSE:
		{
			EndDialog(hWnd, 0);
			break;
		}
		case WM_NCACTIVATE:
		{
			GApp::ActiveDialog = (Msg->a) ? hWnd : 0;
			break;
		}
	}

	return Status;
}

void GDialog::EndModal(int Code)
{
	EndDialog(hWnd, Code);
}

void GDialog::EndModeless(int Code)
{
	DestroyWindow(hWnd);
	DeleteObj(Mem);
}

///////////////////////////////////////////////////////////////////////////////////////////
GControl::GControl()
{
	Pos.ZOff(10, 10);
	SetId(0);
	Sunken(TRUE);
}

GControl::~GControl()
{
}

char *GControl::Name()
{
	return GObject::Name();
}

bool GControl::Name(char *n)
{
	if (hWnd) SetWindowText(hWnd, n);
	else GObject::Name(n);
	return TRUE;
}

bool GControl::Attach(GWindow *parent)
{
	Parent = parent;
	if (Parent)
	{
		int MoreStyles = 0;
		if (NOT (WndFlags & GWF_ENABLE))
		{
			MoreStyles |= WS_DISABLED;
		}
		if (WndFlags & GWF_VISIBLE)
		{
			MoreStyles |= WS_VISIBLE;
		}

		CreateWindowEx(	ExStyle() | ((IsDropTarget()) ? WS_EX_ACCEPTFILES : 0),
						Class(),
						Name(),
						Style() | MoreStyles,
						Pos.x1,
						Pos.y1,
						Pos.X(),
						Pos.Y(),
						Parent->Handle(),
						NULL,
						hAppInst(),
						this);

		if (hWnd)
		{
			SetWindowLong(hWnd, GWL_ID, GetId());
			SetFont(Parent->GetFont());
		}
	}
	return hWnd != 0;
}

int GControl::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
		case WM_SETTEXT:
		{
			GObject::Name((char*)Msg->b);
			break;
		}
	}
	return 0;
}

GdcPt2 GControl::SizeOfStr(char *Str)
{
	GdcPt2 Pt(0, 0);
	if (Str)
	{
		HDC hDC = CreateCompatibleDC(NULL);
		if (hDC)
		{
			HFONT hFnt = GetFont();

			if (NOT hFnt)
			{
				hFnt = SysFont->Handle();
			}

			if (hFnt)
			{
				hFnt = SelectObject(hDC, hFnt);
				SIZE Size;

				if (GetTextExtentPoint32(hDC, Str, strlen(Str), &Size))
				{
					Pt.x = Size.cx;
					Pt.y = Size.cy;
				}

				SelectObject(hDC, hFnt);
			}

			DeleteDC(hDC);
		}
	}

	return Pt;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Button
GWin32Class GButton::WndClass;

GButton::GButton(int id, int x, int y, int cx, int cy, char *name)
{
	Def = FALSE;
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);

	WndClass.SubClass(Class(), "BUTTON");
}

GButton::~GButton()
{
}

bool GButton::Default()
{
	return Def;
}

void GButton::Default(bool b)
{
	Def = b;
}

int GButton::OnEvent(GMessage *Msg)
{
	GControl::OnEvent(Msg);
	return WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
}

///////////////////////////////////////////////////////////////////////////////////////////
// Edit
GWin32Class GEdit::WndClass;

GEdit::GEdit(int id, int x, int y, int cx, int cy, char *name)
{
	GdcPt2 Size = SizeOfStr((name)?name:"A");
	if (cx < 0) cx = Size.x + 6;
	if (cy < 0) cy = Size.y + 4;

	MultiLineEdit = FALSE;
	PasswordEdit = FALSE;
	SetId(id);
	if (name) Name(name);
	Pos.Set(x, y, x+max(cx, 10), y+max(cy, 10));

	WndClass.SubClass(Class(), "EDIT");
}

GEdit::~GEdit()
{
}

int GEdit::SysOnNotify(int Code)
{
	int Len = SendMessage(hWnd, WM_GETTEXTLENGTH, 0, 0);
	char *Str = NEW(char[Len+1]);
	if (Str)
	{
		SendMessage(hWnd, WM_GETTEXT, (WPARAM) (Len+1), (LPARAM) Str);
		GObject::Name(Str);
		DeleteArray(Str);
	}
	return 0;
}

void GEdit::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	Name(Str);
}

int GEdit::Value()
{
	char *n = Name();
	return (n) ? atoi(n) : 0;
}

int GEdit::OnEvent(GMessage *Msg)
{
	GControl::OnEvent(Msg);
	switch (Msg->Msg)
	{
		case WM_KEYDOWN:
		case WM_KEYUP:
		case WM_CHAR:
		{
			GWindow::OnEvent(Msg);
			break;
		}
	}

	return WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
}

///////////////////////////////////////////////////////////////////////////////////////////
// Check box
GWin32Class GCheckBox::WndClass;

GCheckBox::GCheckBox(int id, int x, int y, int cx, int cy, char *name, int InitState)
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	Val = InitState;
	Three = FALSE;
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	WndClass.SubClass(Class(), "BUTTON");
}

GCheckBox::~GCheckBox()
{
}

int GCheckBox::Value()
{
	if (hWnd)
	{
		Val = SendMessage(hWnd, BM_GETCHECK, 0, 0);
	}
	return Val;
}

void GCheckBox::Value(int b)
{
	Val = b;
	if (hWnd)
	{
		SendMessage(hWnd, BM_SETCHECK, Val, 0);
	}
}

int GCheckBox::OnEvent(GMessage *Msg)
{
	if (Msg->Msg == WM_CREATE)
	{
		Value(Val);
	}
	GControl::OnEvent(Msg);
	return WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
}

//////////////////////////////////////////////////////////////////////////////////////////////////
// Text
GWin32Class GText::WndClass;

GText::GText(int id, int x, int y, int cx, int cy, char *name)
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x;
	if (cy < 0) cy = Size.y;

	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);

	WndClass.SubClass(Class(), "STATIC");
}

GText::~GText()
{
}

int GText::OnEvent(GMessage *Msg)
{
	GControl::OnEvent(Msg);
	return WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio group
GWin32Class GRadioGroup::WndClass;
int GRadioGroup::NextId = 20000;

GRadioGroup::GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init)
{
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	InitialValue = Init;

	WndClass.SubClass(Class(), "BUTTON");
}

GRadioGroup::~GRadioGroup()
{
	for (	GRadioButton *But = Buttons.First();
		But;
		But = Buttons.Next())
	{
		DeleteObj(But);
	}
}

bool GRadioGroup::Attach(GWindow *parent)
{
	bool Status = GControl::Attach(parent);
	if (Status)
	{
		for (	GRadioButton *But = Buttons.First();
			But;
			But = Buttons.Next())
		{
			But->Attach(parent);
		}
		Value(InitialValue);
	}
	return Status;
}

int GRadioGroup::Value()
{
	int i=0;
	for (	GRadioButton *But = Buttons.First();
		But;
		But = Buttons.Next(), i++)
	{
		if (But->Value()) return i;
	}
	return 0;
}

void GRadioGroup::Value(int i)
{
	GRadioButton *But = Buttons[i];
	if (But)
	{
		But->Value(TRUE);
	}
}

int GRadioGroup::OnEvent(GMessage *Msg)
{
	GControl::OnEvent(Msg);
	return WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
}

GRadioButton *GRadioGroup::Append(int x, int y, char *name)
{
	GRadioButton *But = NEW(GRadioButton(NextId++, Pos.x1+x, Pos.y1+y, -1, -1, name, Buttons.GetItems() == 0));
	if (But)
	{
		Buttons.Insert(But);
	}
	return But;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio button
GWin32Class GRadioButton::WndClass;

GRadioButton::GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first)
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	First = first;
	Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);

	WndClass.SubClass(Class(), "BUTTON");
}

GRadioButton::~GRadioButton()
{
}

int GRadioButton::Value()
{
	if (hWnd)
	{
		return SendMessage(hWnd, BM_GETCHECK, 0, 0);
	}
	return 0;
}

void GRadioButton::Value(int i)
{
	if (hWnd)
	{
		SendMessage(hWnd, BM_SETCHECK, i, 0);
	}
}

int GRadioButton::OnEvent(GMessage *Msg)
{
	GControl::OnEvent(Msg);
	return WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
}

////////////////////////////////////////////////////////////////////////////////////////////
// Tab Control
GWin32Class GTabControl::WndClass;

GTabControl::GTabControl(int id, int x, int y, int cx, int cy, char *name, int Init)
{
	SetId(id);
	Pos.ZOff(cx, cy);
	Pos.Offset(x, y);
	Name(name);
	InitialValue = Init;
	Current = 0;
	NextTab = 0;

	WndClass.SubClass(LGI_TABCONTROL, WC_TABCONTROL);
}

GTabControl::~GTabControl()
{
	for (	GTabPage *Page = Tabs.First();
		Page;
		Page = Tabs.Next())
	{
		DeleteObj(Page);
	}
}

GControl *GTabControl::FindControl(int Id)
{
	GControl *Ctrl = 0;

	if (GetId() == Id)
	{
		return this;
	}

	for (	GTabPage *Page = Tabs.First();
		Page AND NOT Ctrl;
		Page = Tabs.Next())
	{
		Iterator<GControl> Ctrls(&Page->Children);
		for (GControl *c = Ctrls.First(); c AND NOT Ctrl; c = Ctrls.Next())
		{
			Ctrl = c->FindControl(Id);
		}
	}
	return Ctrl;
}

GControl *GTabControl::GetControl(HWND hCtrl)
{
	GControl *Ctrl = 0;

	if (hCtrl == hWnd)
	{
		return this;
	}

	for (	GTabPage *Page = Tabs.First();
		Page AND NOT Ctrl;
		Page = Tabs.Next())
	{
		Ctrl = Page->GetControl(hCtrl);
	}
	return Ctrl;
}

GdcRegion GTabControl::GetClient()
{
	RECT Temp;
	GdcRegion Client;

	GetWindowRect(hWnd, &Temp);
	Client = Temp;
	Client.Offset(Pos.x1-Client.x1, Pos.y1-Client.y1);
	Temp = Client;
	SendMessage(hWnd, TCM_ADJUSTRECT, FALSE, (LPARAM) &Temp);
	Client = Temp;

	return Client;
}

int GTabControl::SysOnNotify(int Code)
{
	switch (Code)
	{
		case TCN_SELCHANGING:
		{
			// allow change?
			GTabPage *Cur = Tabs[NextTab];
			if (Cur) return NOT Cur->Enabled();
			return FALSE;
		}

		case TCN_SELCHANGE:
		{
			// change visible tab
			GTabPage *Cur = Tabs[Current];
			if (Cur) Cur->Visible(FALSE);
			Current = SendMessage(hWnd, TCM_GETCURSEL, 0, 0);
			Cur = Tabs[Current];
			if (Cur) Cur->Visible(TRUE);
			break;
		}
	}

	return 0;
}

bool GTabControl::Attach(GWindow *parent)
{
	bool Status = GControl::Attach(parent);

	if (hWnd)
	{
		int i=0;
		Iterator<GTabPage> TabList(&Tabs);
		for (	GTabPage *Page = TabList.First();
			Page;
			Page = TabList.Next(), i++)
		{
			Page->Attach(this, i);
			Page->Visible(i == Current);
		}

		SendMessage(hWnd, TCM_SETCURFOCUS, Current, 0);
	}

	return Status;
}

int GTabControl::Value()
{
	return Current;
}

void GTabControl::Value(int i)
{
	int CurItem = SendMessage(hWnd, TCM_GETCURFOCUS, 0, 0);
	i = limit(i, 0, Tabs.GetItems()-1);
	if (CurItem != i)
	{
		GTabPage *Cur = Tabs[Current];
		if (Cur) Cur->Visible(FALSE);

		Cur = Tabs[i];
		if (Cur AND NOT Cur->Enabled()) return;
		NextTab = Current = i;
		SendMessage(hWnd, TCM_SETCURFOCUS, Current, 0);

		if (Cur) Cur->Visible(TRUE);
	}
}

int GTabControl::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
		case WM_LBUTTONDOWN:
		{
			TC_HITTESTINFO Info;
			int Mx = LOWORD(Msg->b);
			int My = HIWORD(Msg->b);

			Info.pt.x = Mx;
			Info.pt.y = My;
			SendMessage(hWnd, TCM_HITTEST, 0, (LPARAM) &Info);
			if (Info.flags & TCHT_ONITEM)
			{
				RECT rc;
				for (int i=0; i<Tabs.GetItems(); i++)
				{
					SendMessage(hWnd, TCM_GETITEMRECT, i, (WPARAM) &rc);
					if (	Mx >= rc.left AND
							My >= rc.top AND
							Mx < rc.right AND
							My < rc.bottom)
					{
						NextTab = i;
						break;
					}
				}
			}
			break;
		}
		case WM_KEYDOWN:
		{
			break;
		}
	}

	return WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
}

GTabPage *GTabControl::Append(char *name)
{
	GTabPage *Page = NEW(GTabPage(name));
	if (Page)
	{
		Tabs.Insert(Page);
	}
	return Page;
}

//////////////////////////////////////////////////////////////////////////////////
GTabPage::GTabPage(char *name)
{
	Parent = 0;
	Index = 0;
	Pos.ZOff(0, 0);
	Active = TRUE;

	GObject::Name(name);
	Item.mask = TCIF_TEXT;
	Item.pszText = GObject::Name();
	Item.cchTextMax = strlen(Item.pszText);
	Item.iImage = 0;
	Item.lParam = 0;
}

GTabPage::~GTabPage()
{
}

GControl *GTabPage::GetControl(HWND hCtrl)
{
	GControl *c = NULL;
	Iterator<GControl> List(&Children);

	for (	GControl *w = List.First();
		w AND NOT c;
		w = List.Next())
	{
		c = w->GetControl(hCtrl);
	}

	return c;
}

bool GTabPage::Attach(GTabControl *parent, int index)
{
	bool Status = FALSE;
	Parent = parent;
	Index = index;

	if (Parent)
	{
		Status = SendMessage(	Parent->Handle(),
					TCM_INSERTITEM,
					(WPARAM) Index,
					(LPARAM) &Item) >= 0;

		if (Status)
		{
			if (parent)
			{
				Pos = parent->GetClient();
			}
			else
			{
				Pos.ZOff(10, 10);
			}

			GWindow *Dlg = Parent->GetParent();
			if (Dlg)
			{
				for (	GControl* w = (GControl*) Children.First();
					w;
					w = (GControl*) Children.Next())
				{
					w->Attach(Dlg);
					GdcRegion p = w->GetPos();
					p.Offset(Pos.x1, Pos.y1);
					w->SetPos(p);
				}
			}
		}
	}
	return Status;
}

char *GTabPage::Name()
{
	return GObject::Name();
}

bool GTabPage::Name(char *name)
{
	bool Status = GObject::Name(name);
	Item.pszText = GObject::Name();
	Item.cchTextMax = strlen(Item.pszText);
	if (Parent)
	{
		SendMessage(	Parent->Handle(),
				TCM_SETITEM,
				(WPARAM) Index,
				(LPARAM) &Item);
	}
	return Status;
}

int GTabPage::OnEvent(GMessage *Msg)
{
	return 0;
}

bool GTabPage::Enabled()
{
	return Active;
}

void GTabPage::Enabled(bool e)
{
	Active = (e != 0);
}

bool GTabPage::Visible()
{
	return TRUE;
}

void GTabPage::Visible(bool v)
{
	for (	GWindow *w = Children.First();
		w;
		w = Children.Next())
	{
		w->Visible(v);
	}
}

bool GTabPage::AppendDialogFromResource(char *Name)
{
	return GDlgFunc::LoadFromResource(Name, Children);
}


//////////////////////////////////////////////////////////////////////////////////
// Slider control
GWin32Class GSlider::WndClass;

GSlider::GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert)
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	Name(name);
	Vertical = vert;
	Min = Max = 0;
	Val = 0;

	WndClass.SubClass(Class(), TRACKBAR_CLASS);
}

GSlider::~GSlider()
{
}

int GSlider::SysOnNotify(int Code)
{
	return 0;
}

void GSlider::Value(int i)
{
	Val = i;
	if (hWnd)
	{
		SendMessage(hWnd, TBM_SETPOS, TRUE, Val);
	}
}

int GSlider::Value()
{
	if (hWnd)
	{
		Val = SendMessage(hWnd, TBM_GETPOS, 0, 0);
	}
	return Val;
}

void GSlider::GetLimits(int &min, int &max)
{
	if (hWnd)
	{
		min = Min = SendMessage(hWnd, TBM_GETRANGEMIN, 0, 0);
		max = Max = SendMessage(hWnd, TBM_GETRANGEMAX, 0, 0);
	}
	else
	{
		min = Min;
		max = Max;
	}
}

void GSlider::SetLimits(int min, int max)
{
	Min = min;
	Max = max;
	if (hWnd)
	{
		SendMessage(hWnd, TBM_SETRANGE, TRUE, (LPARAM) MAKELONG(min, max)); 
	}
}

int GSlider::OnEvent(GMessage *Msg)
{
	int Status = WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);

	switch (Msg->Msg)
	{
		case WM_CREATE:
		{
			SetLimits(Min, Max);
			Value(Val);
			break;
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////////
GWin32Class GCombo::WndClass;

GCombo::GCombo(int id, int x, int y, int cx, int cy, char *name)
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	Name(name);
	SortItems = FALSE;

	WndClass.SubClass(Class(), "COMBOBOX");
}

GCombo::~GCombo()
{
}

int GCombo::SysOnNotify(int Code)
{
	int Len = SendMessage(hWnd, WM_GETTEXTLENGTH, 0, 0);
	char *Str = NEW(char[Len+1]);
	if (Str)
	{
		SendMessage(hWnd, WM_GETTEXT, (WPARAM) (Len+1), (LPARAM) Str);
		GObject::Name(Str);
		DeleteArray(Str);
	}
	return 0;
}

void GCombo::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	Name(Str);
}

int GCombo::Value()
{
	return atoi(Name());
}

int GCombo::OnEvent(GMessage *Msg)
{
	int Status = WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);

	switch (Msg->Msg)
	{
		case WM_CREATE:
		{
			// add string to combo
			for (char *Str = First(); Str; Str = Next())
			{
				SendMessage(hWnd, CB_ADDSTRING, 0, (LPARAM) Str);
			}

			SendMessage(hWnd, WM_SETTEXT, 0, (LPARAM) Name());
			break;
		}
	}

	return Status;
}

bool GCombo::Delete()
{
	int Index = (Cur) ? FindIndex(Cur->Ptr) : -1;
	bool Status = DLinkList::Delete();

	if (Status AND Index >= 0 AND hWnd)
	{
		SendMessage(hWnd, CB_DELETESTRING, Index, 0);
	}

	return Status;
}

bool GCombo::Delete(int i)
{
	bool Status = DLinkList::Delete(i);

	if (Status AND hWnd)
	{
		SendMessage(hWnd, CB_DELETESTRING, i, 0);
	}

	return Status;
}

bool GCombo::Delete(char *p)
{
	int Index = FindIndex(p);
	bool Status = DLinkList::Delete((void*)p);

	if (Status AND Index >= 0 AND hWnd)
	{
		SendMessage(hWnd, CB_DELETESTRING, Index, 0);
	}

	return Status;
}

bool GCombo::Insert(char *p, int Index)
{
	bool Status = DLinkList::Insert((void*)p, Index);
	if (hWnd)
	{
		SendMessage(	hWnd,
				(Index >= 0) ? CB_INSERTSTRING : CB_ADDSTRING,
				(Index >= 0) ? Index : 0,
				(LPARAM) p);
 	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////
GWin32Class GBitmap::WndClass;

GBitmap::GBitmap(int id, int x, int y, char *FileName)
{
	SetId(id);
	Pos.ZOff(1, 1);
	Pos.Offset(x, y);
	pDC = LoadDC(FileName);
	if (pDC)
	{
		Pos.Dimension(pDC->X()+4, pDC->Y()+4);
	}

	WndClass.Create(Class());
}

GBitmap::~GBitmap()
{
	DeleteObj(pDC);
}

int GBitmap::OnEvent(GMessage *Msg)
{
	return GWindow::OnEvent(Msg);
}

void GBitmap::OnPaint(GdcBasePrimitives *pScreen)
{
	GdcRegion a(0, 0, Pos.X()-1, Pos.Y()-1);
	DrawMoveDeepEdge(pScreen, a, SUNKEN);
	if (pDC)
	{
		pScreen->Blt(2, 2, pDC);
	}
	else
	{
		pScreen->Colour(CBit(pScreen->GetBits(), GetSysColor(COLOR_WINDOW)));
		pScreen->Rectangle(&a);
	}
}

void GBitmap::OnMouseClick(VMouse &m)
{
	if (NOT m.Down AND Parent)
	{
		((GDialog*) Parent)->OnNotify(this, 0);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////
// List control

GListItem::GListItem()
{
	Parent = 0;
	Cols = 0;
	Str = 0;
}

GListItem::~GListItem()
{
	if (Str)
	{
		for (int i=0; i<Cols; i++)
		{
			DeleteArray(Str[i]);
		}
		DeleteArray(Str);
	}
}

void GListItem::Update()
{
	if (Parent)
	{
		Parent->SysUpdate(this);
	}
}

bool GListItem::GetItem(LV_ITEM *Item)
{
	if (Item AND Parent)
	{
		LV_FINDINFO Find;
		Find.flags = LVFI_PARAM;
		Find.lParam = (LPARAM) Item;
		int Index = ListView_FindItem(Parent->Handle(), -1, &Find);
		if (Index >= 0)
		{
			Item->iItem = Index;
			ListView_GetItem(Parent->Handle(), Item);
			return TRUE;
		}
	}
	return FALSE;
}

bool GListItem::Selected()
{
	if (Parent)
	{
		LV_ITEM Item;
		ZeroObj(Item);
		Item.mask = LVIF_STATE;
		Item.stateMask = -1;
		if (GetItem(&Item))
		{
			return (Item.state & LVIS_SELECTED) != 0;
		}
	}
	return FALSE;
}

void GListItem::Selected(bool b)
{
	if (Parent)
	{
		LV_ITEM Item;
		ZeroObj(Item);
		if (GetItem(&Item))
		{
			Item.mask = (b) ? LVIF_STATE : 0;
			Item.stateMask = LVIF_STATE;

			ListView_SetItem(Parent->Handle(), &Item);
		}
	}
}

bool GListItem::SetText(char *s, int i)
{
	if (i + 1 > Cols)
	{
		int NewCols = i+1;
		char **NewStr = NEW(char*[NewCols]);
		if (NewStr)
		{
			memset(NewStr, 0, sizeof(char*)*NewCols);

			for (int i=0; i<Cols; i++)
			{
				if (i < NewCols)
				{
					NewStr[i] = Str[i];
				}
				else
				{
					DeleteArray(Str[i]);
				}
			}

			DeleteArray(Str);
			Str = NewStr;
			Cols = NewCols;
		}
	}

	if (Str AND s AND i>=0 AND i<Cols)
	{
		DeleteArray(Str[i]);
		Str[i] = NewStr(s);
	}

	return TRUE;
}

char *GListItem::GetText(int i)
{
	if (Str AND i>=0 AND i<Cols)
	{
		return Str[i];
	}
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////

GWin32Class GList::WndClass;

GList::GList(int id, int x, int y, int cx, int cy, char *name)
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	if (name) Name(name);
	Val = -1;
	EditLabels = FALSE;
	ColumnHeader = TRUE;

	WndClass.SubClass(Class(), WC_LISTVIEW);
}

GList::~GList()
{
	Empty();
}

void GList::SysUpdate(GListItem *Item)
{
	LV_FINDINFO Find;
	Find.flags = LVFI_PARAM;
	Find.lParam = (LPARAM) Item;
	int Index = ListView_FindItem(hWnd, -1, &Find);
	if (Index >= 0)
	{
		ListView_Update(hWnd, Index);
	}
}

void GList::AddColumn(char *Name, int Width, int Align)
{
	LV_COLUMN *Col = NEW(LV_COLUMN);
	if (Col)
	{
		Col->mask = LVCF_FMT | LVCF_SUBITEM | LVCF_TEXT | LVCF_WIDTH;
		Col->fmt = Align;
		Col->cx = Width;
		Col->pszText = NewStr(Name);
		Col->cchTextMax = 0;
		Col->iSubItem = Columns.GetItems();

		if (hWnd)
		{
			SendMessage(hWnd, LVM_INSERTCOLUMN, Columns.GetItems(), (WPARAM) Col); 
		}

		Columns.Insert(Col);
	}
}

void GList::EmptyColumns()
{
	while (ListView_DeleteColumn(hWnd, 0))
	{
	}

	LV_COLUMN *Col;
	while (Col = Columns.First())
	{
		Columns.Delete(Col);
		DeleteObj(Col);
	}
}

void GList::ShowColumnHeader(bool Show)
{
	ColumnHeader = Show;
}

bool GList::ShowColumnHeader()
{
	return ColumnHeader;
}

bool GList::SysAddItem(GListItem *i, int Pos)
{
	bool Status = FALSE;
	if (i)
	{
		for (int Num=0; Num<i->GetCols(); Num++)
		{
			LV_ITEM Item;

			ZeroObj(Item);
			Item.mask = LVIF_TEXT | LVIF_PARAM | LVIF_STATE | LVIF_IMAGE;
			Item.iItem = Pos;
			Item.iSubItem = Num;
			Item.pszText = LPSTR_TEXTCALLBACK;
			Item.iImage = (ImageList) ? I_IMAGECALLBACK : 0;
			Item.lParam = (LPARAM) i;
			Status = SendMessage(hWnd, LVM_INSERTITEM, 0, (LPARAM) &Item);
		}
	}
	return Status;
}

void GList::SysRemoveItem(int Pos)
{
	SendMessage(hWnd, LVM_DELETEITEM, Pos, 0);
}

void GList::OnMouseClick(VMouse &m)
{
	LVHITTESTINFO Info;

	Info.pt.x = m.x;
	Info.pt.y = m.y;
	int Index = ListView_HitTest(hWnd, &Info);
	if (	(Info.flags & LVHT_ONITEM) AND
		Index >= 0)
	{
		LV_ITEM Item;

		ZeroObj(Item);
		Item.mask = LVIF_PARAM;
		Item.iItem = Index;

		if (ListView_GetItem(hWnd, &Item))
		{
			OnItemClick((GListItem*) Item.lParam, m);
		}
	}
	else
	{
		OnItemClick(NULL, m);
	}
}

int GList::OnEvent(GMessage *Msg)
{
	int OldVal = Val;
	switch (Message(Msg))
	{
		case WM_DESTROY:
		{
			if (ImageList)
			{
				ListView_SetImageList(hWnd, 0, LVSIL_SMALL);
			}
			break;
		}
	}

	int Status = WndClass.CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
	switch (Msg->Msg)
	{
		case WM_CREATE:
		{
			if (ImageList)
			{
				ListView_SetImageList(hWnd, ImageList->Handle(), LVSIL_SMALL);
			}

			int i=0;
			for (LV_COLUMN *c = Columns.First(); c; c = Columns.Next(), i++)
			{
				ListView_InsertColumn(hWnd, i, c); 
 			}

			i=0;
			for (GListItem *n = First(); n; n = Next(), i++)
			{
				SysAddItem(n, i);
			}

			Value(Val);
			OnCreate();
			break;
		}
		case WM_CHAR:
		{
			GWindow::OnEvent(Msg);
			break;
		}
		/*
		case WM_NOTIFY:
		{
			LPNMHDR Hdr = (LPNMHDR) b;
			switch (Hdr->code)
			{
				case NM_DBLCLK:
				{
					break;
				}
				case NM_CLICK:
				{
					break;
				}
			}
		}
		*/
		case WM_LBUTTONDBLCLK:
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONDBLCLK:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_MBUTTONDBLCLK:
		case WM_MBUTTONDOWN:
		case WM_MBUTTONUP:
		case WM_KEYDOWN:
		{
			GWindow::OnEvent(Msg);

			// check whether the value has changed
			/*
			INT Selected = SendMessage(hWnd, LVM_GETSELECTIONMARK, 0, 0);
			if (Selected != Val)
			{
				Val = Selected;
				if (Parent)
				{
					Parent->OnEvent(WM_CHANGE, (int)this, 0);


					LV_ITEM Item;

					ZeroObj(Item);
					Item.mask = LVIF_PARAM;
					Item.iItem = Val;

					if (ListView_GetItem(hWnd, &Item))
					{
						OnItemSelect((GListItem*) Item.lParam);
					}
				}
			}
			*/

			for (int i=0; i<GetItems(); i++)
			{
				if (SendMessage(hWnd, LVM_GETITEMSTATE, i, LVIS_FOCUSED))
				{
					if (OldVal != i)
					{
						Val = i;
						if (Parent)
						{
							Parent->OnEvent(&GMessage(WM_CHANGE, (int)this, 0));

							LV_ITEM Item;

							ZeroObj(Item);
							Item.mask = LVIF_PARAM;
							Item.iItem = Val;

							if (ListView_GetItem(hWnd, &Item))
							{
								OnItemSelect((GListItem*) Item.lParam);
							}
						}
					}
					break;
				}
			}
		}
	}
	return Status;
}

bool GList::Remove(GListItem *Obj)
{
	bool Status = FALSE;
	int Index = -1;

	if (hWnd)
	{
		LV_FINDINFO Find;
		Find.flags = LVFI_PARAM;
		Find.lParam = (LPARAM) Obj;
		Index = ListView_FindItem(hWnd, -1, &Find);
	}

	if (Index >= 0)
	{
		SysRemoveItem(Index);
		Status = TRUE;
		Obj->Parent = 0;
	}

	return Status;
}

bool GList::Delete()
{
	bool Status = FALSE;
	GListItem *Cur = (*this)[Val];
	return Delete(Cur);
}

bool GList::Delete(int i)
{
	bool Status = FALSE;
	GListItem *Item = 0;
	
	if (hWnd)
	{
		LV_ITEM it;
		it.mask = LVIF_PARAM;
		it.iItem = i;
		ListView_GetItem(hWnd, &it);
		Item = (GListItem*) it.lParam;
	}

	if (Item)
	{
		SysRemoveItem(i);
		Status = DLinkList::Delete(Item);
		DeleteObj(Item);
		Status = TRUE;
	}

	return Status;
}

bool GList::Delete(GListItem *Item)
{
	bool Status = FALSE;
	int Index = -1;

	if (hWnd)
	{
		LV_FINDINFO Find;
		Find.flags = LVFI_PARAM;
		Find.lParam = (LPARAM) Item;
		Index = ListView_FindItem(hWnd, -1, &Find);
	}

	if (Index >= 0)
	{
		SysRemoveItem(Index);
		Status = DLinkList::Delete(Item);
		DeleteObj(Item);
		Status = TRUE;
	}

	return Status;
}

bool GList::Insert(GListItem *p, int Index)
{
	bool Status = FALSE;
	if (p)
	{
		p->Parent = this;

		if (Index < 0 OR Index > GetItems())
		{
			Index = GetItems();
		}

		Status = DLinkList::Insert(p, Index);
		if (hWnd)
		{
			Status &= SysAddItem(p, Index);
		}
	}
	return Status;
}

void GList::Empty()
{
	for (GListItem *i = First(); i; i = Next())
	{
		DeleteObj(i);
	}

	List<GListItem>::Empty();

	if (hWnd)
	{
		SendMessage(hWnd, LVM_DELETEALLITEMS, 0, 0);
	}
}

void GList::RemoveAll()
{
	List<GListItem>::Empty();

	if (hWnd)
	{
		SendMessage(hWnd, LVM_DELETEALLITEMS, 0, 0);
	}
}

int GList::SysOnNotify(int Code)
{
	return 0;
}

int GList::Value()
{
	if (hWnd)
	{
		for (int i=0; i<GetItems(); i++)
		{
			if (SendMessage(hWnd, LVM_GETITEMSTATE, i, LVIS_FOCUSED))
			{
				Val = i;
				break;
			}
		}
	}
	return Val;
}

void GList::Value(int n)
{
	Val = n;

	if (hWnd)
	{
		LV_ITEM Item;

		ZeroObj(Item);
		for (int i=0; i<GetItems(); i++)
		{
			Item.mask = LVIF_STATE;
			Item.iItem = i;
			Item.stateMask = LVIS_FOCUSED | LVIS_SELECTED;

			if (i == n)
			{
				Item.state |= LVIS_FOCUSED | LVIS_SELECTED;
			}
			else
			{
				Item.state &= ~(LVIS_FOCUSED | LVIS_SELECTED);
			}

 			SendMessage(hWnd, LVM_SETITEM, 0, (LPARAM) &Item);
		}
	}
}

GListItem *GList::Current()
{
	LV_ITEM Item;
	GListItem *i = 0;

	ZeroObj(Item);
	Item.mask = LVIF_PARAM;
	Item.iItem = Val;
	if (SendMessage(hWnd, LVM_GETITEM, 0, (LPARAM) &Item))
	{
		i = (GListItem*) Item.lParam;
	}

	return i;
}

void GList::OnItemClick(GListItem *Item, VMouse &m)
{
	if (Item)
	{
		Item->OnMouseClick(m);
	}
}

void GList::OnItemBeginDrag(GListItem *Item, int Flags)
{
	if (Item)
	{
		VMouse m;
		ZeroObj(m);
		m.Flags = Flags;
		Item->OnBeginDrag(m);
	}
}

bool GList::Select(GListItem *Obj)
{
	bool Status = FALSE;
	if (Obj AND hWnd)
	{
		LV_ITEM Item;

		ZeroObj(Item);
		for (int i=0; i<GetItems(); i++)
		{
			Item.mask = LVIF_PARAM;
			Item.iItem = i;
			ListView_GetItem(hWnd, &Item);

			Item.mask = LVIF_STATE;
			Item.stateMask = LVIS_FOCUSED | LVIS_SELECTED;

			if (Obj == (GListItem*) Item.lParam)
			{
				Item.state |= LVIS_FOCUSED | LVIS_SELECTED;
			}
			else
			{
				Item.state &= ~(LVIS_FOCUSED | LVIS_SELECTED);
			}

 			ListView_SetItem(hWnd, &Item);
		}
	}
	return Status;
}

bool GList::GetAll(List<GListItem> &l)
{
	bool Status = FALSE;
	if (hWnd)
	{
		LV_ITEM Item;

		ZeroObj(Item);
		for (int i=0; i<GetItems(); i++)
		{
			Item.mask = LVIF_PARAM;
			Item.iItem = i;
 			if (ListView_GetItem(hWnd, &Item))
			{
				l.Insert((GListItem*) Item.lParam);
				Status = TRUE;
			}
		}
	}

	return Status;
}

bool GList::GetSelection(List<GListItem> &l)
{
	bool Status = FALSE;
	if (hWnd)
	{
		LV_ITEM Item;

		ZeroObj(Item);
		for (int i=0; i<GetItems(); i++)
		{
			Item.mask = LVIF_STATE | LVIF_PARAM;
			Item.iItem = i;
			Item.stateMask = LVIS_SELECTED;
 			if (SendMessage(hWnd, LVM_GETITEM, 0, (LPARAM) &Item))
			{
				if (Item.state & LVIS_SELECTED)
				{
					l.Insert((GListItem*) Item.lParam);
					Status = TRUE;
				}
			}
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////
GTreeItem::GTreeItem()
{
	hTreeItem = 0;
	Tree = 0;
	Parent = 0;
	Str = 0;
}

GTreeItem::~GTreeItem()
{
	if (Tree)
	{
		/*
		GTreeItem *Item = GetChild();
		DeleteObj(Item);
		Imte = GetNext();
		DeleteObj(Item);

		if (hTreeItem)
		{
			TreeView_DeleteItem(Tree->Handle(), hTreeItem);
		}
		*/
	}

	DeleteArray(Str);
}

GTreeItem *GTreeItem::GetChild()
{
	HTREEITEM hItem = TreeView_GetChild(Tree->Handle(), hTreeItem);
	if (hItem)
	{
		TV_ITEM Item;
		ZeroObj(Item);
		Item.mask = TVIF_PARAM;
		Item.hItem = hItem;

		if (TreeView_GetItem(Tree->Handle(), &Item))
		{
			return (GTreeItem*) Item.lParam;
		}
	}
	return 0;
}

GTreeItem *GTreeItem::GetNext()
{
	HTREEITEM hItem = TreeView_GetNextSibling(Tree->Handle(), hTreeItem);
	if (hItem)
	{
		TV_ITEM Item;
		ZeroObj(Item);
		Item.mask = TVIF_PARAM;
		Item.hItem = hItem;

		if (TreeView_GetItem(Tree->Handle(), &Item))
		{
			return (GTreeItem*) Item.lParam;
		}
	}
	return 0;
}

bool GTreeItem::SetText(char *s, int i)
{
	DeleteArray(Str);
	Str = NewStr(s);
	return TRUE;
}

GTreeItem *GTreeItem::Insert(GTreeItem *Obj, int Pos)
{
	GTreeItem *NewObj = (Obj) ? Obj : NEW(GTreeItem);
	if (NewObj)
	{
		Items.Insert(NewObj, Pos);
		NewObj->Parent = this;
		NewObj->Tree = Tree;
		NewObj->Attach(NULL);
	}

	return NewObj;
}

bool GTreeItem::Attach(GTreeItem *After)
{
	bool Status = FALSE;
	if (Tree)
	{
		TV_INSERTSTRUCT Item;

		Item.hParent = (Parent) ? Parent->ItemHandle() : TVI_ROOT;
		Item.hInsertAfter = (After) ? After->ItemHandle() : 0;
		Item.item.mask = TVIF_TEXT | TVIF_PARAM;
		Item.item.state = 0;
		Item.item.stateMask = 0;
		Item.item.pszText = LPSTR_TEXTCALLBACK;
		Item.item.cchTextMax = 0;
		Item.item.lParam = (DWORD) this;

		if (Tree AND Tree->ImageList)
		{
			Item.item.mask |= TVIF_IMAGE | TVIF_SELECTEDIMAGE;
			Item.item.iImage = I_IMAGECALLBACK;
			Item.item.iSelectedImage = I_IMAGECALLBACK;
		}

		hTreeItem = TreeView_InsertItem(Tree->Handle(), &Item);
		Status = (hTreeItem != NULL);
	}
	return Status;
}

bool GTreeItem::AttachAll()
{
	bool Status = TRUE;
	GTreeItem *Prev = 0;
	for (GTreeItem *i=Items.First(); i; i = Items.Next())
	{
		Status &= i->Attach(Prev);
		i->AttachAll();
		Prev = i;
	}

	return Status;
}

void GTreeItem::Update()
{
}

bool GTreeItem::Selected()
{
	if (Tree)
	{
		TV_ITEM Item;
		
		ZeroObj(Item);
		Item.mask = TVIF_STATE;
		Item.hItem = hTreeItem;
		Item.stateMask = TVIS_SELECTED;
		
		if (TreeView_GetItem(Tree->Handle(), &Item))
		{
			return (Item.state & TVIS_SELECTED) != 0;
		}
	}
	return FALSE;
}

void GTreeItem::Selected(bool b)
{
	if (Tree)
	{
		TV_ITEM Item;
		
		ZeroObj(Item);
		Item.mask = TVIF_STATE;
		Item.hItem = hTreeItem;
		Item.state = (b) ? TVIS_SELECTED : 0;
		Item.stateMask = TVIS_SELECTED;
		
		TreeView_SetItem(Tree->Handle(), &Item);
	}
}

bool GTreeItem::Expanded()
{
	if (Tree)
	{
		TV_ITEM It;

		ZeroObj(It);
		It.mask = TVIF_STATE;
		It.stateMask = TVIS_EXPANDED;
		It.hItem = hTreeItem;
		if (TreeView_GetItem(Tree->Handle(), &It))
		{
			return (It.state & TVIS_EXPANDED) != 0;
		}
	}

	return FALSE;
}

void GTreeItem::Expanded(bool b)
{
	if (Tree)
	{
		TreeView_Expand(Tree->Handle(), hTreeItem, (b) ? TVE_EXPAND : TVE_COLLAPSE); 
	}
}

//////////////////////////////////////////////////////////////////////////////
GWin32Class GTree::WndClass;

GTree::GTree(int id, int x, int y, int cx, int cy, char *name)
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	if (name) GControl::Name(name);

	Lines = TRUE;
	Buttons = TRUE;
	LinesAtRoot = TRUE;
	EditLabels = FALSE;
	Tree = this;

	WndClass.SubClass(Class(), WC_TREEVIEW);
}

GTree::~GTree()
{
	/*
	if (hWnd)
	{
		HTREEITEM hItem = TreeView_GetRoot(hWnd);
		TV_ITEM Item;

		ZeroObj(Item);
		Item.mask = TVIF_PARAM;
		Item.hItem = hItem;
		if (hItem AND TreeView_GetItem(hWnd, &Item))
		{
			GTreeItem *i = (GTreeItem*) Item.lParam;
			if (i)
			{
				DeleteObj(i);
			}
		}
	}
	*/
}

void GTree::OnMouseClick(VMouse &m)
{
	TV_HITTESTINFO Info;

	Info.pt.x = m.x;
	Info.pt.y = m.y;
 
 	HTREEITEM hItem = TreeView_HitTest(hWnd, &Info);

	if (Info.flags & TVHT_ONITEM)
	{
		// over item
		TV_ITEM Item;

		ZeroObj(Item);
		Item.mask = TVIF_HANDLE | TVIF_PARAM;
		Item.hItem = Info.hItem;

		if (TreeView_GetItem(hWnd, &Item))
		{
			OnItemClick((GTreeItem*) Item.lParam, m);
		}		
	}
}

int GTree::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
		case WM_DESTROY:
		{
			if (ImageList)
			{
				TreeView_SetImageList(hWnd, 0, LVSIL_SMALL);
			}
			break;
		}
	}

	GControl::OnEvent(Msg);
	int Status = WndClass.CallParent(hWnd, Msg->Msg, Msg->a, Msg->b);

	switch (Msg->Msg)
	{
		case WM_CREATE:
		{
			if (ImageList)
			{
				TreeView_SetImageList(hWnd, ImageList->Handle(), TVSIL_NORMAL);
			}
			AttachAll();
			OnCreate();
			break;
		}
		case WM_LBUTTONDBLCLK:
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONDBLCLK:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_MBUTTONDBLCLK:
		case WM_MBUTTONDOWN:
		case WM_MBUTTONUP:
		{
			GWindow::OnEvent(Msg);
			break;
		}
	}

	return Status;
}

GTreeItem *GTree::Insert(GTreeItem *Obj, int Pos)
{
	GTreeItem *NewObj = GTreeItem::Insert(Obj, Pos);
	if (NewObj)
	{
		NewObj->Tree = this;
	}
	return NewObj;
}

bool GTree::Remove(GTreeItem *Obj)
{
	if (hWnd AND Obj)
	{
		return TreeView_DeleteItem(hWnd, Obj->hTreeItem); 
	}
	return FALSE;
}

void GTree::RemoveAll()
{
	if (hWnd)
	{
		TreeView_DeleteAllItems(hWnd); 
	}
}

bool GTree::Delete(GTreeItem *Obj)
{
	bool Status = FALSE;
	if (Obj AND hWnd)
	{
		// Status = TreeView_DeleteItem(hWnd, Obj->ItemHandle());
		Status = TRUE;
		DeleteObj(Obj);
	}
	return Status;
}

bool GTree::Select(GTreeItem *Obj)
{
	bool Status = FALSE;
	if (Obj AND hWnd)
	{
		TreeView_SelectItem(hWnd, Obj->ItemHandle());
		Status = TRUE;
	}
	return Status;
}

void GTree::OnItemClick(GTreeItem *Item, VMouse &m)
{
	if (Item)
	{
		Item->OnMouseClick(m);
	}
}

void GTree::OnItemBeginDrag(GTreeItem *Item, int Flags)
{
	if (Item)
	{
		VMouse m;
		ZeroObj(m);
		m.Flags = Flags;
		Item->OnBeginDrag(m);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
GItemContainer::GItemContainer()
{
	Flags = 0;
	ImageList = 0;
}

GItemContainer::~GItemContainer()
{
	if (OwnList())
	{
		DeleteObj(ImageList);
	}
	else
	{
		ImageList = 0;
	}
}

bool GItemContainer::SetImageList(GImageList *list, bool Own)
{
	ImageList = list;
	OwnList(Own);
	return ImageList != NULL;
}

bool GItemContainer::Load(char *File)
{
	bool Status = FALSE;
	GdcBmp f;
	if (f.Open(File, O_READ))
	{
		GdcDibSection *pDC = NEW(GdcDibSection);
		if (pDC)
		{
			Status = f.ReadImage(pDC);
		}
	}
	return Status;
}

///////////////////////////////////////////////////////////////////////////////////////////
GWin32Class GProgress::WndClass;

GProgress::GProgress(int id, int x, int y, int cx, int cy, char *name)
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	if (name) GControl::Name(name);

	WndClass.SubClass(Class(), PROGRESS_CLASS);
}

GProgress::~GProgress()
{
}

void GProgress::SetLimits(int l, int h)
{
	Low = l;
	High = h;
	if (hWnd)
	{
		SendMessage(hWnd, PBM_SETRANGE, 0, MAKELPARAM(Low, High));
	}
}

void GProgress::Value(int v)
{
	Val = v;
	if (hWnd)
	{
		SendMessage(hWnd, PBM_SETPOS, (WPARAM) Val, 0);
	}
}

int GProgress::OnEvent(GMessage *Msg)
{
	return WndClass.CallParent(hWnd, Msg->Msg, Msg->a, Msg->b);
}


